package mrl.ambulance;

import javolution.util.FastMap;
import mrl.ambulance.marketLearnerStrategy.AmbulanceConditionChecker;
import mrl.ambulance.marketLearnerStrategy.AmbulanceUtilities;
import mrl.ambulance.structures.CivilianValue;
import mrl.common.MRLConstants;
import mrl.common.Util;
import mrl.partitioning.DefaultAmbulancePartitionManager;
import mrl.partitioning.Partition;
import mrl.world.MrlWorld;
import rescuecore2.misc.Pair;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.worldmodel.EntityID;

import java.util.*;

/**
 * Author: Pooya Deldar Gohardani
 * Date: 6/18/12
 * Time: 9:15 AM
 */
public class PartitionBasedCoordination {

    private MrlWorld world;
    private AmbulanceUtilities ambulanceUtilities;
    private AmbulanceConditionChecker conditionChecker;
    private DefaultAmbulancePartitionManager ambulancePartitionManager;
    private Human myTarget;
    private Partition myPartition;
    private Pair<Integer, Integer> pairPosition = null;

    public PartitionBasedCoordination(MrlWorld world, AmbulanceUtilities ambulanceUtilities, AmbulanceConditionChecker conditionChecker) {
        this.world = world;
        this.ambulanceUtilities = ambulanceUtilities;
        this.conditionChecker = conditionChecker;

        ambulancePartitionManager = new DefaultAmbulancePartitionManager(world, ambulanceUtilities);
        ambulancePartitionManager.initialise();
        world.setPartitionManager(ambulancePartitionManager);

        myPartition = ambulancePartitionManager.findHumanPartition(world.getSelfHuman());


        if (MRLConstants.LAUNCH_VIEWER && myPartition != null) {
            System.out.println(world.getSelf().getID() + " myPartition is:" + myPartition.getId());
        }


    }


    public Human getNextTarget(List<StandardEntity> goodHumans) {

        if (!goodHumans.contains(myTarget)) {
            myTarget = null;
        }

        myTarget = (Human) selectTarget(goodHumans, myTarget);
        return myTarget;
    }

    private StandardEntity selectTarget(List<StandardEntity> myGoodHumans, StandardEntity myTarget) {
        if (myTarget != null) {
            Human human = (Human) myTarget;
            if (conditionChecker.isPassable(world.getSelfPosition().getID(), human.getPosition())) {
                return myTarget;
            } else {
                myGoodHumans.remove(myTarget);
            }
        }
        if (myGoodHumans.isEmpty()) {
            return null;
        }


        ArrayList<CivilianValue> civilianTTAs;
        //look in partition
        Map<EntityID, List<StandardEntity>> partitionVictimMap = findVictimPartitionMap(myGoodHumans, true);

        StandardEntity victimToRescue = findVictimToRescue(partitionVictimMap, myGoodHumans);

//        System.out.println(world.getPlatoonAgent().getDebugString() + " victimToRescue: " + victimToRescue);


        return victimToRescue;


    }

    private StandardEntity findVictimToRescue(Map<EntityID, List<StandardEntity>> partitionVictimMap, List<StandardEntity> myGoodHumans) {

        List<StandardEntity> victims;
        StandardEntity bestVictim = null;
        if (myPartition == null) {
            bestVictim = findBestPartitionToGo(partitionVictimMap, myGoodHumans);
        } else {
            victims = partitionVictimMap.get(myPartition.getId());
            if (victims == null || victims.isEmpty()) {
//                System.out.println(world.getPlatoonAgent().getDebugString() + " myPartitionVictims are empty ");
                bestVictim = findBestPartitionToGo(partitionVictimMap, myGoodHumans);
            } else {

                bestVictim = findBestVictim(victims);
            }
        }

        return bestVictim;
    }

    private StandardEntity findBestVictim(List<StandardEntity> victims) {

        StandardEntity victimEntityToRescue = null;
        if (victims == null || victims.isEmpty()) {
            victimEntityToRescue = null;
        } else {

            Collections.sort(victims, VICTIM_BURIENDNESS_COMPARATOR);

            EntityID selfPositionID = world.getSelfPosition().getID();
            Human victim;

            for (StandardEntity victimEntity : victims) {
                victim = (Human) victimEntity;
                if (conditionChecker.isPassable(selfPositionID, victim.getPosition())) {
                    victimEntityToRescue = victimEntity;
                    break;
                }
            }

        }
        return victimEntityToRescue;
    }

    private StandardEntity findBestPartitionToGo(Map<EntityID, List<StandardEntity>> partitionVictimMap, List<StandardEntity> myGoodHumans) {
        List<StandardEntity> workerAmbulances;
        List<StandardEntity> victims;
        Partition partitionToGo = null;
        int maxFoundCivilians = myGoodHumans.size();
        int maxAvailableAgents = ambulanceUtilities.getReadyAmbulances().size();
        int numberOfWorkers;
        int numberOfVictims;
        double value;
        double normalRatio = (double) maxFoundCivilians / (double) (maxAvailableAgents + 1);


//        System.out.println(world.getPlatoonAgent().getDebugString() + " normalRatio: " + normalRatio);


        List<Partition> partitions = new ArrayList<Partition>(world.getPartitionManager().getPartitions());
        Map<EntityID, List<StandardEntity>> partitionAmbulanceMap = findPartitionAmbulanceMap();


        for (Partition partition : partitions) {
            workerAmbulances = partitionAmbulanceMap.get(partition.getId());
            victims = partitionVictimMap.get(partition.getId());

            if (workerAmbulances == null || workerAmbulances.isEmpty()) {
                numberOfWorkers = 0;
            } else {
                numberOfWorkers = workerAmbulances.size();
            }

            if (victims == null || victims.isEmpty()) {
                numberOfVictims = 0;
            } else {
                numberOfVictims = victims.size();
            }


            value = computePartitionValue(numberOfWorkers, numberOfVictims);
//            System.out.println(world.getPlatoonAgent().getDebugString() + "partition:" + partition.getId() + " partitionValue: " + value);

            partition.setValue(value);
        }


        Collections.sort(partitions, DISTANCE_TO_PARTITION_COMPARATOR);
        StandardEntity bestVictim = null;
        StandardEntity tempVictim = null;
        int minNeededAgents;

        for (Partition partition : partitions) {
            if (partition.getValue() > normalRatio) {

                victims = partitionVictimMap.get(partition.getId());
                while (!victims.isEmpty()) {
                    tempVictim = findBestVictim(victims);
                    if (tempVictim == null) {
                        break;
                    }
                    minNeededAgents =ambulanceUtilities.computeMinimumNeededAgent((Human)tempVictim);

                    if (shouldBeSelected(tempVictim, minNeededAgents)) {
//                        System.out.println(world.getPlatoonAgent().getDebugString()+" ID: "+tempVictim.getID()+" min:"+minNeededAgents);
                        bestVictim=tempVictim;
                        break;
                    }

                    victims.remove(tempVictim);
                }
                if (bestVictim != null) {
                    break;
                }
            }
        }

        return bestVictim;
    }

    private boolean shouldBeSelected(StandardEntity bestVictim, int minNeededAgents) {

        List<StandardEntity> agents = new ArrayList<StandardEntity>(ambulanceUtilities.getReadyAmbulances());
        pairPosition = bestVictim.getLocation(world);


        Collections.sort(agents, VICTIM_DISTANCE_COMPARATOR);



        if (agents.indexOf(world.getEntity(world.getSelf().getID())) < minNeededAgents) {
//            for (StandardEntity entity:agents){
////                System.out.println(" distance "+world.getPlatoonAgent().getDebugString()+" id:"+entity.getID());
//            }

            return true;
        } else {
            return false;
        }

    }

    private Map<EntityID, List<StandardEntity>> findPartitionAmbulanceMap() {

        Map<EntityID, List<StandardEntity>> partitionAmbulanceMap = new FastMap<EntityID, List<StandardEntity>>();

        Partition partition;
        Human human;
        List<StandardEntity> ambulanceList;
        for (StandardEntity entity : ambulanceUtilities.getReadyAmbulances()) {

            human = (Human) entity;
            partition = world.getPartitionManager().findPartitionAtArea(human.getPosition());
            if (partition == null) {
                // do nothing
            } else {
                ambulanceList = partitionAmbulanceMap.get(partition.getId());
                if (ambulanceList == null) {
                    ambulanceList = new ArrayList<StandardEntity>();
                }
                ambulanceList.add(entity);
                partitionAmbulanceMap.put(partition.getId(), ambulanceList);
            }

        }

        return partitionAmbulanceMap;

    }


    private double computePartitionValue(int workerAmbulances, int victims) {
        double ratio;
        if (workerAmbulances == 0) {
            ratio = 0;
        } else {
            ratio = (double) victims / (double) workerAmbulances;
        }
        return ratio;
    }


    private Map<EntityID, List<StandardEntity>> findVictimPartitionMap(List<StandardEntity> myGoodHumans, boolean inPartition) {
        int tta;
        Pair<Integer, Integer> ttd;
        ArrayList<CivilianValue> civilianTTAs = new ArrayList<CivilianValue>();
        CivilianValue civilianTTA;

        // a map of each partition and victims in it
        Map<EntityID, List<StandardEntity>> partitionVictimMap = new FastMap<EntityID, List<StandardEntity>>();

        Human human;
        Partition partition;
        List<StandardEntity> victims;
        for (StandardEntity victimEntity : myGoodHumans) {
            human = (Human) victimEntity;
            partition = world.getPartitionManager().findPartitionAtArea(human.getPosition());
            if (partition == null) {
                // so where the hell is this victim !!?
            } else {
                victims = partitionVictimMap.get(partition.getId());
                if (victims == null) {
                    victims = new ArrayList<StandardEntity>();
                }
                victims.add(victimEntity);
                partitionVictimMap.put(partition.getId(), victims);
            }
        }


        return partitionVictimMap;
    }


    public Comparator<Partition> DISTANCE_TO_PARTITION_COMPARATOR = new Comparator<Partition>() {
        public int compare(Partition r1, Partition r2) {

            int firstDistance = Util.distance(r1.getCenter(), world.getSelfLocation());
            int secondDistance = Util.distance(r2.getCenter(), world.getSelfLocation());

            if (firstDistance > secondDistance)
                return 1;
            if (firstDistance == secondDistance)
                return 0;

            return -1;
        }
    };


    public Comparator<StandardEntity> VICTIM_BURIENDNESS_COMPARATOR = new Comparator<StandardEntity>() {
        public int compare(StandardEntity r1, StandardEntity r2) {

            Human h1 = (Human) r1;
            Human h2 = (Human) r2;

            if (h1.getBuriedness() > h2.getBuriedness())
                return 1;
            if (h1.getBuriedness() == h2.getBuriedness())
                return 0;

            return -1;
        }
    };


    public Comparator<StandardEntity> VICTIM_DISTANCE_COMPARATOR = new Comparator<StandardEntity>() {
        public int compare(StandardEntity r1, StandardEntity r2) {

            Human h1 = (Human) r1;
            Human h2 = (Human) r2;
            int firstDistance = Util.distance(h1.getLocation(world), pairPosition);
            int secondDistance = Util.distance(h2.getLocation(world), pairPosition);

            if (firstDistance > secondDistance)
                return 1;
            if (firstDistance == secondDistance)
                return 0;

            return -1;
        }
    };

}
