package mrl.common;

import mrl.world.MrlWorld;
import mrl.world.object.Wall;
import rescuecore2.misc.Pair;
import rescuecore2.misc.geometry.GeometryTools2D;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;
import sun.reflect.generics.reflectiveObjects.NotImplementedException;

import java.awt.*;
import java.awt.Point;
import java.awt.geom.Line2D;
import java.awt.geom.PathIterator;
import java.awt.geom.Point2D;
import java.io.*;
import java.util.*;
import java.util.List;

/**
 * User: mrl
 * Date: May 5, 2010
 * Time: 2:17:52 PM
 */
public class Util {

    public static Point getPoint(String value) {
        String[] values = value.split(",");
        return new Point(Integer.parseInt(values[0]), Integer.parseInt(values[1]));

    }

    public static int getConditionCount(Collection col, Condition cond) {
        int res = 0;
        for (Object aCol : col) {
            if (cond.eval(aCol)) res++;
        }
        return res;
    }

    public static int distance(int x1, int y1, int x2, int y2) {
        float dx = x1 - x2;
        float dy = y1 - y2;
        return (int) Math.sqrt(dx * dx + dy * dy);
    }

    public static double distance(double x1, double y1, double x2, double y2) {
        double dx = x1 - x2;
        double dy = y1 - y2;
        return (double) Math.sqrt(dx * dx + dy * dy);
    }

    public static int distance(Point p1, Point p2) {
        double dx = p1.getX() - p2.getX();
        double dy = p1.getY() - p2.getY();
        return (int) Math.sqrt(dx * dx + dy * dy);
    }

    public static int distance(Point2D p1, Point2D p2) {
        double dx = p1.getX() - p2.getX();
        double dy = p1.getY() - p2.getY();
        return (int) Math.sqrt(dx * dx + dy * dy);
    }

    public static int distance(rescuecore2.misc.geometry.Point2D start, rescuecore2.misc.geometry.Point2D end) {
        double dx = start.getX() - end.getX();
        double dy = start.getY() - end.getY();
        return (int) Math.sqrt(dx * dx + dy * dy);
    }

    public static int distance(Point point, Pair<Integer, Integer> pair) {
        double dx = point.getX() - pair.first();
        double dy = point.getY() - pair.second();
        return (int) Math.sqrt(dx * dx + dy * dy);
    }

    public static int distance(Pair<Integer, Integer> pair, Point point) {
        double dx = point.getX() - pair.first();
        double dy = point.getY() - pair.second();
        return (int) Math.sqrt(dx * dx + dy * dy);
    }

    public static int max(int x, int y) {
        return x >= y ? x : y;
    }

    public static int min(int x, int y) {
        return x <= y ? x : y;
    }

    public static boolean Position(double x0, double y0, double ang, double tg, Wall edg) {
        double a = edg.x1;
        double b = edg.y1;
        double c = edg.x2;
        double d = edg.y2;

        if (ang == 0f || ang == Math.PI) {
            if (isBetween(a, x0, c)) {
                if (a == c) {
                    edg.distance = Util.distance((int) x0, (int) y0, (int) a, (int) b);
                    if (a > x0) {
                        edg.right = true;

                    } else {
                        edg.right = false;
                    }
                } else {
                    edg.distance = Util.distance((int) x0, (int) y0, (int) x0, (int) ((b - d) * (x0 - c) / (a - c) + d));
                    edg.right = false;
                }
                return true;
            }
            edg.distance = 0;
            return false;
        } else {
            if (a == c) {
                double y = tg * (a - x0) + y0;
                if (isBetween(b, y, d)) {
                    edg.distance = Util.distance((int) x0, (int) y0, (int) a, (int) y);
                    if (a > x0) {
                        edg.right = true;
                    } else {
                        edg.right = false;
                    }
                    return true;
                }
                edg.distance = 0;
                return false;
            }
            if (a * tg - c * tg + d - b == 0) {
                edg.distance = 0;
                return false;
            }
            double x = (-b * c + d * a + c * y0 + a * tg * x0 - a * y0 - c * tg * x0) / (a * tg - c * tg + d - b);
            if (isBetween(a, x, c)) {
                if (a == c) {
                    edg.distance = Util.distance((int) x0, (int) y0, (int) a, (int) b);
                    if (a > x0) {
                        edg.right = true;
                    } else {
                        edg.right = false;
                    }
                } else {
                    edg.distance = Util.distance((int) x0, (int) y0, (int) x, (int) ((b - d) * (x - c) / (a - c) + d));
                    if (x > x0) {
                        edg.right = true;
                    } else {
                        edg.right = false;
                    }
                }
                return true;
            }
            edg.distance = 0;
            return false;
        }
    }


    public static boolean isBetween(double a, double x, double b) {
        if (a > b) {
            double k = a;
            a = b;
            b = k;
        }
        if (x >= a && x <= b) {
            return true;
        }
        return false;
    }

    public static int distance(Area obj1, Area obj2) {
        return distance(obj1.getX(), obj1.getY(), obj2.getX(), obj2.getY());
    }

    public static int distance(Pair<Integer, Integer> obj1, Pair<Integer, Integer> obj2) {
        return distance(obj1.first(), obj1.second(), obj2.first(), obj2.second());
    }

//    public static WithCoordinates nearestTo(Collection<? extends WithCoordinates> collection,WithCoordinates from){
//        int minDistance=Integer.MAX_VALUE;
//        WithCoordinates nearest=null;
//        for(WithCoordinates obj:collection){
//            if(distance(from,obj)<minDistance){
//                minDistance=distance(from,obj);
//                nearest=obj;
//            }
//        }
//        return nearest;
//    }

    public static Area nearestTo(Collection<Area> collection, Pair<Integer, Integer> from) {
        int minDistance = Integer.MAX_VALUE;
        Area nearest = null;
        int tempDist;

        for (Area obj : collection) {

            tempDist = distance(from.first(), from.second(), obj.getX(), obj.getY());
            if (tempDist < minDistance) {
                minDistance = tempDist;
                nearest = obj;
            }
        }
        return nearest;
    }

    public static StandardEntity nearestToCenterOfPolygon(MrlWorld world, Collection<StandardEntity> collection, Polygon polygon, Pair<Integer, Integer> from) {
        int minDistance = Integer.MAX_VALUE;
        StandardEntity nearest = null;
        int tempDist;

        for (StandardEntity entity : collection) {

            if (!(entity instanceof Area) || !polygon.contains(new Point(entity.getLocation(world).first(), entity.getLocation(world).second())))
                continue;
            tempDist = distance(from.first(), from.second(), entity.getLocation(world).first(), entity.getLocation(world).second());
            if (tempDist < minDistance) {
                minDistance = tempDist;
                nearest = entity;
            }
        }
        return nearest;
    }

    public static <T> java.util.List<T> sortByValueInc(final Map<T, Integer> map) {
        java.util.List<T> keys = new ArrayList<T>();
        keys.addAll(map.keySet());
        Collections.sort(keys, new Comparator<T>() {
            public int compare(T o1, T o2) {
                Integer v1 = map.get(o1);
                Integer v2 = map.get(o2);
                return v1.compareTo(v2);
            }
        });
        return keys;
    }

    public static <T> java.util.List<T> sortByValueDec(final Map<T, Integer> map) {
        java.util.List<T> keys = new ArrayList<T>();
        keys.addAll(map.keySet());
        Collections.sort(keys, new Comparator<T>() {
            public int compare(T o1, T o2) {
                Integer v1 = map.get(o1);
                Integer v2 = map.get(o2);
                return v2.compareTo(v1);
            }
        });
        return keys;
    }

    public static Object readObject(String filePath) throws IOException, ClassNotFoundException {
        ObjectInputStream objInput = new ObjectInputStream(new FileInputStream(filePath));
        Object o = objInput.readObject();
        objInput.close();
        return o;
    }

    public static void writeObject(Object object, String filePath) throws IOException {
        ObjectOutputStream objectOutputStream = new ObjectOutputStream(new FileOutputStream(filePath));
        objectOutputStream.writeObject(object);
        objectOutputStream.close();

    }

    private static double gaussmf(double x, double sig, double c) {
        return Math.exp(-((x - c) * (x - c)) / (2.0 * sig * sig));
    }

    public static double gauss2mf(double x, double sig1, double c1, double sig2, double c2) {
        if (x <= c1) {
            return gaussmf(x, sig1, c1);
        } else {
            return gaussmf(x, sig2, c2);
        }
    }

    /**
     * Returns a list of {@link rescuecore2.standard.entities.Road} containing roads that ends to {@code building}
     *
     * @param world    the model
     * @param building building to find entrance roads
     * @return List of entrance roads
     */
    public static List<Road> getEntranceRoads(MrlWorld world, Building building) {
        ArrayList<Road> entranceRoads = new ArrayList<Road>();
        if (building != null && building.getNeighbours() != null) {
            for (EntityID entityID : building.getNeighbours()) {
                Area area = (Area) world.getEntity(entityID);
                if (area instanceof Road) {
                    entranceRoads.add((Road) area);
                }
            }
        }
        return entranceRoads;
    }

    public static List<com.poths.rna.data.Point> getPointList(int[] apexList) {

        List<com.poths.rna.data.Point> points = new ArrayList<com.poths.rna.data.Point>();
        for (int i = 0; i < apexList.length; i += 2) {
            points.add(new com.poths.rna.data.Point(apexList[i], apexList[i + 1]));
        }

        return points;
    }

    public static List<com.poths.rna.data.Point> getPointList(int[] xs, int[] ys) {

        List<com.poths.rna.data.Point> points = new ArrayList<com.poths.rna.data.Point>();
        for (int i = 0; i < xs.length; i++) {
            points.add(new com.poths.rna.data.Point(xs[i], ys[i]));
        }

        return points;
    }


    public static List<rescuecore2.misc.geometry.Point2D> getPoint2DList(int[] xs, int[] ys) {

        List<rescuecore2.misc.geometry.Point2D> points = new ArrayList<rescuecore2.misc.geometry.Point2D>();
        for (int i = 0; i < xs.length; i++) {
            points.add(new rescuecore2.misc.geometry.Point2D(xs[i], ys[i]));
        }

        return points;
    }


    /**
     * @param poly main cluster Polygon
     * @param line sub cluster polygon lines
     * @return Set of intersect Point
     */
    public static Set<Point2D> getIntersections(final Polygon poly, final Line2D line)/* throws Exception */ {

        final PathIterator polyIt = poly.getPathIterator(null); //Getting an iterator along the polygon path
        final double[] coords = new double[6]; //Double array with length 6 needed by iterator
        final double[] firstCoords = new double[2]; //First point (needed for closing polygon path)
        final double[] lastCoords = new double[2]; //Previously visited point
        final Set<Point2D> intersections = new HashSet<Point2D>(); //List to hold found intersections
        polyIt.currentSegment(firstCoords); //Getting the first coordinate pair
        lastCoords[0] = firstCoords[0]; //Priming the previous coordinate pair
        lastCoords[1] = firstCoords[1];
        polyIt.next();
        while (!polyIt.isDone()) {
            final int type = polyIt.currentSegment(coords);
            switch (type) {
                case PathIterator.SEG_LINETO: {
                    final Line2D.Double currentLine = new Line2D.Double(lastCoords[0], lastCoords[1], coords[0], coords[1]);
                    if (currentLine.intersectsLine(line)) {
                        intersections.add(getIntersection(currentLine, line));
                    }
                    lastCoords[0] = coords[0];
                    lastCoords[1] = coords[1];
                    break;
                }
                case PathIterator.SEG_CLOSE: {
                    final Line2D.Double currentLine = new Line2D.Double(coords[0], coords[1], firstCoords[0], firstCoords[1]);
                    if (currentLine.intersectsLine(line)) {
                        intersections.add(getIntersection(currentLine, line));
                    }
                    break;
                }
                default: {
                    throw new NoSuchElementException("Unsupported PathIterator segment type.");
                }
            }
            polyIt.next();
        }
        return intersections;

    }

//    public static List<rescuecore2.misc.geometry.Point2D> getIntersections(final Polygon poly, final rescuecore2.misc.geometry.Line2D line){
//        List<rescuecore2.misc.geometry.Point2D> intersections = new ArrayList<rescuecore2.misc.geometry.Point2D>();
//        List<rescuecore2.misc.geometry.Line2D> polyLines = getLines(poly);
//        for (rescuecore2.misc.geometry.Line2D polyLine : polyLines){
//            rescuecore2.misc.geometry.Point2D intersect =  polyLine.getIntersection(line);
//        }
//    }


    /**
     * this function calculate the intersect point of polygon and line
     *
     * @param line1 main cluster polygon line
     * @param line2 sub cluster polygon line
     * @return the intersection of two lines point
     */
    public static Point2D getIntersection(final Line2D line1, final Line2D line2) {

        final double x1, y1, x2, y2, x3, y3, x4, y4;
        x1 = line1.getX1();
        y1 = line1.getY1();
        x2 = line1.getX2();
        y2 = line1.getY2();
        x3 = line2.getX1();
        y3 = line2.getY1();
        x4 = line2.getX2();
        y4 = line2.getY2();
        final double x = ((x2 - x1) * (x3 * y4 - x4 * y3) - (x4 - x3) * (x1 * y2 - x2 * y1))
                / ((x1 - x2) * (y3 - y4) - (y1 - y2) * (x3 - x4));
        final double y = ((y3 - y4) * (x1 * y2 - x2 * y1) - (y1 - y2) * (x3 * y4 - x4 * y3))
                / ((x1 - x2) * (y3 - y4) - (y1 - y2) * (x3 - x4));

        return new Point2D.Double(x, y);

    }

    public static rescuecore2.misc.geometry.Point2D getIntersection(final rescuecore2.misc.geometry.Line2D line1, final rescuecore2.misc.geometry.Line2D line2) {

        final double x1, y1, x2, y2, x3, y3, x4, y4;
        x1 = line1.getOrigin().getX();
        y1 = line1.getOrigin().getY();
        x2 = line1.getEndPoint().getX();
        y2 = line1.getEndPoint().getY();
        x3 = line2.getOrigin().getX();
        y3 = line2.getOrigin().getY();
        x4 = line2.getEndPoint().getX();
        y4 = line2.getEndPoint().getY();
        final double x = ((x2 - x1) * (x3 * y4 - x4 * y3) - (x4 - x3) * (x1 * y2 - x2 * y1))
                / ((x1 - x2) * (y3 - y4) - (y1 - y2) * (x3 - x4));
        final double y = ((y3 - y4) * (x1 * y2 - x2 * y1) - (y1 - y2) * (x3 * y4 - x4 * y3))
                / ((x1 - x2) * (y3 - y4) - (y1 - y2) * (x3 - x4));

        return new rescuecore2.misc.geometry.Point2D(x, y);

    }


    public static boolean isOnBlockade(MrlWorld world, Human human) {
        if (human != world.getSelfHuman() && !human.isPositionDefined()) {
            return false;
        }
        StandardEntity se = world.getEntity(human.getPosition());
        if (se instanceof Road) {
            Blockade blockade;
            Road road = (Road) se;
            Shape shape;
            if (road.isBlockadesDefined()) {
                for (EntityID id : road.getBlockades()) {
                    blockade = (Blockade) world.getEntity(id);
                    if (blockade != null && blockade.isApexesDefined()) {
                        shape = Util.getPolygon(blockade.getApexes());
                        if (shape.contains(human.getX(), human.getY())) {
                            return true;
                        }
                    }
                }
            }
        }
        return false;

    }

    public static boolean isNearBlockade(MrlWorld world, Human human) {
        if (human != world.getSelfHuman() && !human.isPositionDefined()) {
            return false;
        }
        StandardEntity se = world.getEntity(human.getPosition());
        if (se instanceof Road) {
            Blockade blockade;
            Road road = (Road) se;
            int distance;
            if (road.isBlockadesDefined()) {
                for (EntityID id : road.getBlockades()) {
                    blockade = (Blockade) world.getEntity(id);
                    if (blockade != null && blockade.isApexesDefined()) {
                        distance = findDistanceTo(blockade, human.getX(), human.getY());
                        if (distance < 500) {
                            return true;
                        }
                    }
                }
            }
        }
        return false;

    }

    public static int findDistanceTo(Blockade b, int x, int y) {
        //        Logger.debug("Finding distance to " + b + " from " + x + ", " + y);
        List<rescuecore2.misc.geometry.Line2D> lines = GeometryTools2D.pointsToLines(GeometryTools2D.vertexArrayToPoints(b.getApexes()), true);
        double best = Double.MAX_VALUE;
        rescuecore2.misc.geometry.Point2D origin = new rescuecore2.misc.geometry.Point2D(x, y);
        for (rescuecore2.misc.geometry.Line2D next : lines) {
            rescuecore2.misc.geometry.Point2D closest = GeometryTools2D.getClosestPointOnSegment(next, origin);
            double d = GeometryTools2D.getDistance(origin, closest);
            //            Logger.debug("Next line: " + next + ", closest point: " + closest + ", distance: " + d);
            if (d < best) {
                best = d;
                //                Logger.debug("New best distance");
            }

        }
        return (int) best;
    }

    public static int findDistanceTo(Area area, int x, int y) {
        //        Logger.debug("Finding distance to " + b + " from " + x + ", " + y);
        List<rescuecore2.misc.geometry.Line2D> lines = GeometryTools2D.pointsToLines(GeometryTools2D.vertexArrayToPoints(area.getApexList()), true);
        double best = Double.MAX_VALUE;
        rescuecore2.misc.geometry.Point2D origin = new rescuecore2.misc.geometry.Point2D(x, y);
        for (rescuecore2.misc.geometry.Line2D next : lines) {
            rescuecore2.misc.geometry.Point2D closest = GeometryTools2D.getClosestPointOnSegment(next, origin);
            double d = GeometryTools2D.getDistance(origin, closest);
            //            Logger.debug("Next line: " + next + ", closest point: " + closest + ", distance: " + d);
            if (d < best) {
                best = d;
                //                Logger.debug("New best distance");
            }

        }
        return (int) best;
    }


    public static boolean isOnBlockade(MrlWorld world) {
        return isOnBlockade(world, world.getSelfHuman());
    }


    public static Polygon getPolygon(int[] apexes) {
        Polygon polygon = new Polygon();
        for (int i = 0; i < apexes.length; i += 2) {
            polygon.addPoint(apexes[i], apexes[i + 1]);
        }

        return polygon;
    }

    public static rescuecore2.misc.geometry.Point2D getMiddle(rescuecore2.misc.geometry.Point2D start, rescuecore2.misc.geometry.Point2D end) {

        double cx = ((start.getX() + end.getX()) / 2);
        double cy = ((start.getY() + end.getY()) / 2);

        return new rescuecore2.misc.geometry.Point2D(cx, cy);

    }

    public static rescuecore2.misc.geometry.Point2D getMiddle(rescuecore2.misc.geometry.Line2D line) {
        return getMiddle(line.getOrigin(), line.getEndPoint());
    }

    public static com.poths.rna.data.Point getMiddle(com.poths.rna.data.Point start, com.poths.rna.data.Point end) {
        double cx = ((start.getX() + end.getX()) / 2);
        double cy = ((start.getY() + end.getY()) / 2);

        return new com.poths.rna.data.Point(cx, cy);
    }

    public static double lineSegmentAndPointDistance(Line2D line, Point2D point) {
        return Line2D.ptSegDist(line.getP1().getX(), line.getP1().getY(), line.getP2().getX(), line.getP2().getY(), point.getX(), point.getY());
    }

    public static double lineSegmentAndPointDistance(rescuecore2.misc.geometry.Line2D line, rescuecore2.misc.geometry.Point2D point) {
        return Line2D.ptSegDist(line.getOrigin().getX(), line.getOrigin().getY(), line.getEndPoint().getX(), line.getEndPoint().getY(), point.getX(), point.getY());
    }

    public static double edgeAndPointDistance(Edge edge, Point2D point) {
        return Line2D.ptSegDist(edge.getStartX(), edge.getStartY(), edge.getEndX(), edge.getEndY(), point.getX(), point.getY());
    }

    /**
     * MTN
     * get closest point of line from a point
     *
     * @param line  target point that we want get nearest point of it from another point
     * @param point a point2D
     * @return Point2D
     */
    public static Point2D closestPoint(Line2D line, Point2D point) {
//        return GeometryTools2D.getClosestPoint(line,point);
        return null;
//        double slope = slope(line);
//        if(Double.isInfinite(slope)){
//            slope = Double.MAX_VALUE;
//        }
//        double perpendicularSlope = -1/slope;
//        if(Double.isInfinite(perpendicularSlope)){
//            perpendicularSlope = Double.MAX_VALUE;
//        }
//
//
////        double xPer = (point.getY() - line.getY1() -perpendicularSlope*point.getX() + slope*line.getX1())/(slope-perpendicularSlope);
//        double xPer = ((perpendicularSlope*point.getX() - point.getY()) - (slope*line.getX1() - line.getY1()))/(perpendicularSlope-slope);
//        double yPer = perpendicularSlope*(xPer-point.getX()) + point.getY();
//        double yPer2 = slope*(xPer-line.getX1()) + line.getY1();
//
//        return new Point2D.Double(xPer,yPer);
    }

    /**
     * MTN
     * calculate slope of a line
     *
     * @param line line that we want calculate slope of it
     * @return line slope
     */
    public static double slope(Line2D line) {
        double x1 = line.getX1();
        double y1 = line.getY1();
        double x2 = line.getX2();
        double y2 = line.getY2();
        return ((y1 - y2) / (x1 - x2));
    }

    public static double slope(rescuecore2.misc.geometry.Line2D line) {
        double x1 = line.getOrigin().getX();
        double y1 = line.getOrigin().getY();
        double x2 = line.getEndPoint().getX();
        double y2 = line.getEndPoint().getY();
        return ((y1 - y2) / (x1 - x2));
    }


    public static double lineLength(rescuecore2.misc.geometry.Line2D line) {
        double x1 = line.getOrigin().getX(), y1 = line.getOrigin().getY();
        double x2 = line.getEndPoint().getX(), y2 = line.getEndPoint().getY();
        return Math.sqrt(Math.pow(x1 - x2, 2) + Math.pow(y1 - y2, 2));
    }

    public static boolean containsEach(List list1, List list2) {
        for (Object object : list1) {
            if (list2.contains(object)) {
                return true;
            }
        }
        return false;
    }

    public static boolean contains(rescuecore2.misc.geometry.Line2D line2D, rescuecore2.misc.geometry.Point2D point) {
        return contains(line2D, point, 0);
    }

    public static boolean contains(rescuecore2.misc.geometry.Line2D line2D, rescuecore2.misc.geometry.Point2D point, double threshold) {
        return !(line2D == null || point == null) && Line2D.ptSegDist(line2D.getOrigin().getX(), line2D.getOrigin().getY(), line2D.getEndPoint().getX(), line2D.getEndPoint().getY(), point.getX(), point.getY()) <= threshold;
    }


    public static boolean contains(Line2D line2D, com.poths.rna.data.Point point) {
        return Line2D.ptSegDist(line2D.getX1(), line2D.getY1(), line2D.getX2(), line2D.getY2(), point.getX(), point.getY()) == 0;
    }

    public static double distance(rescuecore2.misc.geometry.Line2D line1, rescuecore2.misc.geometry.Line2D line2) {
        double dist1 = distance(line1, line2.getOrigin());
        double dist2 = distance(line1, line2.getEndPoint());
        double dist3 = distance(line2, line1.getOrigin());
        double dist4 = distance(line2, line1.getEndPoint());
        double min = Math.min(dist1, dist2);
        min = Math.min(min, dist3);
        min = Math.min(min, dist4);
        return min;
    }

    public static double distance(rescuecore2.misc.geometry.Line2D line, Polygon polygon) {
        List<rescuecore2.misc.geometry.Line2D> polyLines = getLines(polygon);
        double minDist = Double.MAX_VALUE;
        for (rescuecore2.misc.geometry.Line2D polyLine : polyLines) {
            minDist = Math.min(minDist, distance(line, polyLine));
            if (minDist == 0.0) {
                break;
            }
        }
        return minDist;
    }

    public static boolean isPassable(Polygon polygon, Polygon polygon1, int agentPassingThreshold) {

        int count = polygon1.npoints;
        int j;
        double tempDistance;
        boolean isPassable=false;
        for (int i = 0; i < count; i++) {
            j = (i + 1) % count;
            rescuecore2.misc.geometry.Point2D startPoint = new rescuecore2.misc.geometry.Point2D(polygon1.xpoints[i], polygon1.ypoints[i]);
            rescuecore2.misc.geometry.Point2D endPoint = new rescuecore2.misc.geometry.Point2D(polygon1.xpoints[j], polygon1.ypoints[j]);
            rescuecore2.misc.geometry.Line2D poly2Line = new rescuecore2.misc.geometry.Line2D(startPoint, endPoint);
            tempDistance = Util.distance(poly2Line, polygon);
            if(tempDistance<agentPassingThreshold){
                isPassable=true;
                break;
            }
        }
        return isPassable;

    }


    public static double distance(Polygon polygon1, Polygon polygon2) {
        int count = polygon2.npoints;
        double minDistance = Double.MAX_VALUE;
        int j;
        double distance;
        for (int i = 0; i < count; i++) {
            j = (i + 1) % count;
            rescuecore2.misc.geometry.Point2D startPoint = new rescuecore2.misc.geometry.Point2D(polygon2.xpoints[i], polygon2.ypoints[i]);
            rescuecore2.misc.geometry.Point2D endPoint = new rescuecore2.misc.geometry.Point2D(polygon2.xpoints[j], polygon2.ypoints[j]);
            rescuecore2.misc.geometry.Line2D poly2Line = new rescuecore2.misc.geometry.Line2D(startPoint, endPoint);
            distance = distance(poly2Line, polygon1);
            minDistance = Math.min(minDistance, distance);
            if (minDistance == 0.0) {
                break;
            }
        }
        return minDistance;
    }

    public static double distance(rescuecore2.misc.geometry.Line2D line, rescuecore2.misc.geometry.Point2D point) {
        return Line2D.ptSegDist(line.getOrigin().getX(), line.getOrigin().getY(), line.getEndPoint().getX(), line.getEndPoint().getY(), point.getX(), point.getY());
    }

    public static double distance(Polygon polygon, rescuecore2.misc.geometry.Point2D point) {
        int count = polygon.npoints;
        double minDistance = Double.MAX_VALUE;
        for (int i = 0; i < count; i++) {
            int j = (i + 1) % count;
            rescuecore2.misc.geometry.Point2D stPoint = new rescuecore2.misc.geometry.Point2D(polygon.xpoints[i], polygon.ypoints[i]);
            rescuecore2.misc.geometry.Point2D endPoint = new rescuecore2.misc.geometry.Point2D(polygon.xpoints[j], polygon.ypoints[j]);
            rescuecore2.misc.geometry.Line2D poly2Line = new rescuecore2.misc.geometry.Line2D(stPoint, endPoint);
            double distance = distance(poly2Line, point);
            minDistance = Math.min(minDistance, distance);
            if (minDistance == 0.0) {
                break;
            }
        }
        return minDistance;
    }

    public static List<rescuecore2.misc.geometry.Point2D> intersections(Polygon polygon, rescuecore2.misc.geometry.Line2D line) {
        List<rescuecore2.misc.geometry.Point2D> intersections = new ArrayList<rescuecore2.misc.geometry.Point2D>();
        List<rescuecore2.misc.geometry.Line2D> polyLines = getLines(polygon);

        for (rescuecore2.misc.geometry.Line2D ln : polyLines) {
            rescuecore2.misc.geometry.Point2D intersectPoint = GeometryTools2D.getIntersectionPoint(line, ln);
            if (contains(ln, intersectPoint)) {
                intersections.add(intersectPoint);
            }
        }
        return intersections;
    }

    public static List<rescuecore2.misc.geometry.Line2D> getLines(Polygon polygon) {
        List<rescuecore2.misc.geometry.Line2D> lines = new ArrayList<rescuecore2.misc.geometry.Line2D>();
        int count = polygon.npoints;
        for (int i = 0; i < count; i++) {
            int j = (i + 1) % count;
            rescuecore2.misc.geometry.Point2D p1 = new rescuecore2.misc.geometry.Point2D(polygon.xpoints[i], polygon.ypoints[i]);
            rescuecore2.misc.geometry.Point2D p2 = new rescuecore2.misc.geometry.Point2D(polygon.xpoints[j], polygon.ypoints[j]);
            rescuecore2.misc.geometry.Line2D line = new rescuecore2.misc.geometry.Line2D(p1, p2);
            lines.add(line);
        }
        return lines;
    }

    public static List<rescuecore2.misc.geometry.Line2D> getIntersectionLines(Polygon polygon, rescuecore2.misc.geometry.Line2D line) {
        List<rescuecore2.misc.geometry.Line2D> polyLines = Util.getLines(polygon);
        List<rescuecore2.misc.geometry.Line2D> intersectionLines = new ArrayList<rescuecore2.misc.geometry.Line2D>();
        for (rescuecore2.misc.geometry.Line2D polyLine : polyLines) {
            rescuecore2.misc.geometry.Point2D intersect = getIntersection(line, polyLine);
            if (contains(polyLine, intersect)) {
                intersectionLines.add(polyLine);
            }
        }
        return intersectionLines;
    }

    public static rescuecore2.misc.geometry.Line2D nearestLine(List<rescuecore2.misc.geometry.Line2D> lineList, rescuecore2.misc.geometry.Point2D point) {
        rescuecore2.misc.geometry.Line2D nearestLine = null;
        double minDistance = Double.MAX_VALUE;
        for (rescuecore2.misc.geometry.Line2D line : lineList) {
            double distance = lineSegmentAndPointDistance(line, point);
            if (distance < minDistance) {
                minDistance = distance;
                nearestLine = line;
            }
        }
        return nearestLine;
    }

    public static rescuecore2.misc.geometry.Point2D getPoint(Pair<Integer, Integer> position) {
        return new rescuecore2.misc.geometry.Point2D(position.first(), position.second());
    }

    public static rescuecore2.misc.geometry.Line2D improveLine(rescuecore2.misc.geometry.Line2D line, double size) {
        double x0 = line.getOrigin().getX(), x1 = line.getEndPoint().getX();
        double y0 = line.getOrigin().getY(), y1 = line.getEndPoint().getY();
        double deltaY = y1 - y0;
        double deltaX = x1 - x0;
        double xF, yF;
        double slope;
        if (deltaX != 0)
            slope = deltaY / deltaX;
        else {
            if (deltaY > 0)
                slope = Double.MAX_VALUE;
            else
                slope = -Double.MAX_VALUE;
        }

        double theta = Math.atan(slope);
        if (deltaX > 0) {
            xF = x1 + size * Math.abs(Math.cos(theta));
        } else {
            xF = x1 - size * Math.abs(Math.cos(theta));
        }
        if (deltaY > 0) {
            yF = y1 + size * Math.abs(Math.sin(theta));
        } else {
            yF = y1 - size * Math.abs(Math.sin(theta));
        }
        rescuecore2.misc.geometry.Point2D endPoint = new rescuecore2.misc.geometry.Point2D(xF, yF);
        return new rescuecore2.misc.geometry.Line2D(line.getOrigin(), endPoint);
    }

    private static double degToRad(double degree) {
        return degree * Math.PI / 180;
    }

    public static boolean equals(rescuecore2.misc.geometry.Line2D line1, rescuecore2.misc.geometry.Line2D line2) {
        return line1.getOrigin().getX() == line2.getOrigin().getX() && line1.getOrigin().getY() == line2.getOrigin().getY() && line1.getEndPoint().getX() == line2.getEndPoint().getX() && line1.getEndPoint().getY() == line2.getEndPoint().getY();
    }
}
