package mrl.common.clustering;

import javolution.util.FastMap;
import mrl.world.MrlWorld;
import mrl.world.object.MrlBuilding;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.StandardEntity;

import java.awt.*;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

/**
 * @author Vahid Hooshangi
 */
public class CivilianCluster extends Cluster {

    private MrlWorld world;
    private IClusterMembershipChecker membershipChecker;
    private int numberOfCivAndAgent;
    private int numberOfBuildings;
    private double civilianClusterCoef;
    private double civilianClusterDensity;
    private double civilianClusterValue;
    private final double roundingRadius = 1000;
    private List<Map<FireCluster, Double>> distanceMap;
    private Map<FireCluster, Double> finalValues;

    private double finalValue;

    public CivilianCluster(MrlWorld world, IClusterMembershipChecker membershipChecker) {
        this.world = world;
        this.membershipChecker = membershipChecker;
        numberOfCivAndAgent = 0;
        numberOfBuildings = 0;
        distanceMap = new ArrayList<Map<FireCluster, Double>>();
        finalValues = new FastMap<FireCluster, Double>();
//        convexHull = new ConvexHull();
    }

    private void setCentre() {
        int sumX = 0;
        int sumY = 0;
        for (int x : convexObject.getConvexPolygon().xpoints) {
            sumX += x;
        }

        for (int y : convexObject.getConvexPolygon().ypoints) {
            sumY += y;
        }

        if (convexObject.getConvexPolygon().npoints > 0) {
            center = new Point(sumX / convexObject.getConvexPolygon().npoints, sumY / convexObject.getConvexPolygon().npoints);
        }else {
            center = new Point(0,0);
        }

    }


    @Override
    public void updateConvexHull() {
        MrlBuilding mrlBuilding;
        convexObject.setConvexPolygon(convexHull.getConvexPolygon());

        if (convexObject != null && convexObject.getConvexPolygon().npoints != 0) {
            for (int i = 0; i < convexObject.getConvexPolygon().npoints; i++) {
                convexHull.addPoint(convexObject.getConvexPolygon().xpoints[i],
                        convexObject.getConvexPolygon().ypoints[i]);
            }
        }


        for (StandardEntity entity : entities) {
            if (entity instanceof Building) {
                mrlBuilding = world.getMrlBuilding(entity.getID());
                try {
                    if (membershipChecker.checkMembership(mrlBuilding)) {
                            numberOfCivAndAgent += removeBadBuilding(mrlBuilding);
                            numberOfBuildings += 1;
                            for (int i = 0; i < mrlBuilding.getSelfBuilding().getApexList().length; i += 2) {
                                convexHull.addPoint(mrlBuilding.getSelfBuilding().getApexList()[i],
                                        mrlBuilding.getSelfBuilding().getApexList()[i + 1]);
                            }
                        }
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
        }


        convexObject.setConvexPolygon(convexHull.getConvexPolygon());
        setCivilianClusterCoef();
        setCivilianClusterDensity();
        setCivilianClusterValue();
        setCentre();

    }


    private int removeBadBuilding(MrlBuilding building) {
        int count = 0;
        Human human;
        for (StandardEntity se : world.getHumans()) {
            human = (Human) se;
            if (human.isPositionDefined() && human.getPosition() == building.getID()) {
                if (human.isBuriednessDefined() && human.getBuriedness() > 0) {
                    if (human.isHPDefined() && human.getHP() > 0) {
                        count++;
                    }
                }
            }
        }
        return count;
    }

    private void setCivilianClusterCoef() {
        civilianClusterCoef = (numberOfCivAndAgent * (numberOfCivAndAgent - 1)) / 3;
    }

    private void setCivilianClusterDensity() {
        try {
            civilianClusterDensity = numberOfCivAndAgent / numberOfBuildings;
        } catch (Exception e) {
            civilianClusterDensity = 0;
        }
    }

    private void setCivilianClusterValue() {
        civilianClusterValue = civilianClusterCoef * civilianClusterDensity;
        civilianClusterValue = Math.round(civilianClusterValue * roundingRadius) / roundingRadius;
    }

    public void setDistanceMap(FireCluster fireCluster, double distance) {
        if (fireCluster == null)
            return;
        Map<FireCluster, Double> map = new FastMap<FireCluster, Double>();

        map.put(fireCluster, distance);

        for (Map<FireCluster, Double> m : distanceMap) {
            if (m.containsKey(fireCluster)) {
                distanceMap.remove(m);
                break;
            }
        }


        distanceMap.add(map);
    }

    public void setFinalValue(FireCluster fireCluster) {

        finalValue = civilianClusterValue + fireCluster.getCoef();
    }

    private void setFinalValue(FireCluster fireCluster, Double finalValue) {
        finalValues.remove(fireCluster);
        finalValues.put(fireCluster, finalValue);
    }

    public int getNumberOfCivAndAgent() {
        return numberOfCivAndAgent;
    }

    public int getNumberOfBuildings() {
        return numberOfBuildings;
    }

    public double getCivilianClusterCoef() {
        return civilianClusterCoef;
    }

    public double getCivilianClusterDensity() {
        return civilianClusterDensity;
    }

    public double getCivilianClusterValue() {
        return civilianClusterValue;
    }

    public double getFinalValue() {
        return finalValue;
    }
}
