package mrl.common.clustering;

import javolution.util.FastSet;
import mrl.viewer.layers.MrlCivilianClusterLayer;
import mrl.world.MrlWorld;
import mrl.world.object.MrlBuilding;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.StandardEntity;

import java.util.ArrayList;
import java.util.Set;

/**
 * @author vahid Hooshangi
 */
public class CivilianClusterManager extends ClusterManager {
    public static int CLUSTER_RANGE_THRESHOLD;


    public CivilianClusterManager(MrlWorld world) {
        super(world);
        CLUSTER_RANGE_THRESHOLD = (world.getPlatoonAgent().viewDistance / 2);
    }

    @Override
    public void updateClusters() {

        Cluster cluster;
        Cluster tempCluster;
        Set<Cluster> adjacentClusters = new FastSet<Cluster>();

        for (MrlBuilding building : world.getMrlBuildings()) {
            if (civilianConditionChecker.checkMembership(building)) {
                cluster = getCluster(building.getID());
                if (cluster == null) {
                    cluster = new CivilianCluster(world, civilianConditionChecker);
                    cluster.add(building.getSelfBuilding());
                    entityClusterMap.put(building.getID(), cluster);
                    for (StandardEntity entity : world.getObjectsInRange(building.getID(), CLUSTER_RANGE_THRESHOLD)) {
                        if (!(entity instanceof Building)) {
                            continue;
                        }

                        tempCluster = getCluster(entity.getID());
                        if (tempCluster == null) {
                            if (civilianConditionChecker.checkMembership(world.getMrlBuilding(entity.getID()))) {
                                cluster.add(entity);
                                entityClusterMap.put(entity.getID(), cluster);
                            }
                        } else {
                            adjacentClusters.add(tempCluster);
                        }
                    }

                    if (adjacentClusters.isEmpty()) {
                        addToClusterSet(cluster,building.getID());
                    } else {
                        merge(adjacentClusters, cluster, building.getID());
                    }
                }
            }
            adjacentClusters.clear();
        }

    }

    @Override
    public void updateConvexHullsForViewer() {
//        convexObjectsForViewer.clear();
//        for (Cluster cluster : clusters) {
//            convexObjectsForViewer.add(cluster.getConvexHullObject());
//        }
//        MrlCivilianClusterLayer.CONVEX_HULLS_MAP.put(world.getSelf().getID(), convexObjectsForViewer);
        java.util.List<Cluster> clusterList = new ArrayList<Cluster>();
        clusterList.addAll(clusters);
        MrlCivilianClusterLayer.CONVEX_HULLS_MAP.put(world.getSelf().getID(), clusterList);
    }
}
