package mrl.common.clustering;

import javolution.util.FastMap;
import mrl.viewer.layers.MrlConvexHullLayer;
import mrl.world.MrlWorld;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.worldmodel.EntityID;

import java.awt.*;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;


/**
 * @version 1.0
 */
public abstract class ClusterManager {

    protected MrlWorld world;
    protected List<Cluster> clusters;
    protected Map<EntityID, Cluster> entityClusterMap;
    /**
     * Checks MrlBuilding fire cluster membership based on real temperature information
     */
    protected IClusterMembershipChecker fireClusterMembershipChecker;

    /**
     * Checks (estimates) MrlBuilding fire cluster membership based on Estimated temperature information
     */
    protected IClusterMembershipChecker fireClusterMembershipEstimator;
    //protected IClusterMembershipChecker civilianConditionChecker;
    protected CivilianClusterMembershipChecker civilianConditionChecker;
    //protected CivilianClusterManager civilianClusterManager;
    protected List<ConvexObject> convexObjectsForViewer;
    private List<Polygon> bigBorderHulls;
    private List<Polygon> smallBorderHulls;
    protected List<StandardEntity> borders;

    public ClusterManager(MrlWorld world) {
        this.world = world;
        clusters = new ArrayList<Cluster>();
        entityClusterMap = new FastMap<EntityID, Cluster>();
        convexObjectsForViewer = new ArrayList<ConvexObject>();
        fireClusterMembershipChecker = new FireClusterMembershipChecker();
        fireClusterMembershipEstimator = new FireClusterMembershipEstimator();
        civilianConditionChecker = new CivilianClusterMembershipChecker(this.world);

        bigBorderHulls = new ArrayList<Polygon>();
        smallBorderHulls = new ArrayList<Polygon>();
        borders = new ArrayList<StandardEntity>();
    }

    public abstract void updateClusters();

    /**
     * merge new cluster to others and replace the result with all others
     *
     * @param adjacentClusters adjacent clusters to the new cluster
     * @param cluster          new constructed cluster
     * @param entityID
     */
    protected void merge(Set<Cluster> adjacentClusters, Cluster cluster,EntityID entityID) {

        for (Cluster c : adjacentClusters) {
            cluster.eat(c);

            // refreshing EntityClusterMap
            for (StandardEntity entity : c.entities) {
                entityClusterMap.remove(entity.getID()); //added 25 khordad! by sajjad & peyman
                entityClusterMap.put(entity.getID(), cluster);
            }
            clusters.remove(c);
        }

        addToClusterSet(cluster,entityID);
    }


    protected Cluster getCluster(EntityID id) {
        return entityClusterMap.get(id);
    }

    public List<Cluster> getClusters() {
        return clusters;
    }

    protected void addToClusterSet(Cluster cluster,EntityID entityID) {
//        cluster.updateConvexHull();
        entityClusterMap.put(entityID, cluster);
        clusters.add(cluster);
    }


    public void updateConvexHullsForViewer() {
        convexObjectsForViewer.clear();
        bigBorderHulls.clear();
        smallBorderHulls.clear();
        borders.clear();
        MrlConvexHullLayer.BIG_BORDER_HULLS.clear();
        MrlConvexHullLayer.SMALL_BORDER_HULLS.clear();
        for (Cluster cluster : clusters) {
            convexObjectsForViewer.add(cluster.getConvexHullObject());
            borders.addAll(cluster.getBorderEntities());
            MrlConvexHullLayer.BORDER_BUILDINGS.put(world.getSelf().getID(),borders);
            bigBorderHulls.add(cluster.getBigBorderPolygon());
            smallBorderHulls.add(cluster.getSmallBorderPolygon());
        }
        MrlConvexHullLayer.CONVEX_HULLS_MAP.put(world.getSelf().getID(), convexObjectsForViewer);
        MrlConvexHullLayer.BIG_BORDER_HULLS.addAll(bigBorderHulls);
        MrlConvexHullLayer.SMALL_BORDER_HULLS.addAll(smallBorderHulls);
    }




}
