package mrl.common.clustering;

import mrl.common.ConvexHull;
import mrl.common.Util;
import mrl.geometry.CompositeConvexHull;
import mrl.world.MrlWorld;
import mrl.world.object.MrlBuilding;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.StandardEntity;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;


/**
 * Created by IntelliJ IDEA.
 * User: Siavash
 * Date: 2/26/12
 * Time: 2:05 PM
 * Edited by Sajjad
 */
public class FireCluster extends Cluster {

    MrlWorld world;
    IClusterMembershipChecker membershipChecker;
    private Collection<Cluster> civilianClusters;
    private double coef;
    private double value;
    int idCounter;

    private List<MrlBuilding> buildings;

    public FireCluster(MrlWorld world, IClusterMembershipChecker membershipChecker, Collection<Cluster> civilianClusters) {
        super();
        this.world = world;
        this.membershipChecker = membershipChecker;
        this.civilianClusters = civilianClusters;
        this.buildings = new ArrayList<MrlBuilding>();
        idCounter = 0;
    }

    public FireCluster(MrlWorld world, IClusterMembershipChecker membershipChecker) {
        super();
        this.world = world;
        this.membershipChecker = membershipChecker;
        this.civilianClusters = civilianClusters;
        this.buildings = new ArrayList<MrlBuilding>();
        idCounter = 0;
    }

    /**
     * any building with temperature bigger than zero should be considered in convex hull
     */
    @Override
    public void updateConvexHull() {
        MrlBuilding mrlBuilding;
        convexHull = new CompositeConvexHull();
        convexObject.setConvexPolygon(convexHull.getConvexPolygon());
        if (convexObject != null && convexObject.getConvexPolygon() != null &&convexObject.getConvexPolygon().npoints != 0) {
            for (int i = 0; i < convexObject.getConvexPolygon().npoints; i++) {
                convexHull.addPoint(convexObject.getConvexPolygon().xpoints[i],
                        convexObject.getConvexPolygon().ypoints[i]);
            }
        }

        for (StandardEntity entity : entities) {
            if (entity instanceof Building) {
                mrlBuilding = world.getMrlBuilding(entity.getID());

                if (isDying && mrlBuilding.getEstimatedFieryness() > 0 && mrlBuilding.getEstimatedFieryness() < 3) {
                    setDying(false);
                }

                if (isEdge && !world.getBorderBuildings().contains(mrlBuilding.getID())) {
                    setEdge(false);
                }

                try {
//                    if (membershipChecker.checkMembership(mrlBuilding)) {
//                        convexHull.addPoint(mrlBuilding.getSelfBuilding().getX(),
//                                mrlBuilding.getSelfBuilding().getY());
                        for (int i = 0; i < mrlBuilding.getSelfBuilding().getApexList().length; i += 2) {
                            convexHull.addPoint(mrlBuilding.getSelfBuilding().getApexList()[i],
                                    mrlBuilding.getSelfBuilding().getApexList()[i + 1]);
                        }

//                    }
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
        }



        convexObject.setConvexPolygon(convexHull.getConvexPolygon());
        setBorderEntities();
//        sizeOfBuildings();
//        setTotalDistance();
    }

    private void setBorderEntities() {
        Building building;
        borderEntities.clear();

        if (convexObject.getConvexPolygon().npoints == 0) // I don't know why this happens, this should be checked TODO check this if something comes wrong here
            return;

        smallBorderPolygon = scalePolygon(convexObject.getConvexPolygon(), 0.9);
        bigBorderPolygon = scalePolygon(convexObject.getConvexPolygon(), 1.1);

        for (StandardEntity entity : entities) {

            if (entity instanceof Refuge) {
                continue;
            }
            if (!(entity instanceof Building)) {
                continue;
            }
            building = (Building) entity;
            int vertexes[] = building.getApexList();
            for (int i = 0; i < vertexes.length; i += 2) {

                if ((bigBorderPolygon.contains(vertexes[i],vertexes[i+1])) && !(smallBorderPolygon.contains(vertexes[i],vertexes[i+1])))
                {
                    borderEntities.add(building);
                    break;
                }
            }
        }
    }


    /**
     * this function calculate total distance between this fire cluster and All of the civilian Cluster
     */
    private void setTotalDistance() {
        double minDistance;
        double totalDistance = 0.0;

        for (Cluster cluster : civilianClusters) {
            minDistance = Integer.MAX_VALUE;
            int dis = 0;
            for (int i = 0; i < convexObject.getConvexPolygon().npoints; i++) {
                for (int j = 0; j < cluster.getConvexHullObject().getConvexPolygon().npoints; j++) {
                    dis = Util.distance(cluster.getConvexHullObject().getConvexPolygon().xpoints[j], cluster.getConvexHullObject().getConvexPolygon().ypoints[j], convexObject.getConvexPolygon().xpoints[i], convexObject.getConvexPolygon().ypoints[i]) / 1000;
                    if (minDistance > dis) {
                        minDistance = dis;
                    }
                }
            }
            if (cluster instanceof CivilianCluster) {
                minDistance = minDistance * ((CivilianCluster) cluster).getCivilianClusterValue();
            }
            totalDistance += minDistance;
        }

        setCoef(totalDistance);

    }

    /**
     * this function set the filre cluster Coef
     *
     * @param total the sum of minimum distance between this fire cluster and all civilian cluster
     */
    private void setCoef(double total) {
        int minDistance = Integer.MAX_VALUE;
        int dis = 0;
        for (int i = 0; i < convexObject.getConvexPolygon().npoints; i++) {
            dis = Util.distance(convexObject.getConvexPolygon().xpoints[i], convexObject.getConvexPolygon().ypoints[i], world.getMapWidth() / 2, world.getMapHeight() / 2) / 1000;
            if (dis < minDistance) {
                minDistance = dis;
            }
        }


        coef = total + minDistance + (buildings.size() / world.getBuildings().size()) + (getTotalBuildingArea(buildings) / getTotalBuildingArea(world.getMrlBuildings()));
    }

    private double getTotalBuildingArea(List<MrlBuilding> mrlBuildings) {
        double totalArea = 0;
        for (MrlBuilding building : mrlBuildings) {
            totalArea = building.getSelfBuilding().getShape().getBounds().getWidth() * building.getSelfBuilding().getShape().getBounds().getHeight();
        }
        return totalArea;
    }

    /**
     * this is function calculate sum of the fire building that have fireness 1, 2, 3
     */
    private void sizeOfBuildings() {
        int count = 0;
        for (StandardEntity entity : entities) {
            if (entity instanceof Building) {
                if (((Building) entity).isFierynessDefined() && ((Building) entity).getFieryness() >= 1 && ((Building) entity).getFieryness() <= 3) {
                    buildings.add(world.getMrlBuilding(entity.getID()));
                }
            }
        }

        value(count);
    }

    private void value(int count) {
        value = coef * count;
    }

    public double getCoef() {
        return coef;
    }

    public double getValue() {
        return value;
    }


    public List<MrlBuilding> getBuildings() {
        return buildings;
    }
}
