package mrl.common.clustering;

import javolution.util.FastSet;
import mrl.common.MRLConstants;
import mrl.common.TimestampThreadLogger;
import mrl.platoon.MrlPlatoonAgent;
import mrl.util.SortUtil;
import mrl.world.MrlWorld;
import mrl.world.object.MrlBuilding;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.worldmodel.EntityID;

import java.util.Collection;
import java.util.Set;

/**
 * @author Siavash
 */
public class FireClusterManager extends ClusterManager {

    public static int CLUSTER_RANGE_THRESHOLD;
    private Set<Cluster> civilianClusterSet;
    private Set<EntityID> lastBurningBuildings;
    private Set<EntityID> lastExtinguishedBuildings;
    private Set<EntityID> tempBurns;


    public FireClusterManager(MrlWorld world, Collection<Cluster> civilianClusters) {
        super(world);
        if (world.isMapHuge()) {
            CLUSTER_RANGE_THRESHOLD = world.getPlatoonAgent().viewDistance * 3;
        } else {
            CLUSTER_RANGE_THRESHOLD = world.getPlatoonAgent().viewDistance;
        }
//        civilianClusterSet = clusterSet;
        lastBurningBuildings = new FastSet<EntityID>();
        lastBurningBuildings = new FastSet<EntityID>();
    }

    public FireClusterManager(MrlWorld world) {
        super(world);
        if (world.getPlatoonAgent() != null) {
            if (world.isMapHuge()) {
                CLUSTER_RANGE_THRESHOLD = world.getPlatoonAgent().viewDistance * 3;
            } else {
                CLUSTER_RANGE_THRESHOLD = world.getPlatoonAgent().viewDistance;
            }
        }
        lastBurningBuildings = new FastSet<EntityID>();
        lastBurningBuildings = new FastSet<EntityID>();
        tempBurns = new FastSet<EntityID>();
    }

    @Override
    public void updateClusters() {
        Cluster cluster;
        Cluster tempCluster;
        Set<Cluster> adjacentClusters = new FastSet<Cluster>();

//        entityClusterMap.clear();
//        clusterSet.clear();

        for (MrlBuilding building : world.getMrlBuildings()) {
            if (MRLConstants.LAUNCH_VIEWER) {
                building.BUILDING_VALUE = Double.NaN;
            }
            //TODO: isEligible_Estimated ok or it should be isEligible
            //TODO @Siavash Check if the following condition is right
            if (fireClusterMembershipEstimator.checkMembership(building)) {
                cluster = getCluster(building.getID());
                if (cluster == null) {
                    cluster = new FireCluster(world, fireClusterMembershipChecker);
                    cluster.add(building.getSelfBuilding());

                    //checking neighbour clusters
                    for (StandardEntity entity : world.getObjectsInRange(building.getID(), CLUSTER_RANGE_THRESHOLD)) {
                        if (!(entity instanceof Building)) {
                            continue;
                        }

                        tempCluster = getCluster(entity.getID());
                        if (tempCluster == null) {
                            //TODO: isEligible_Estimated ok or it should be isEligible
                            /*if (fireClusterMembershipChecker.isEligible_Estimated(world.getMrlBuilding(entity.getID()))) {
                                cluster.add(entity);
                                entityClusterMap.put(entity.getID(), cluster);
                            }*/
                        } else {
                            adjacentClusters.add(tempCluster);
                        }
                    }

                    if (adjacentClusters.isEmpty()) {
                        addToClusterSet(cluster, building.getID());
                    } else {
                        merge(adjacentClusters, cluster, building.getID());
                    }


                } else {
                    //do noting
                }
            } else {
                // Was it previously in any cluster?
                cluster = getCluster(building.getID());
                if (cluster == null) {
                    //do nothing
                } else {
                    cluster.remove(building.getSelfBuilding());
                    entityClusterMap.remove(building.getID());//edited by sajjad, 2 lines shifted up
                    if (cluster.entities.isEmpty()) {
                        clusters.remove(cluster);
                    }
                }
            }
            adjacentClusters.clear();
        }


        // updating convexHull of each cluster
        for (Cluster c : clusters) {
            c.updateConvexHull();
        }


    }

}

