package mrl.firebrigade;

import mrl.common.*;
import mrl.common.clustering.CivilianCluster;
import mrl.common.clustering.Cluster;
import mrl.common.clustering.ConvexObject;
import mrl.communication.Packet;
import mrl.firebrigade.CLStrategy.FBCLStrategy;
import mrl.firebrigade.simulator.WaterCoolingEstimator;
import mrl.firebrigade.sterategy.SelectionHelper;
import mrl.helper.HumanHelper;
import mrl.helper.PropertyHelper;
import mrl.platoon.MrlPlatoonAgent;
import mrl.platoon.State;
import mrl.platoon.genericsearch.*;
import mrl.viewer.layers.MrlConvexHullLayer;
import mrl.world.object.Entrance;
import mrl.world.object.MrlBuilding;
import rescuecore2.messages.Command;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.ChangeSet;
import rescuecore2.worldmodel.EntityID;

import java.util.*;

/**
 * MRL fire brigade agent.
 */
public class MrlFireBrigade extends MrlPlatoonAgent<FireBrigade> {

    protected MrlFireBrigadeWorld world;
    private FireBrigadeMessageHelper fireBrigadeMessageHelper;

    protected MrlBuilding target;
//    protected MrlZone targetMrlZone;
//    private FireBrigadeDecisionManager decisionManager;
    private WaterCoolingEstimator coolingEstimator;

    private int maxDistance;
    private int maxPower;
    private int maxWater;
    private int waterRefillRate;
    private int timeoutExceptionCounter = 0;

    private MrlBuilding lastTarget;
    private int targetCycle;

    private MrlFireBrigadeDirectionManager fireBrigadeDirectionManager;
    private SelectionHelper selectionHelper;

    private ConvexObject convexObject;

    private FBCLStrategy FBCLStrategy;

    public BurningBuildingSearchManager fireSearcher;
    private BurningBuildingSearchDecisionMaker decisionMaker;

    @Override
    public String toString() {
        return "MRL fire brigade ID: " + this.getID().getValue();
    }

    @Override
    protected void postConnect() {
        super.postConnect();
        this.world = (MrlFireBrigadeWorld) super.world;

        model.indexClass(StandardEntityURN.BUILDING, StandardEntityURN.REFUGE);
        maxWater = config.getIntValue(MAX_WATER_KEY);
        maxDistance = config.getIntValue(MAX_EXTINGUISH_DISTANCE_KEY);
        maxPower = config.getIntValue(MAX_EXTINGUISH_POWER_KEY);
        waterRefillRate = WATER_REFILL_RATE;//config.getIntValue(WATER_REFILL_RATE_KEY);
        fireBrigadeMessageHelper = new FireBrigadeMessageHelper(world, this, messageManager);
//        directionManager = new DirectionManager(world, maxWater);
//        this.decisionManager = new FireBrigadeDecisionManager(world, directionManager);
        this.coolingEstimator = new WaterCoolingEstimator();
        this.fireBrigadeDirectionManager = new MrlFireBrigadeDirectionManager(world, (mrl.common.clustering.FireClusterManager) world.getFireClusterManager());

        if (MRLConstants.LAUNCH_VIEWER) {
            MrlConvexHullLayer.EXTINGUISH_RANGE = maxDistance;
        }

        FBCLStrategy = new FBCLStrategy(world, this, pathPlanner);

        selectionHelper = new SelectionHelper(world, fireBrigadeDirectionManager, maxDistance);
        //--------------SEARCH INITIATION-------------------

        decisionMaker = new BurningBuildingSearchDecisionMaker(world);
        possibleBuildingSearchDecisionMaker = new PossibleBuildingSearchDecisionMaker(world);
        stupidSearchDecisionMaker = new StupidSearchDecisionMaker(world);
        heatTracerDecisionMaker = new HeatTracerDecisionMaker(world);
        fireSearcher = new BurningBuildingSearchManager(world, world.getPlatoonAgent(), decisionMaker, defaultSearchStrategy);

//        myPower = maxPower;
        possibleBuildingSearchManager = new PossibleBuildingSearchManager(world, this, possibleBuildingSearchDecisionMaker, defaultSearchStrategy);
        stupidSearchManager = new StupidSearchManager(world, this, stupidSearchDecisionMaker, defaultSearchStrategy);
        heatTracerSearchManager = new HeatTracerSearchManager(world, this, heatTracerDecisionMaker, defaultSearchStrategy);
        defaultSearchManager = new DefaultSearchManager(world, this, stupidSearchDecisionMaker, defaultSearchStrategy);
        //--------------SEARCH INITIATION-------------------
        //--------------Border Initiation------------------
        world.setBorderBuildings();
        //--------------Border Initiation------------------

        System.out.println("   success");
    }

    @Override
    protected void think(int time, ChangeSet changed, Collection<Command> heard) {

        super.think(time, changed, heard);
    }

    @Override
    public void processMessage(Packet packet, EntityID sender) {
        fireBrigadeMessageHelper.processMessage(packet, sender);
    }

    @Override
    protected EnumSet<StandardEntityURN> getRequestedEntityURNsEnum() {
        return EnumSet.of(StandardEntityURN.FIRE_BRIGADE);
    }

    @Override
    public void act() throws CommandException, TimeOutException {

        if (world.getHelper(HumanHelper.class).isBuried(getID())) {
            setAgentState(State.BURIED);
            return;
        }
        if (world.isCommunicationLess()) {
            communicationLessAct();
        } else {
            commonAct();
        }
    }

    private void communicationLessAct() throws CommandException, TimeOutException {
        moveToRefugeIfDamagedOrTankIsEmpty();
        isThinkTimeOver("moveToRefugeIfDamagedOrTankIsEmpty");

        FBCLStrategy.execute();

    }

    private void commonAct() throws CommandException, TimeOutException {
        if (timeoutExceptionCounter < 15) {

            setCenterDirection();
            ourPreciousAct();

            setAgentState(State.Searching);
            heatTracerSearchManager.execute();
            fireSearcher.execute();
            defaultSearchManager.execute();
        }
    }

    public void setDirection() {
        double maxValue = Double.MIN_VALUE;
        Cluster bestCluster = null;
        for (Cluster fireCluster : world.getFireClusterManager().getClusters()) {
            for (Cluster civCluster : world.getCivilianClusterManager().getClusters()) {
                if (((CivilianCluster) civCluster).getCivilianClusterValue() > maxValue) {
                    maxValue = ((CivilianCluster) civCluster).getCivilianClusterValue();
                    bestCluster = civCluster;
                }
            }
            try {
                fireBrigadeDirectionManager.getBuildingInDirectionCivilianCluster(bestCluster.getCenter(), fireCluster.getConvexHullObject());
            } catch (NullPointerException e) {
                fireBrigadeDirectionManager.getBuildingInDirectionWorldCentre(fireCluster.getConvexHullObject());
            }
            this.convexObject = fireCluster.getConvexHullObject();
        }
    }

    public void setCenterDirection() {
        for (Cluster fireCluster : world.getFireClusterManager().getClusters()) {
            fireBrigadeDirectionManager.setCenterBasedDirection(fireCluster.getConvexHullObject());
        }
    }


    /*private void veryOldAct() throws CommandException, TimeOutException {
        moveToRefugeIfDamagedOrTankIsEmpty();
        isThinkTimeOver("moveToRefugeIfDamagedOrTankIsEmpty");

        extinguishNearbyWhenStuck();
        isThinkTimeOver("extinguishNearbyWhenStuck");

        directionSelection();
        isThinkTimeOver("directionSelection");


        if (directionManager.isDirectionEnabled()) {
            targetMrlZone = decisionManager.selectFireZones(directionManager.getTargetDirectZones());
            if (targetMrlZone == null) {
                targetMrlZone = decisionManager.selectFireZones(world.getZones().getBurningZones());
            }
        } else {
            targetMrlZone = decisionManager.selectFireZones(world.getZones().getBurningZones());
        }
        //if(fireClusterManager.getAssignedCluster() == null)
//        long s = System.nanoTime();
//        for (MrlFireCluster cluster : fireClusterManager.getClusters()) {
//            fireBrigadeDirectionManager.getBuildingInDirectionWorldCentre(cluster.getConvexHullObject());
//        }
//        long f = System.nanoTime();
//        System.out.println("Process time = " + (f - s) + " ns  " + ((f - s) / 1000000) + " milli second");

//        fireClusterManager.assignMeToCluster(decisionMaker.chooseNearestCluster(fireClusterManager.getClusters(), new Point(me().getX(), me().getY())));
//        if (fireClusterManager.getAssignedCluster() != null) {
//            fireBrigadeMessageHelper.sendGoToClusterMessage(fireClusterManager.getAssignedCluster().getCenter());
//        }
        isThinkTimeOver("selectFireCluster");

        target = decisionManager.selectBuilding();
//        target = decisionMaker.chooseTarget(fireClusterManager.getAssignedCluster());
        isThinkTimeOver("selectBuilding");

        extinguishTarget();
        isThinkTimeOver("extinguishTarget");

        explore();
        isThinkTimeOver("explore");

        afterWork();
        isThinkTimeOver("afterWork");
    }*/

    public void ourPreciousAct() throws CommandException, TimeOutException {
//        try{
        moveToRefugeIfDamagedOrTankIsEmpty();
        isThinkTimeOver("moveToRefugeIfDamagedOrTankIsEmpty");

        extinguishNearbyWhenStuck();
        isThinkTimeOver("extinguishNearbyWhenStuck");

//            extinguishMyself();
//            isThinkTimeOver("extinguish myself");
        target = selectionHelper.selectTarget();
        isThinkTimeOver("selectBuilding");

        if (MRLConstants.LAUNCH_VIEWER && target != null) {
            MrlConvexHullLayer.MY_TARGET.remove(me().getID());
            List<Building> trgt = new ArrayList<Building>();
            trgt.add(target.getSelfBuilding());
            MrlConvexHullLayer.MY_TARGET.put(me().getID(), trgt);
        }
//            moveToRefugeIfNeedWater();
//            isThinkTimeOver("just moving to refuge");

        extinguishTarget();
        isThinkTimeOver("extinguishTarget");
//        }catch(TimeOutException e){
//            timeoutExceptionCounter++;
//        }
    }


    private void moveToRefugeIfNeedWater() throws CommandException { //do not put this function before target selection
        if (target == null && me().getWater() < maxWater / 2) {
            restAtRefuge();
        } //if we are sure that there is no other work to do... we can go and refill water
    }

    private void extinguishMyself() throws CommandException {
        if ((world.getSelfPosition() instanceof Building) && ((Building) world.getSelfPosition()).isOnFire()) {
            target = world.getMrlBuilding(world.getSelfPosition().getID());
            if (target != null) {

                if (((FireBrigade) world.getSelfHuman()).getWater() != 0) {
                    int waterPower = calculateWaterPower(target);

                    fireBrigadeMessageHelper.sendWaterMessage(target.getID(), waterPower);
                    target.increaseWaterQuantity(waterPower);

                    sendExtinguishAct(world.getTime(), target.getID(), waterPower);
                }
            }
        }
    }

    private void moveToRefugeIfDamagedOrTankIsEmpty() throws CommandException {
        int requiredWater;
        if (world.getRefuges().isEmpty()) {
            return;
        }

        int fbCount = 0;
        if (me().getPosition(world) instanceof Refuge) {
            for (StandardEntity standardEntity : world.getFireBrigades()) {
                FireBrigade fb = (FireBrigade) standardEntity;
                if (fb.getPosition().equals(me().getPosition()) && world.isVisible(fb)) {
                    fbCount++;
                }
            }

            requiredWater = (maxWater + waterRefillRate) - (fbCount * waterRefillRate);

            requiredWater = Util.max(requiredWater, maxWater * 2 / 3);

//            if (target != null) {
//                if (world.getDistance(me(), target) <= 140000) {
//                    if (target.getTotalArea() <= 150) {
//                        requiredWater = maxWater / 2;
//                    } else {
//                        requiredWater = maxWater;
//                    }
//                }
//            }

            if (me().getWater() < requiredWater) {
                sendRestAct(world.getTime());
            }
        }

        if ((me().getDamage() > 10 && me().getHP() < 5000) || me().getWater() == 0) {


//            fireBrigadeMessageHelper.sendGoToClusterMessage(null);
            restAtRefuge();
        }

    }


    private void extinguishTarget() throws CommandException {
        if (target == null) {
            targetCycle = 0;
            return;
        }

        if (lastTarget == target) {
            targetCycle++;
        } else {
            targetCycle = 0;
            lastTarget = target;
        }
//        FireBrigade nearestFb = (FireBrigade) world.getSelfHuman();
        int myDistanceToTarget = world.getDistance(world.getSelfHuman(), target.getSelfBuilding());
        EntityID targetId = target.getSelfBuilding().getID();

        moveToEntranceForCheck();

        // move to target if I can't see this.
        if (myDistanceToTarget > maxDistance) {
            StandardEntity targetToGo = chooseBestWayToFire();

            if (targetToGo == null) {
                move(target.getSelfBuilding(), maxDistance, false);
            } else {
                if (MRLConstants.LAUNCH_VIEWER) {
//                    MrlConvexHullLayer.STAND.remove(world.getSelf().getID());
                    List<StandardEntity> stands = new ArrayList<StandardEntity>();
                    stands.add(targetToGo);
                    MrlConvexHullLayer.STAND.put(world.getSelf().getID(), stands);
                }
                if (!move((Area) targetToGo, maxDistance, false)) {
                    //age jayee ke mikhast bere rahi behesh nabood miad in khate code
                    Collection<StandardEntity> inRange = world.getObjectsInRange(targetToGo, maxDistance / 3);
                    int counter = 0;
                    for (StandardEntity entity : inRange) {
                        if (entity instanceof Area && world.getDistance(target.getSelfBuilding(), entity) < maxDistance) {
                            if (MRLConstants.LAUNCH_VIEWER) {
                                List<StandardEntity> stands = new ArrayList<StandardEntity>();
                                stands.add(entity);
                                MrlConvexHullLayer.STAND.put(world.getSelf().getID(), stands);
                            }
                            counter++;
                            move((Area) entity, maxDistance, false);
                            if (counter > 3) {
                                break;
                            }
                        }
                    }

                }
            }
        }

//        exitFromBuilding();
        if (myDistanceToTarget <= maxDistance) {
//            if (location() instanceof Refuge) {
//                Refuge ref = (Refuge) location();
//                Area area = getConnectedAreaToRefugeEntrance(ref);
//                if (area != null) {
//                    move(area, IN_TARGET, false);
//                }
//            }

//            if (target.isTemperatureDefined() && getCyclesRemainTarget() * maxPower > calculateWaterPower(target) * 1.5) {
////                System.out.println(getDebugString() + " needed Water used for : " + target);
//
//                int f = target.getFieryness();
//                if (f <= 3 && f >= 1)
//                    target.setFieryness(f + 4);// set to   put off
//            }

            if (((FireBrigade) world.getSelfHuman()).getWater() != 0) {
                int waterPower = calculateWaterPower(target);

                fireBrigadeMessageHelper.sendWaterMessage(targetId, waterPower);
                target.increaseWaterQuantity(waterPower);

                sendExtinguishAct(world.getTime(), targetId, waterPower);
            }

        }
//        if (target != null) {
//            fireBrigadeMessageHelper.sendTargetToGoMessage(target.getID());
//        }
//        System.out.println(getDebugString() + " No way to Extinguish:" + target);

    }

    private StandardEntity chooseBestWayToFire() {
        double minDistance = Integer.MAX_VALUE;
        StandardEntity targetToExtinguish = null;
        double dis;
        Collection<StandardEntity> inRange = world.getObjectsInRange(target.getID(), maxDistance);

        if (MRLConstants.LAUNCH_VIEWER) {
            Set<StandardEntity> inRangeSet = new HashSet<StandardEntity>(inRange);
            MrlConvexHullLayer.BEST_PLACE_TO_STAND.put(world.getSelf().getID(), inRangeSet);
        }
//        System.out.println("In Range: " + inRange.size());
        for (StandardEntity se : inRange) {
            if (se instanceof Building) {
                dis = Util.distance(me().getX(), me().getY(), ((Building) se).getX(), ((Building) se).getY());
                if (dis < minDistance && ((Building) se).isFierynessDefined() && (((Building) se).getFieryness() >= 4 || ((Building) se).getFieryness() <= 1)) {
                    minDistance = dis;
                    targetToExtinguish = se;
                }
            } else if (se instanceof Road) {
                dis = Util.distance(me().getX(), me().getY(), ((Road) se).getX(), ((Road) se).getY());
                if (dis < minDistance) {
                    minDistance = dis;
                    targetToExtinguish = se;
                }
            }
        }

        return targetToExtinguish;
    }

    public Entrance getAnOpenEntrance(MrlBuilding building) {
        Entrance nearestEntrance = null;
        int minDistance = Integer.MAX_VALUE;

        for (Entrance entrance : building.getEntrances()) {
            int dis = Util.distance(world.getSelfHuman().getX(), world.getSelfHuman().getY(), entrance.getNeighbour().getX(), entrance.getNeighbour().getY());
            if (entrance.isBlockedOrNotSeen(world.getRoadHelper()) && dis < minDistance) {
                minDistance = dis;
                nearestEntrance = entrance;
            }
        }
        return nearestEntrance;
    }

    private boolean aFireBrigadeIsOnEntrances(MrlBuilding building) {
        for (StandardEntity standardEntity : world.getFireBrigades()) {
            FireBrigade fb = (FireBrigade) standardEntity;
            if (fb.isPositionDefined() && building.getID().equals(fb.getPosition())) {
                return true;
            }
        }
        return false;
    }

    private boolean isMyLocationOnEntrances(MrlBuilding building) {
        for (Entrance entrance : building.getEntrances()) {
            if (location().equals(entrance.getNeighbour())) {
                return true;
            }
        }
        return false;
    }

    private void directionSelection() {
//        directionManager.chooseFireCluster();
//        if (directionManager.directionEnabled) {
//            world.printData("  Direction: " + directionManager.directionEnabled + " UP:" + directionManager.directValue[0] + " RIGHT:" + directionManager.directValue[1] + " DOWN:" + directionManager.directValue[2] + " LEFT:" + directionManager.directValue[3]);
//        }

    }

    protected int calculateWaterPower(MrlBuilding building) {
//        return Math.min(((FireBrigade) world.getSelfHuman()).getWater(), maxPower);
//        myPower = fireClusterManager.getMyPower();
        return Math.min(((FireBrigade) world.getSelfHuman()).getWater(), Math.min(maxPower, Math.max(500, waterNeededToExtinguish(building))));
    }

    public int waterNeededToExtinguish(MrlBuilding building) {
        return coolingEstimator.getWaterNeeded(building.getSelfBuilding().getGroundArea(), building.getSelfBuilding().getFloors(),
                building.getSelfBuilding().getBuildingCode(), building.getEstimatedTemperature(), 20);
    }

    private void extinguishNearbyWhenStuck() throws CommandException {

        if (world.getPlatoonAgent().isStuck()) {
            if (((FireBrigade) world.getSelfHuman()).getWater() == 0) {
                target = null;
                return;
            }
            Collection<StandardEntity> collection = world.getObjectsInRange(world.getSelfHuman(), maxDistance);
            List<MrlBuilding> firedBuildings = new ArrayList<MrlBuilding>();


            for (StandardEntity entity : collection) {
                if ((entity instanceof Building) && Util.distance(entity.getLocation(world), world.getSelfLocation()) < maxDistance) {
                    MrlBuilding building = world.getMrlBuilding(entity.getID());
                    if (/*building.isBurning()*/ building.getEstimatedTemperature() > 10) {
                        firedBuildings.add(building);
                    }
                }
            }


            target = selectionHelper.selectOneOfThese(firedBuildings);

            if (target != null) {

                if (((FireBrigade) world.getSelfHuman()).getWater() != 0) {
                    int waterPower = calculateWaterPower(target);

                    fireBrigadeMessageHelper.sendWaterMessage(target.getID(), waterPower);
                    target.increaseWaterQuantity(waterPower);

                    sendExtinguishAct(world.getTime(), target.getID(), waterPower);
                }

//                sendExtinguishAct(world.getTime(), target.getID(), Math.min(2000, maxPower));
            }
        }
    }

    // it was an old function, commented by Sajjad
    public MrlBuilding getBestBuilding(List<MrlBuilding> burningBuildings) {
        MrlBuilding best = null;
        double maxVal = Double.MIN_VALUE;
        for (MrlBuilding b : burningBuildings) {

            double val = (b.getBuildingAreaTempValue() * 1) + (b.getBuildingRadiation() * 2) - (b.getNeighbourRadiation() * 1);
//                   if (preTargetBuilding != null && b.equals(preTargetBuilding)) {
//                       val *= 1.2;
//                   }
            int civCount = 0;

            for (StandardEntity entity : world.getCivilians()) {
                Civilian civilian = (Civilian) entity;
                if (civilian.isPositionDefined() && b.getNeighbourIdBuildings().contains(civilian.getPosition())) {
                    civCount++;
                }
            }
            for (EntityID id : world.getBuriedAgents()) {
                Human human = (Human) world.getEntity(id);
                if (human.isPositionDefined() && b.getNeighbourIdBuildings().contains(human.getPosition())) {
                    civCount += 2;
                }
            }

            val += val * (((double) civCount * 5) / 100.0);
            if (val > maxVal) {
                maxVal = val;
                best = b;
            }
        }
        return best;
    }

    private void moveToEntranceForCheck() throws CommandException {
        PropertyHelper propertyHelper = world.getHelper(PropertyHelper.class);
        int lastUpdate = propertyHelper.getEntityLastUpdateTime(target.getSelfBuilding());

        if (        //Commented by Sajjad
            /*(targetCycle > (random.nextInt(3) + 2)
            || (world.isCommunicationLess()
            || world.isCommunicationLimited())
            ||*/ (world.getTime() - lastUpdate) > 2
                && !world.isVisible(target.getSelfBuilding())) {
            Entrance openEntrance = getAnOpenEntrance(target);
            if (openEntrance != null && !location().equals(openEntrance.getNeighbour())) {
                move(openEntrance.getNeighbour(), IN_TARGET, false);
            }
        }

    }

    public ConvexObject getConvexObject() {
        return convexObject;
    }

}