package mrl.helper;

import javolution.util.FastMap;
import mrl.world.MrlWorld;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.worldmodel.Property;

import java.util.Collection;
import java.util.Map;

/**
 * User: mrl
 * Date: Dec 3, 2010
 * Time: 10:54:57 AM
 */
public class PropertyHelper implements IHelper {

    protected MrlWorld world;
    protected Map<Property, Integer> propertyTimeMap = new FastMap<Property, Integer>();

    public PropertyHelper(MrlWorld world) {
        this.world = world;
    }

    public void init() {
        Collection<StandardEntity> entities = world.getEntitiesOfType(
                StandardEntityURN.AMBULANCE_TEAM,
                StandardEntityURN.AMBULANCE_CENTRE,
                StandardEntityURN.FIRE_BRIGADE,
                StandardEntityURN.FIRE_STATION,
                StandardEntityURN.POLICE_FORCE,
                StandardEntityURN.POLICE_OFFICE,
                StandardEntityURN.CIVILIAN,
                StandardEntityURN.BUILDING,
                StandardEntityURN.REFUGE,
                StandardEntityURN.BLOCKADE,
                StandardEntityURN.ROAD);

        propertyTimeMap.clear();

        for (StandardEntity entity : entities) {
            for (Property property : entity.getProperties()) {
                propertyTimeMap.put(property, 0);
            }
        }
    }

    public void update() {

    }

    public int getEntityLastUpdateTime(StandardEntity entity) {
        int maxTime = Integer.MIN_VALUE;
        for (Property property : entity.getProperties()) {
            Integer value = propertyTimeMap.get(property);
            if (value > maxTime) {
                maxTime = value;
            }
        }

        return maxTime;
    }

    public Integer getPropertyTime(Property property) {
        return propertyTimeMap.get(property);
    }

    public void setPropertyTime(Property property, Integer time) {
        propertyTimeMap.put(property, time);
    }

    public void addEntityProperty(StandardEntity entity, int time) {
        for (Property property : entity.getProperties()) {
            propertyTimeMap.put(property, time);
        }

    }
}
