package mrl.platoon;

import javolution.util.FastMap;
import mrl.ambulance.MrlAmbulanceTeam;
import mrl.ambulance.MrlAmbulanceTeamWorld;
import mrl.common.*;
import mrl.communication.MessageManager;
import mrl.communication.Packet;
import mrl.firebrigade.MrlFireBrigade;
import mrl.firebrigade.MrlFireBrigadeWorld;
import mrl.helper.HumanHelper;
import mrl.helper.RoadHelper;
import mrl.partition.Partition;
import mrl.platoon.genericsearch.*;
import mrl.police.MrlPoliceForce;
import mrl.police.MrlPoliceForceWorld;
import mrl.world.MrlWorld;
import mrl.world.object.MrlBuilding;
import mrl.world.routing.pathPlanner.AverageTools;
import mrl.world.routing.pathPlanner.IPathPlanner;
import mrl.world.routing.pathPlanner.PathPlanner;
import rescuecore2.Constants;
import rescuecore2.config.Config;
import rescuecore2.log.Logger;
import rescuecore2.messages.Command;
import rescuecore2.messages.control.KASense;
import rescuecore2.misc.Pair;
import rescuecore2.standard.components.StandardAgent;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.ChangeSet;
import rescuecore2.worldmodel.EntityID;

import java.util.*;

/**
 * User: mrl
 * Date: Apr 28, 2010
 * Time: 10:56:00 PM
 */
public abstract class MrlPlatoonAgent<T extends StandardEntity> extends StandardAgent<T> implements MRLConstants {
    // for viewer
    public static Map<EntityID, Integer> VALUE_FOR_DISPLAY_IN_VIEWER = new FastMap<EntityID, Integer>();
    public static EntityID CHECK_ID = new EntityID(0);
    public static EntityID CHECK_ID2 = new EntityID(0);
    //-------------------------------------------
    protected MrlWorld world;
    protected IPathPlanner pathPlanner;
    protected MessageManager messageManager;
    //    protected ISearchHelper searchHelper;
    protected long thinkStartTime_;
    protected long thinkTime;
    protected long thinkTimeThreshold;
    protected boolean useSpeak;
    public int viewDistance;
    public int clearDistance;
    //Sajjad
    public double refillRate = 0;

    protected int ignoreCommandTime;
    private String lastCommand;
    protected State agentState = State.Searching;
    private List<EntityID> visitedBuildings;
    private List<EntityID> visitedBuildingsHelp;

    protected boolean isStuck = false;
    private boolean movingRendezvous = false;
    private boolean onRendezvous = false;

    protected PossibleBuildingSearchManager possibleBuildingSearchManager;
    protected DefaultSearchManager defaultSearchManager;
    protected StupidSearchManager stupidSearchManager;
    protected CheckBlockadesManager checkBlockadesManager;
    protected ManualSearchManager manualSearchManager;
    public HeatTracerSearchManager heatTracerSearchManager;
    protected DefaultSearchStrategy defaultSearchStrategy;
    protected PossibleBuildingSearchDecisionMaker possibleBuildingSearchDecisionMaker;
    protected StupidSearchDecisionMaker stupidSearchDecisionMaker;
    protected HeatTracerDecisionMaker heatTracerDecisionMaker;
    protected CheckBlockadesDecisionMaker checkBlockadesDecisionMaker;
    protected ManualSearchDecisionMaker manualSearchDecisionMaker;
    protected AverageTools averageTools;


    public AmIStuckChecker amIStuckChecker;
    private int tryCount=5;

    public void sendSubscribe(int... channel) {
        sendSubscribe(world.getTime(), channel);
    }

    public void sendMessage(int channel, byte[] message) {
        sendSpeak(world.getTime(), channel, message);
    }

    public void sendMoveAct(int time, List<EntityID> path) throws CommandException {
        super.sendMove(time, path);
        throw new CommandException("Move");
    }

    public void sendMoveAct(int time, List<EntityID> path, int destinationX, int destinationY) throws CommandException {
        super.sendMove(time, path, destinationX, destinationY);
        throw new CommandException("Move To Point");
    }

    public void sendRandomWalkAct(int time, List<EntityID> path) throws CommandException {
        super.sendMove(time, path);
        throw new CommandException("Random Walk");
    }

    public void sendRestAct(int time) throws CommandException {
        super.sendRest(time);
        throw new CommandException("Rest");
    }

    protected void sendRescueAct(int time, EntityID target) throws CommandException {
        super.sendRescue(time, target);
        throw new CommandException("Rescue");
    }

    protected void sendLoadAct(int time, EntityID target) throws CommandException {
        super.sendLoad(time, target);
        throw new CommandException("Load");
    }

    protected void sendUnloadAct(int time) throws CommandException {
        super.sendUnload(time);
        throw new CommandException("Unload");
    }

    public void sendClearAct(int time, EntityID target) throws CommandException {
        super.sendClear(time, target);
        throw new CommandException("Clear");
    }

    protected void sendExtinguishAct(int time, EntityID target, int water) throws CommandException {
        super.sendExtinguish(time, target, water);
        throw new CommandException("Extinguish");
    }

    protected void updateSelfPosition(ChangeSet changeSet) {
        for (EntityID entity : changeSet.getChangedEntities()) {
            if (getID().equals(entity)) {
                world.getSelfHuman().setPosition(((Human) world.getEntity(entity)).getPosition());
                return;
            }
        }
    }

    protected void restAtRefuge() throws CommandException {
        setAgentState(State.Resting);
        if (world.getSelfHuman().getBuriedness() > 0) {
            throw new CommandException("I'm Buried");
        }
        if (world.getSelfPosition() instanceof Refuge) {
            sendRestAct(world.getTime());
        } else if (!world.getRefuges().isEmpty()) {
            if (world.getSelfHuman() instanceof PoliceForce) {
//                ((MrlPoliceForce) world.getSelf()).clearHere(PoliceActionStyle.CLEAR_NORMAL);
                ((MrlPoliceForce) world.getSelf()).getClearHelper().clearWay(pathPlanner.getNextPlan(), null);
            }
            moveToRefuge();
        }
    }

    public boolean move(Area target, int maxDistance, boolean force) throws CommandException {
        pathPlanner.move(target, maxDistance, force);
        return false;
    }

    protected boolean move(Collection<? extends Area> targets, int maxDistance, boolean force) throws CommandException {

//        pathPlanner.move(targets, maxDistance, force);
        throw new UnsupportedOperationException();
    }

    protected void moveToRefuge() throws CommandException {
        pathPlanner.moveToRefuge();
    }

    public void moveToPoint(EntityID area, int destX, int destY) throws CommandException {
        pathPlanner.moveToPoint(area, destX, destY);
    }

    public Config getConfig() {
        return config;
    }

    public IPathPlanner getPathPlanner() {
        return pathPlanner;
    }

    public MessageManager getMessageManager() {
        return messageManager;
    }

    public Partition getMyPartition() {
        return world.getPartitions().getMyPartition();
    }

    public List<EntityID> getVisitedBuildings() {
        return visitedBuildings;
    }

    public int getIgnoreCommandTime() {
        return ignoreCommandTime;
    }

    public Collection<StandardEntity> getObjectsInRange(StandardEntity entity, int range) {
        //TAVAJOH: baraye gereftane blockade ha hargez az in estefade nashe.
        return model.getObjectsInRange(entity, range);
    }

    public String getLastCommand() {
        return lastCommand;
    }

    public State getAgentState() {
        return agentState;
    }

    public void setAgentState(State state) {
        world.getAgentStateMap().put(world.getTime(), state);
        world.getHelper(HumanHelper.class).setAgentSate(me().getID(), state);
        agentState = state;
    }

    public String getDebugString() {
        return "Time:" + world.getTime() + " Me:" + me() + " ";
    }

    @Override
    protected EnumSet<StandardEntityURN> getRequestedEntityURNsEnum() {
        return EnumSet.of(StandardEntityURN.FIRE_BRIGADE,
                StandardEntityURN.AMBULANCE_TEAM,
                StandardEntityURN.POLICE_FORCE);
    }

    @Override
    protected void postConnect() {
        super.postConnect();
        System.out.print(this);

        this.thinkTime = config.getIntValue(THINK_TIME_KEY);
        this.thinkTimeThreshold = (long) (thinkTime * 0.9);
        this.useSpeak = config.getValue(Constants.COMMUNICATION_MODEL_KEY).equals(SPEAK_COMMUNICATION_MODEL_KEY);
//        this.viewDistance = config.getIntValue(MAX_VIEW_DISTANCE_KEY);
        this.viewDistance = config.getIntValue(MRLConstants.MAX_VIEW_DISTANCE_KEY);
        this.ignoreCommandTime = getConfig().getIntValue(MRLConstants.IGNORE_AGENT_COMMANDS_KEY);
        this.clearDistance = config.getIntValue(MAX_CLEAR_DISTANCE_KEY);
        Logger.debug("Communication model: " + config.getValue(Constants.COMMUNICATION_MODEL_KEY));
        Logger.debug(useSpeak ? "Using speak model" : "Using say model");

        if (this instanceof MrlPoliceForce) {
            world = new MrlPoliceForceWorld(this, model.getAllEntities());
        } else if (this instanceof MrlFireBrigade) {
            world = new MrlFireBrigadeWorld(this, model.getAllEntities());
        } else if (this instanceof MrlAmbulanceTeam) {
            world = new MrlAmbulanceTeamWorld(this, model.getAllEntities());
        } else {
            world = new MrlWorld(this, model.getAllEntities());
        }

        updateAgentStates(0);

//        world.setKernelTimeSteps(getConfig().getIntValue(KERNEL_TIMESTEPS));
        model = world;
        this.pathPlanner = new PathPlanner(world);

        if (HIGHWAY_STRATEGY) {
            pathPlanner.getGraph().setHighWayStrategy();
        }
//        world.getPathLength();


        world.partitionMakingOperations(world.getSelfHuman());


        world.preRoutingPartitions();

        this.messageManager = new MessageManager(world);

//        dataToFile = new LearnerIO("data/" + me().getID() + "data.txt", true);
//        dataToFile2 = new LearnerIO("data/civilianFound.txt", true);

        this.random = new Random(System.currentTimeMillis());
        visitedBuildings = new ArrayList<EntityID>();
        visitedBuildingsHelp = new ArrayList<EntityID>();
        amIStuckChecker = new AmIStuckChecker(world, 80, 6);

        defaultSearchStrategy = new DefaultSearchStrategy(this, world);
        checkBlockadesDecisionMaker = new CheckBlockadesDecisionMaker(world);
        checkBlockadesManager = new CheckBlockadesManager(world, this, checkBlockadesDecisionMaker, defaultSearchStrategy);
        averageTools = new AverageTools(world);
    }

    @Override
    protected void processSense(KASense sense) {
        long start = System.currentTimeMillis();
        Collection<Command> heard = sense.getHearing();

        think(sense.getTime(), sense.getChangeSet(), heard);

        long end = System.currentTimeMillis();
        if (end - start > thinkTime) {
//            Logger.warn("Time:" + sense.getTime() + " cycle needed:" + (end - start) + "ms");
            System.err.println("Time:" + sense.getTime() + " Agent:" + this + " cycle needed:" + (end - start) + "ms");
        }
    }

    @Override
    protected void think(int time, ChangeSet changed, Collection<Command> heard) {
        thinkStartTime_ = System.currentTimeMillis();

        try{
        if (MRLConstants.LAUNCH_VIEWER) {
            VALUE_FOR_DISPLAY_IN_VIEWER.put(getID(), world.getSelfHuman().getDamage());// for viewer
        }
        }catch (Exception e){
            e.printStackTrace();
        }

        world.setTime(time);

        updateAgentStates(time);


//        rendezvousOutPut();

        if (getID().equals(CHECK_ID)) {
            System.out.print("");
        }

        if (DEBUG_RENDEZVOUS_ACTION) {
            System.out.println(me() + " my partition is: " + world.getPartitions().getMyPartition().getId());
        }

        updateSelfPosition(changed);
        scanChannel(heard);
        messageManager.processHearing(heard);

        world.merge(changed);
        markVisitedBuildings();

        TimestampThreadLogger threadLogger = TimestampThreadLogger.getCurrentThreadLogger();
        world.updateEveryCycle();
        pathPlanner.update();

        world.updatePreviousChangeSet();

        int xPosition = world.getSelfLocation().first();
        int yPosition = world.getSelfLocation().second();

        if (time < getIgnoreCommandTime()) {
            return;
        } else if (time == 4 ) {
            isStuck = amIStuckChecker.amIStuck(xPosition, yPosition);
        } else if (time == 7) {
            isStuck = amIStuckChecker.amIStuck(xPosition, yPosition);
        } else if (time == 11) {
            isStuck = amIStuckChecker.amIStuck(xPosition, yPosition);
        }else if(isStuck && tryCount > random.nextInt(10)){
            tryCount = 0;
            isStuck = amIStuckChecker.amIStuck(xPosition, yPosition);
        }
        tryCount++;

        if (isStuck && !(world.getSelf() instanceof MrlPoliceForce)) {

            world.getHelper(HumanHelper.class).setLockedByBlockade(world.getSelf().getID(), true);

            try {
                checkBlockadesManager.execute();
            } catch (CommandException e) {
                lastCommand = e.getMessage();
                Logger.info(getDebugString() + " - ACT: " + e.getMessage());
            } catch (Exception e) {
                e.printStackTrace();
            }
        }

        if (!isStuck) {
            world.getHelper(HumanHelper.class).setLockedByBlockade(world.getSelf().getID(), false);
        }

        if (shouldRestAtRefuge()) {
            try {
                restAtRefuge();
            } catch (CommandException e) {
                lastCommand = e.getMessage();
                Logger.info(getDebugString() + " - ACT: " + e.getMessage());
                // age ye agent damage dasht va nemitoonest kari bokone hade aghal message bede.
                messageManager.sendMessages(onRendezvous);
                return;
            }
        }
//        ThreadPoolExecutor threadPoolExecutor=new ThreadPoolExecutor();
//        threadPoolExecutor.awaitTermination(10, TimeUnit.SECONDS);

//        long tt = System.currentTimeMillis();
        try {
            act();
//            setAgentState(State.Inactive);
//            world.getHelper(HumanHelper.class).setAmIStuck(true);
        } catch (CommandException e) {
            lastCommand = e.getMessage();
            Logger.info(getDebugString() + " - ACT: " + e.getMessage());
        } catch (TimeOutException e) {
            System.err.println(getDebugString() + "TimeOutException" + e.getMessage());
        } catch (Exception e) {
            e.printStackTrace();
        }

        if (world.getSelfHuman().getBuriedness() > 0) {
            setAgentState(State.BURIED);
        }

//        tt = System.currentTimeMillis() - tt;
//        System.out.println("process time = " + tt);

        messageManager.sendMessages(onRendezvous);

    }

    private void updateAgentStates(int time) {

        Human human;
        HumanHelper humanHelper = world.getHelper(HumanHelper.class);
        if (time == 0) {
            for (StandardEntity entity : world.getPlatoonAgents()) {
                human = (Human) entity;
                // if it is in building at the beginning of simulation, so it might be buried

                world.getAgentPositionMap().put(entity.getID(), human.getPosition());

                if (human.getPosition(world) instanceof Building) {
                    world.getBuriedAgents().add(entity.getID());
                } else if (!(entity instanceof PoliceForce)) {// if position is instance of Road , so it might be blocked
                    humanHelper.setLockedByBlockade(entity.getID(), true);
                }
            }
        } else {
            for (StandardEntity entity : world.getPlatoonAgents()) {
                human = (Human) entity;
                // if it is in building at the beginning of simulation, so it might be buried
                if (!human.getPosition().equals(world.getAgentPositionMap().get(human.getID()))) {
                    world.getBuriedAgents().remove(entity.getID());
                    humanHelper.setLockedByBlockade(entity.getID(), false);
                    world.getAgentPositionMap().put(entity.getID(), human.getPosition());
                }
            }
        }


    }

    public abstract void act() throws CommandException, TimeOutException;

    public abstract void processMessage(Packet packet, EntityID sender);

    public boolean isThinkTimeOver(String s) throws TimeOutException {
        if (!MRLConstants.LAUNCH_VIEWER && (System.currentTimeMillis() - thinkStartTime_) > thinkTimeThreshold) {
            throw new TimeOutException("  Timeout(" + thinkTimeThreshold + ") on: " + s);
        }
        return false;
    }

    private boolean shouldRestAtRefuge() {
        Human self = world.getSelfHuman();
        return (self.getBuriedness() == 0
                && ((self.getDamage() > 4 && self.getHP() < 7000)
                || self.getDamage() > 9
                || (self.getDamage() > 0 && self.getHP() < 5000)));
    }

    private void scanChannel(Collection<Command> commands) {
        if (world.getTime() < getIgnoreCommandTime()) {
            return;
        }
        messageManager.scanChannels(commands);
    }

    /**
     * in method jahaei ro ke dide ro be ye list ezafe mikone ke bahash message dorost mikonim.
     */
    private void markVisitedBuildings() {
        Area position = (Area) world.getSelfPosition();
        markVisitedBuildings(position);
    }
    
    public void markVisitedBuildings(Area area){


        if (area instanceof Road) {
            List<Building> buildings = RoadHelper.getConnectedBuildings(world, (Road) area);
            for (Building building : buildings) {
                if (!world.getVisitedBuildings().contains(building.getID()) && !world.getShouldCheckInsideBuildings().contains(world.getMrlBuilding(building.getID()))) {
                    removeVisitedBuildingFromWorld(building);
                }
            }
        }

        if (area instanceof Building) {
            if (!world.getVisitedBuildings().contains(area.getID()))
                removeVisitedBuildingFromWorld((Building) area);
        }
    }

    public void removeVisitedBuildingFromWorld(Building building) {
//        if (visitedBuildings.size() > 3) {
//            visitedBuildings.remove(0);
//        }
        if (!visitedBuildings.contains(building.getID())) {
            visitedBuildings.add(building.getID());
        }
        if (!visitedBuildingsHelp.contains(building.getID())) {
            visitedBuildingsHelp.add(building.getID());
        }
        world.getUnvisitedBuildings().remove(building.getID());
        world.getVisitedBuildings().add(building.getID());

        MrlBuilding mrlBuilding = world.getMrlBuilding(building.getID());
        mrlBuilding.setVisited();
    }

    public List<EntityID> getVisitedBuildingsHelp() {
        return visitedBuildingsHelp;
    }

    public boolean isMovingRendezvous() {
        return movingRendezvous;
    }

    public void setMovingRendezvous(boolean movingRendezvous) {
        this.movingRendezvous = movingRendezvous;
    }

    public void setIAmOnRendezvousPlace(boolean onRendezvous) {
        this.onRendezvous = onRendezvous;
    }

    public Random getRandom() {
        return random;
    }

    public boolean isStuck() {
        return isStuck;
    }

    public void setStuck(boolean stuck) {
        isStuck = stuck;
    }

    public AverageTools getAverageTools() {
        return averageTools;
    }

    StandardEntity loopTarget;

    public void loopBetweenTwoArea(StandardEntity area1, StandardEntity area2) throws CommandException {
        if (world.getSelfPosition().getID().equals(area2.getID()) || loopTarget == null) {
            loopTarget = area1;
        } else if (world.getSelfPosition().getID().equals(area1.getID())) {
            loopTarget = area2;
        }
        move((Area) loopTarget, 0, true);
    }

    Pair<Integer, Integer> loopTargetPoint;

    public void loopBetweenTwoPoint(Pair<Integer, Integer> p1, Pair<Integer, Integer> p2) throws CommandException {
        if (world.getSelfLocation().equals(p2) || loopTargetPoint == null) {
            loopTargetPoint = p1;
        } else if (world.getSelfLocation().equals(p1)) {
            loopTargetPoint = p2;
        }
        moveToPoint(world.getSelfPosition().getID(), loopTargetPoint.first(), loopTargetPoint.second());
    }
}