package mrl.platoon.genericsearch;

import mrl.common.CommandException;
import mrl.common.MRLConstants;
import mrl.helper.BuildingHelper;
import mrl.platoon.MrlPlatoonAgent;
import mrl.world.MrlWorld;
import mrl.world.object.Entrance;
import mrl.world.object.MrlBuilding;
import mrl.world.routing.path.Path;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.worldmodel.EntityID;

import java.util.ArrayList;
import java.util.List;

import static mrl.platoon.genericsearch.SearchStatus.*;

/**
 * @author Siavash
 * @see SearchStrategy
 */
public class DefaultSearchStrategy extends SearchStrategy {

//    private static Log //logger = LogFactory.getLog(DefaultSearchStrategy.class);
    public static List<MrlBuilding> SHOULD_CHECK_INSIDE_BUILDINGS = new ArrayList<MrlBuilding>();
    ArrayList<Area> entrances;
    int tryCount;
    Area target;
    List<Building> shouldVisitInsideBuildings;

    public DefaultSearchStrategy(MrlPlatoonAgent agent, MrlWorld world) {
        super(agent, world);
        entrances = new ArrayList<Area>();
        shouldVisitInsideBuildings = new ArrayList<Building>();
        target = null;
        tryCount = 3;
        if (MRLConstants.LAUNCH_VIEWER) {
            SHOULD_CHECK_INSIDE_BUILDINGS = new ArrayList<MrlBuilding>(world.getShouldCheckInsideBuildings());
        }
    }

    public void updateEntrances(Path path, boolean searchUnvisited) {
        //logger.debug("Updating Building And Entrances...");
        if (path != null && path.getEntrances() != null) {
            for (Entrance entrance : path.getEntrances()) {
                Area area = (Area) world.getEntity(entrance.getID());
                entrances.add(area);
            }
            if (searchUnvisited) {
                entrances.removeAll(visitedAreas);
            }
        }
    }


    /**
     * @see DefaultSearchStrategy#search(boolean, boolean)
     */
    @Override
    public SearchStatus search(boolean inPartition, boolean searchUnvisited) throws CommandException {
        throw new UnsupportedOperationException();
    }

    /**
     * @see DefaultSearchStrategy#searchPath()
     */
    @Override
    public SearchStatus searchPath() throws CommandException {
        StandardEntity myPosition = world.getSelfPosition();

        if (myPosition.equals(target) || target == null) {
            //logger.debug("I've reached target: [" + target + "]");
            if (target != null) {
                addVisitedArea(target);
                for (EntityID entityID : target.getNeighbours()) {
                    Area area = (Area) world.getEntity(entityID);
                    if (area instanceof Building) {
                        updateVisitedBuildings((Building) area);
                    }
                }

            }
            if (!entrances.isEmpty()) {
                target = entrances.get(0);
                entrances.remove(0);
            }
        }

        if (target != null) {
            if (tryCount > 0) {
//                if(agent instanceof MrlPoliceForce){
//                    ((MrlPoliceForce) agent).clearHere();
//                }
                agent.move(target, MRLConstants.IN_TARGET, false);
                tryCount--;
            } else {
                tryCount = 3;
                target = null;
            }
        }

        if (entrances.isEmpty()) {
            target = null;
            return FINISHED;
        }
        return CANCELED;
    }


//    private boolean isInFire(Building building) {
//        return (building.isFierynessDefined() && !(building.getFieryness() == 0 || building.getFieryness() == 4 || building.getFieryness() == 5));
//    }

    private boolean shouldICheckInside(Building building) {
        MrlBuilding mrlBuilding = world.getMrlBuilding(building.getID());
        boolean searchInside = mrlBuilding.shouldCheckInside();
        boolean searchInsideTemp = searchInside;
        for (Area entArea : BuildingHelper.getEntranceAreas(world, building)) {
            if (entArea instanceof Road) {
                searchInsideTemp = searchInside;
                break;
            }
            if (entArea instanceof Building) {
                if (BuildingHelper.hasPossibleAliveHuman((Building) entArea)) {
                    searchInsideTemp = false;
                }
            }
        }
        searchInside = searchInsideTemp;
        if (BuildingHelper.hasPossibleAliveHuman(building)) {
            searchInside = false;
        }
        return searchInside;
    }

    /**
     * @see DefaultSearchStrategy#searchBuilding(rescuecore2.standard.entities.Building)
     */
    @Override
    public SearchStatus searchBuilding(Building building) throws CommandException {

        if (building == null) {
            return CANCELED;
        }

        boolean searchInside = shouldICheckInside(building);

        if (searchInside) {
            if (world.getSelfPosition().equals(building)) {
                updateVisitedBuildings();
                blackList.clear();
                return FINISHED;
            }
        } else {
            List<Road> entranceRoads = BuildingHelper.getEntranceRoads(world, building);
            for (Road road : entranceRoads) {
                if (world.getSelfPosition().equals(road)) {
                    updateVisitedBuildings(building);
                    blackList.clear();
                    return FINISHED;
                }
            }
        }
        manualMoveToSearchingBuilding(building, searchInside, false);
        return CANCELED;
    }

    /**
     * @see DefaultSearchStrategy#search()
     */
    @Override
    public SearchStatus search() {
        throw new UnsupportedOperationException();
    }

    @Override
    public void setSearchingPath(Path searchingPath, boolean searchUnvisited) {
        this.path = searchingPath;
        updateEntrances(searchingPath, searchUnvisited);
    }

    @Override
    public void setSearchingBuilding(Building searchingBuilding) {
        this.building = searchingBuilding;
    }
}