package mrl.platoon.genericsearch;

import javolution.util.FastSet;
import mrl.partition.Partition;
import mrl.world.MrlWorld;
import mrl.world.object.MrlBuilding;
import mrl.world.routing.path.Path;
import rescuecore2.worldmodel.EntityID;

import java.util.Set;

/**
 * Default implementation for {@link ISearchDecisionMaker}
 *@author Siavash
 * @see SearchManager
 */
public abstract class SearchDecisionMaker implements ISearchDecisionMaker {

    protected MrlWorld world;
    protected boolean searchInPartition;

    protected Set<Path> validPaths;
    protected Set<EntityID> validBuildings;
    protected Partition partition;
    //TODO: handle unvisited buildings check

    public SearchDecisionMaker(MrlWorld world) {
        this.world = world;
        validBuildings = new FastSet<EntityID>();
        validPaths = new FastSet<Path>();

    }

    @Override
    public void update() {
        if(searchInPartition){
            validBuildings.clear();
            for(MrlBuilding mrlBuilding : world.getPartitionManager().findHumanPartition(world.getSelfHuman()).getBuildings()){
                validBuildings.add(mrlBuilding.getID());
            }

            validPaths.clear();
            for(Path path : world.getPartitionManager().findHumanPartition(world.getSelfHuman()).getPaths()){
                validPaths.add(path);
            }

        }
    }

    public void setSearchInPartition(boolean searchInPartition) {
        this.searchInPartition = searchInPartition;
    }

    public void setPartition(Partition partition) {
        this.partition = partition;
    }

}
