package mrl.police;

import mrl.common.Util;
import mrl.communication.messages.ClearedRoadMessage;
import mrl.helper.HumanHelper;
import mrl.helper.RoadHelper;
import mrl.partitioning.Partition;
import mrl.platoon.State;
import mrl.police.moa.PoliceForceUtilities;
import mrl.police.moa.Target;
import mrl.task.Task;
import mrl.world.MrlWorld;
import mrl.world.routing.pathPlanner.CheckAreaPassable;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;

import java.util.List;
import java.util.Map;

/**
 * Author: Pooya Deldar Gohardani
 * Date: 2/7/12
 * Time: 3:02 PM
 */
public class PoliceConditionChecker {

    private MrlWorld world;
    private PoliceForceUtilities utilities;
    private boolean isFirstTime;

    private int numberOfHealthyPoliceForces;
    private int numberOfHealthyCoworkers;
    private RoadHelper roadHelper;
    private ClearHereHelper clearHereHelper;
    private int minTimeToSplitPartitions = 10;

    public PoliceConditionChecker(MrlWorld world, PoliceForceUtilities utilities) {
        this.world = world;
        this.utilities = utilities;
        this.isFirstTime = true;
        roadHelper = world.getHelper(RoadHelper.class);
        numberOfHealthyPoliceForces = 0;
    }


    /**
     * Is target reachable from a way?
     *
     * @param task the task that should be examined
     * @return true if target is reachable, otherwise false
     */
    public boolean isTargetCleared(Task task) {

        if (task == null) {
            return false;
        }
        //TODO: should be done better...
//        return Util.distance(world.getSelfLocation(), world.getEntity(task.getTarget().getId()).getLocation(world))
//                <= world.getPlatoonAgent().viewDistance / 5;


//        Blockade blockade=clearHereHelper.getBlockadeToClear(task.getTarget(), task.getActionStyle());
        CheckAreaPassable checkAreaPassable = world.getPlatoonAgent().getPathPlanner().getAreaPassably();
        ClearedRoadMessage clearedRoadMessage;
        List<EntityID> blockades;
        Road road;
        boolean isTargetDone = true;
        for (EntityID targetID : task.getTarget().getRoadsToMove().keySet()) {
            road = (Road) world.getEntity(targetID);
            blockades = checkAreaPassable.policeCheckPassably(road, null);
            if (roadHelper.isPassable(targetID)) {
                if (road.isBlockadesDefined() && road.getBlockades().isEmpty()) {
                    continue;
                } else {
                    isTargetDone = false;
                }
            } else if (!roadHelper.isPassable(targetID) || blockades != null && !blockades.isEmpty()) {
                isTargetDone = false;
            } else {
                // this road is passable
                roadHelper.setRoadPassable(targetID, true);
                task.getTarget().getRoadsToMove().put(targetID, true);
                clearedRoadMessage = new ClearedRoadMessage(world.getIndexes().getAreaIndex(targetID));
                world.getPlatoonAgent().getMessageManager().addMessageToQueue(clearedRoadMessage, clearedRoadMessage.getType());

            }
        }


        return isTargetDone;

//        return amIOnTarget(task.getTarget()) || roadHelper.isPassable(task.getTarget().getNearestRoadID());

    }

    private boolean amIOnTarget(Target target) {
        return world.getSelfPosition().getID().equals(target.getNearestRoadID());
    }

    public boolean isTimeToPartitionAssignment() {

        if (world.getTime() >= 12 && (isFirstTime || isNumberOfHealthyPolicesChanged())) {
            if (isFirstTime) {
                isFirstTime = false;
                numberOfHealthyPoliceForces = utilities.getHealthyPoliceForces().size();
                if (numberOfHealthyPoliceForces == 0) {
                    return false;
                }
            }
            return true;
        } else {
            return false;
        }

    }

    public boolean isTimeToPartitionSplitting(Partition myOriginalPartition, Map<Human, Partition> humanPartitionMap) {

        if (world.getTime() < minTimeToSplitPartitions) {
            return false;
        }

        int i = 0;
        for (StandardEntity entity : utilities.getHealthyCoworkers()) {
            if (!entity.getID().equals(world.getSelf().getID()) && humanPartitionMap.get(entity).equals(myOriginalPartition)) {
                i++;
            }

        }
        return utilities.getHealthyCoworkers().contains(world.getSelfHuman()) && i > 0;
    }

    private boolean isNumberOfHealthyPolicesChanged() {

        if (utilities.getHealthyPoliceForces().size() == 0) {
            return false;
        }

        if (numberOfHealthyPoliceForces != utilities.getHealthyPoliceForces().size()) {
            numberOfHealthyPoliceForces = utilities.getHealthyPoliceForces().size();
            return true;
        } else {
            return false;
        }

    }


    public boolean isInBadBuilding() {
        if (world.getSelfPosition() instanceof Building) {
            Building building = (Building) world.getSelfPosition();
            if (building.isTemperatureDefined() && building.getTemperature() > 0) {
                return true;
            }
        }
        return false;
    }
}
