package mrl.police.strategies;

import mrl.platoon.MrlPlatoonAgent;
import mrl.police.PoliceConditionChecker;
import mrl.police.PoliceMessageHelper;
import mrl.police.moa.Target;
import mrl.police.moa.PoliceForceUtilities;
import mrl.task.MoveStyle;
import mrl.task.Task;
import mrl.world.MrlWorld;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

/**
 * User: Pooya
 * Date: 3/1/12
 * Time: 7:45 PM
 */
public class PolicePrioritizationStrategy extends PoliceStrategy {


    public PolicePrioritizationStrategy(MrlWorld world, MrlPlatoonAgent self, boolean inPartition, PoliceForceUtilities utilities,
                                        PoliceMessageHelper messageHelper, PoliceConditionChecker conditionChecker) {
        super(inPartition, messageHelper, world, utilities, conditionChecker, self);
    }


    @Override
    public Task act() {
//        if (myTask != null && conditionChecker.isTargetCleared(myTask)) {
//            System.err.println(world.getTime() + " " + self.getID() + " I have done my assigned task");
//            myTask = null;
//        }

        List<Target> targetList = new ArrayList<Target>(utilities.getTargets(inPartition).values());
        Collections.sort(targetList, utilities.TARGET_IMPORTANCE_COMPARATOR);

//        for (Target target:targetList){
//            System.out.println(world.getTime()+" "+world.getSelf().getID()+" target:"+target.getId()+" imp:"+target.getImportance()+" dist:"+target.getDistanceToIt());
//        }

        if (myTask != null) {
            if (conditionChecker.isTargetCleared(myTask)) {
                doneTaskOperations(targetList);
            }
        }

        Target mostImportantTarget = null;
        int i = 0;
        while (i < targetList.size()) {
            if (!targetList.isEmpty()) {
                mostImportantTarget = targetList.get(i);
            }

            if (myTask == null || isThereMoreImportantTask(myTask.getTarget(), mostImportantTarget)) {
                if (mostImportantTarget != null) {
                    mostImportantTarget.setRoadsToMove(utilities.getRoadsToMove(mostImportantTarget));
                    myTask = new Task(mostImportantTarget, utilities.getActionStyle(mostImportantTarget), MoveStyle.FASTEST_PATH);

                    if (conditionChecker.isTargetCleared(myTask)) {
                        doneTaskOperations(targetList);
                        i++;
                    } else {
                        break;//found new task, so break te loop immediately
                    }
                }
            } else if (conditionChecker.isTargetCleared(myTask)/*utilities.isDoneTask(myTask)*/) {

                doneTaskOperations(targetList);
                i++;

            } else {
                break;
            }
        }
        return myTask;

    }

    private void doneTaskOperations(List<Target> targetList) {
        utilities.getDoneTargets().add(myTask.getTarget().getId());
        targetList.remove(myTask.getTarget());
        myTask = null;
    }


    private boolean isThereMoreImportantTask(Target myTarget, Target mostImportantTarget) {

        return mostImportantTarget != null && myTarget.getImportance() < mostImportantTarget.getImportance();

    }


}
