package mrl.viewer;

import javolution.util.FastMap;
import javolution.util.FastSet;
import mrl.common.MRLConstants;
import mrl.firebrigade.FireClusterManager;
import mrl.platoon.MrlPlatoonAgent;
import mrl.world.MrlWorld;
import mrl.world.object.MrlBuilding;
import org.jfree.ui.tabbedui.VerticalLayout;
import rescuecore2.Constants;
import rescuecore2.Timestep;
import rescuecore2.log.Logger;
import rescuecore2.messages.control.KVTimestep;
import rescuecore2.score.ScoreFunction;
import rescuecore2.standard.components.StandardViewer;
import rescuecore2.standard.entities.*;
import rescuecore2.view.RenderedObject;
import rescuecore2.view.ViewComponent;
import rescuecore2.view.ViewLayer;
import rescuecore2.view.ViewListener;
import rescuecore2.worldmodel.EntityID;
import rescuecore2.worldmodel.Property;

import javax.swing.*;
import javax.swing.border.EtchedBorder;
import javax.swing.tree.DefaultMutableTreeNode;
import java.awt.*;
import java.awt.event.*;
import java.lang.reflect.Method;
import java.text.NumberFormat;
import java.util.*;
import java.util.List;

import static rescuecore2.misc.java.JavaTools.instantiate;

/**
 * A simple viewer.
 */
public class MrlViewer extends StandardViewer implements MRLConstants {

    public static Map<EntityID, List<Integer>> civilianPassableBuildingMap = new FastMap<EntityID, List<Integer>>();
    public static Map<EntityID, MrlBuilding> TEMP_BUILDING_MAP = new FastMap<EntityID, MrlBuilding>();
    private static final int FONT_SIZE = 20;
    private static final int MAX_PROPERTIES = 11;
    private MrlAnimatedWorldModelViewer viewer;
    private JLabel timeLabel;
    private JLabel scoreLabel;
    private JPanel panel = new JPanel(new BorderLayout());
    private JPanel topPanel = new JPanel(new BorderLayout());
    private JPanel bottomPanel = new JPanel(new BorderLayout());
    private final JTable propertiesTable = new JTable(new Object[MAX_PROPERTIES][2], new String[]{"Property", "Value"});
    DefaultMutableTreeNode rootNode = new DefaultMutableTreeNode("Agents Data");
    JTree tree = new JTree(rootNode);
    JPopupMenu popup = new JPopupMenu("popupMenu");
    StandardEntity selectedObject;
    JCheckBox jCheckBoxUpdateAgentData = new JCheckBox("update Agent Data");
    public static long randomValue;
    boolean shouldUpdateAgentData = false;
    boolean pause = false;
    int time;
    private NumberFormat format;
    private ScoreFunction scoreFunction;
    Map<Integer, KVTimestep> dataMap = new FastMap<Integer, KVTimestep>();
    boolean isSetToManual = false;
    int currentTime;
    private String scoreValue;

    public static Map<EntityID,StandardEntity> BUILDINGS;//TODO :just for test

//    public static FileOutputStream writer;
//    static{
//        try {
//            writer=new FileOutputStream("t.xml");
//        } catch (IOException e) {
//            e.printStackTrace();
//        }
//    }

    public void selectObject(int id) {

//        String s="<rcr:civilian rcr:location=\""+id +"\"/>\n";
//        try {
//            writer.write(s.getBytes());
//        } catch (IOException e) {
//            e.printStackTrace();
//        }
        selectedObject = model.getEntity(new EntityID(id));
        MrlPlatoonAgent.CHECK_ID = new EntityID(id);
        StaticViewProperties.selectedObject = selectedObject;
        prepareExtraDataPanel();

        repaintSelectedObject();

        for (int i = 3; i < MAX_PROPERTIES; i++) {
            propertiesTable.setValueAt("", i, 0);
            propertiesTable.setValueAt("", i, 1);
        }

        propertiesTable.setValueAt(selectedObject.getURN().replace("urn:rescuecore2.standard:entity:", ""), 0, 0);
        propertiesTable.setValueAt(selectedObject.getID().getValue(), 0, 1);
        propertiesTable.setValueAt("x", 1, 0);
        propertiesTable.setValueAt(selectedObject.getProperty(X_URN).getValue(), 1, 1);
        propertiesTable.setValueAt("y", 2, 0);
        propertiesTable.setValueAt(selectedObject.getProperty(Y_URN).getValue(), 2, 1);

        if (selectedObject instanceof Human) {

            propertiesTable.setValueAt("position", 3, 0);
            propertiesTable.setValueAt(selectedObject.getProperty(POSITION_URN).getValue(), 3, 1);
            propertiesTable.setValueAt("hp", 4, 0);
            propertiesTable.setValueAt(selectedObject.getProperty(HP_URN).getValue(), 4, 1);
            propertiesTable.setValueAt("damage", 5, 0);
            propertiesTable.setValueAt(selectedObject.getProperty(DAMAGE_URN).getValue(), 5, 1);
            propertiesTable.setValueAt("buriedness", 6, 0);
            propertiesTable.setValueAt(selectedObject.getProperty(BURIEDNESS_URN).getValue(), 6, 1);
            propertiesTable.setValueAt("stamina", 7, 0);
            propertiesTable.setValueAt(selectedObject.getProperty(STAMINA_URN).getValue(), 7, 1);
//            try {
//                if (!(selectedObject instanceof Civilian)) {
//                    propertiesTable.setValueAt("act", 8, 0);
//                    propertiesTable.setValueAt(debugDataObject.act(), 8, 1);
//                }
//            } catch (NullPointerException ignore) {
//                propertiesTable.setValueAt(" -null- ", 8, 1);
//            }
            if (selectedObject instanceof FireBrigade) {
                propertiesTable.setValueAt("waterquantity", 9, 0);
                propertiesTable.setValueAt(selectedObject.getProperty(WATER_QUANTITY_URN).getValue(), 9, 1);

            }
            if (selectedObject instanceof Civilian) {
                List<Integer> list = civilianPassableBuildingMap.get(selectedObject.getID());
                propertiesTable.setValueAt("possible buildings", 8, 0);
                if (list != null) {
                    propertiesTable.setValueAt(list.toString(), 8, 1);
                }
            }
        } else if (selectedObject instanceof Area) {
            propertiesTable.setValueAt("neighbours", 3, 0);
            Property p = selectedObject.getProperty(EDGES_URN);
            Collection edges = (Collection) p.getValue();
            for (Object o : edges) {
                if (((Edge) o).isPassable()) {
                    propertiesTable.setValueAt(propertiesTable.getValueAt(3, 1).toString() + ((Edge) o).getNeighbour() + ", ", 3, 1);
                }
            }
            if (selectedObject instanceof Building) {
                propertiesTable.setValueAt("areaground", 4, 0);
                propertiesTable.setValueAt(selectedObject.getProperty("urn:rescuecore2.standard:property:buildingareaground").getValue(), 4, 1);
                propertiesTable.setValueAt("totalarea", 5, 0);
                propertiesTable.setValueAt(selectedObject.getProperty("urn:rescuecore2.standard:property:buildingareatotal").getValue(), 5, 1);
                propertiesTable.setValueAt("floors", 6, 0);
                propertiesTable.setValueAt(selectedObject.getProperty("urn:rescuecore2.standard:property:floors").getValue(), 6, 1);
                propertiesTable.setValueAt("temperature", 7, 0);
                propertiesTable.setValueAt(selectedObject.getProperty(TEMPERATURE_URN).getValue(), 7, 1);
                propertiesTable.setValueAt("ignition", 8, 0);
                propertiesTable.setValueAt(selectedObject.getProperty("urn:rescuecore2.standard:property:ignition").getValue(), 8, 1);
                propertiesTable.setValueAt("fieryness", 9, 0);
                propertiesTable.setValueAt(selectedObject.getProperty(FIERYNESS_URN).getValue(), 9, 1);
                propertiesTable.setValueAt("broknness", 10, 0);
                propertiesTable.setValueAt(selectedObject.getProperty("urn:rescuecore2.standard:property:brokenness").getValue(), 10, 1);

//                try {
//                    propertiesTable.setValueAt("EstimatedTemp", 12, 0);
//                    propertiesTable.setValueAt(TEMP_BUILDING_MAP.get(selectedObject.getID()).getEstimatedTemperature(), 12, 1);
//                    propertiesTable.setValueAt("fieriness", 13, 0);
//                    propertiesTable.setValueAt(TEMP_BUILDING_MAP.get(selectedObject.getID()).getEstimatedFieryness(), 13, 1);
//                } catch (NullPointerException ignored) {
//                }

            } else if (selectedObject instanceof Road) {
                propertiesTable.setValueAt("blockades", 4, 0);
                propertiesTable.setValueAt(selectedObject.getProperty("urn:rescuecore2.standard:property:blockades").getValue(), 4, 1);
            }
        } else if (selectedObject instanceof Blockade) {
            propertiesTable.setValueAt("repair cost", 3, 0);
            propertiesTable.setValueAt(selectedObject.getProperty("urn:rescuecore2.standard:property:repaircost").getValue(), 3, 1);
        }

    }

    private void repaintSelectedObject() {
        viewer.repaint();
    }

    @Override
    protected void postConnect() {
        super.postConnect();
        JFrame frame = new JFrame("Viewer " + getViewerID() + " (" + model.getAllEntities().size() + " entities)");
        frame.setExtendedState(JFrame.MAXIMIZED_BOTH);
        viewer = new MrlAnimatedWorldModelViewer();
        viewer.initialise(config);
        viewer.view(model);
        // CHECKSTYLE:OFF:MagicNumber
//        viewer.setPreferredSize(new Dimension(900, 500));
        // CHECKSTYLE:ON:MagicNumber
        timeLabel = new JLabel("Time: Not started", JLabel.CENTER);
        timeLabel.setBackground(Color.WHITE);
        timeLabel.setOpaque(true);
        timeLabel.setFont(timeLabel.getFont().deriveFont(Font.PLAIN, FONT_SIZE));
        scoreFunction = makeScoreFunction();
        format = NumberFormat.getInstance();
        format.setMaximumFractionDigits(5);
        scoreLabel = new JLabel("   Score: Not started", JLabel.CENTER);
        scoreLabel.setBackground(Color.WHITE);
        scoreLabel.setOpaque(true);
        scoreLabel.setFont(scoreLabel.getFont().deriveFont(Font.PLAIN, FONT_SIZE));

        topPanel.setPreferredSize(new Dimension(1000, 25));
        topPanel.add(timeLabel, BorderLayout.CENTER);
        topPanel.add(scoreLabel, BorderLayout.WEST);

        JScrollPane s = new JScrollPane(propertiesTable);
        s.setMinimumSize(new Dimension(80, 195));
//        s.setBorder(new EtchedBorder(EtchedBorder.RAISED));
        s.setBorder(null);

        final JMenu menu = new JMenu();
        menu.setLayout(new VerticalLayout());
        menu.setBorder(null);
        menu.setEnabled(false);

        BUILDINGS=new FastMap<EntityID,StandardEntity>();
        for(StandardEntity entity:model.getEntitiesOfType(StandardEntityURN.BUILDING)){
            BUILDINGS.put(entity.getID(),entity);
        }


        for (ViewLayer next : viewer.getLayers()) {
            Action action = viewer.getLayersActions().get(next);

            final JMenuBar checkBoxMenuItem = new JMenuBar();
            final JMenuItem menuItem = new JMenuItem(action);
            checkBoxMenuItem.add(next.getName(), menuItem);
            final JPopupMenu innerPopUp;
            innerPopUp = new JPopupMenu();

            List<JMenuItem> items = next.getPopupMenuItems();
            if (items != null && !items.isEmpty()) {
                for (JMenuItem item : items) {
                    innerPopUp.add(item.getName(), item);
                }
            }

            menuItem.addMouseListener(new MouseListener() {
                @Override
                public void mouseClicked(MouseEvent e) {
                }

                @Override
                public void mousePressed(MouseEvent e) {
                }

                @Override
                public void mouseReleased(MouseEvent e) {
                }

                @Override
                public void mouseEntered(MouseEvent e) {
                    if (e.getSource() instanceof JMenuItem) {
                        if (!innerPopUp.isShowing()) {
//                            innerPopUp.show(checkBoxMenuItem, e.getX() + 10, e.getY());
                            int over = 0;
                            if (innerPopUp.getComponents().length > 0) {
                                over = menuItem.getText().length() * 10;
                            }
                            innerPopUp.show(checkBoxMenuItem, menuItem.getLocation().x + over, menuItem.getLocation().y);
                        }
                    }
                }

                @Override
                public void mouseExited(MouseEvent e) {
                    if (e.getSource() instanceof JMenuItem) {
                        if (innerPopUp.isShowing() && !innerPopUp.contains(e.getPoint())) {
                            innerPopUp.setVisible(false);
                        }
                    }
                }
            });
            menu.add(next.getName(), checkBoxMenuItem);
            menu.addSeparator();
        }


        JScrollPane t = new JScrollPane(menu);
        t.setBorder(new EtchedBorder(EtchedBorder.RAISED));
        t.setMinimumSize(new Dimension(80, 500));

        JSplitPane pane = new JSplitPane(JSplitPane.VERTICAL_SPLIT);
        pane.setDividerSize(5);
        pane.setDividerLocation(195);
        pane.setTopComponent(s);
        pane.setBottomComponent(t);

        panel.add(pane);
        panel.setPreferredSize(new Dimension(80, 500));
        final JSplitPane splitPane = new JSplitPane();
        splitPane.setLeftComponent(viewer);
        splitPane.setRightComponent(panel);
        splitPane.setDividerSize(5);

        bottomPanel.setPreferredSize(new Dimension(1000, 30));

        jCheckBoxUpdateAgentData.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent actionEvent) {
                shouldUpdateAgentData = !shouldUpdateAgentData;
            }
        });
        bottomPanel.add(jCheckBoxUpdateAgentData, BorderLayout.EAST);


        JTextField timeTextField = new JTextField();
        timeTextField.setPreferredSize(new Dimension(40, 25));
        timeTextField.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent actionEvent) {
                JTextField field = (JTextField) actionEvent.getSource();
                try {
                    manualTimeStep(Integer.parseInt(field.getText()));
                } catch (Exception ignore) {
                }
            }
        });
        JTextField idTextField = new JTextField();
        idTextField.setPreferredSize(new Dimension(60, 25));
        idTextField.addActionListener(new ActionListener() {

            @Override
            public void actionPerformed(ActionEvent actionEvent) {
                JTextField field = (JTextField) actionEvent.getSource();
                try {
                    selectObject(Integer.parseInt(field.getText()));
                } catch (Exception ignore) {
                }
            }
        });

        JTextField checkIdTextField = new JTextField();
        checkIdTextField.setPreferredSize(new Dimension(60, 25));
        checkIdTextField.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent actionEvent) {
                JTextField field = (JTextField) actionEvent.getSource();
                try {
                    MrlPlatoonAgent.CHECK_ID2 = new EntityID(Integer.parseInt(field.getText()));
                } catch (Exception ignore) {
                }
            }
        });


        JCheckBox jCheckBoxPause = new JCheckBox("Pause");
        jCheckBoxPause.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent actionEvent) {
                pause = !pause;
            }
        });

        JButton refreshColor = new JButton("refresh color");
        refreshColor.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent actionEvent) {
                randomValue = System.currentTimeMillis();
                viewer.repaint();
            }
        });
        JPanel bPanel = new JPanel();

        bPanel.add(refreshColor);

        bPanel.add(new JLabel("Time:"));
        bPanel.add(timeTextField);

        bPanel.add(jCheckBoxPause);

        bPanel.add(new JLabel(" ID:"));
        bPanel.add(idTextField);

        bPanel.add(new JLabel(" check:"));
        bPanel.add(checkIdTextField);

        bPanel.setPreferredSize(new Dimension(600, 25));

        bottomPanel.add(bPanel, BorderLayout.WEST);


        frame.add(topPanel, BorderLayout.NORTH);
        frame.add(splitPane, BorderLayout.CENTER);
        frame.add(bottomPanel, BorderLayout.SOUTH);

        frame.pack();
        frame.setExtendedState(JFrame.MAXIMIZED_BOTH);
        splitPane.setDividerLocation(0.99);
        frame.setVisible(true);
        frame.addComponentListener(new ComponentListener() {
            @Override
            public void componentResized(ComponentEvent e) {
                splitPane.setDividerLocation(0.8);
            }

            @Override
            public void componentMoved(ComponentEvent e) {
            }

            @Override
            public void componentShown(ComponentEvent e) {
            }

            @Override
            public void componentHidden(ComponentEvent e) {
            }
        });

        frame.addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent e) {
                System.exit(0);
            }
        });

        viewer.addViewListener(new ViewListener() {

            @Override
            public void objectsClicked(ViewComponent view, List<RenderedObject> objects) {

                if (objects.isEmpty()) {
                    StaticViewProperties.selectedObject = null;
                    selectedObject = null;
                    MrlPlatoonAgent.CHECK_ID = null;
                    MrlPlatoonAgent.CHECK_ID2 = null;
                    viewer.repaint();
                }
                if (objects.size() == 1) {
                    StandardEntity entity = (StandardEntity) objects.get(0).getObject();
                    selectObject(entity.getID().getValue());
                } else {
                    int c = 0;
                    int id = -1;
                    popup.removeAll();
                    for (RenderedObject next : objects) {
                        StandardEntity entity = (StandardEntity) next.getObject();
                        if (entity instanceof Human) {
                            c++;
                            if (c == 1) {
                                id = entity.getID().getValue();
                            }
                        }
                        JMenuItem jmi = new JMenuItem(entity.getID().getValue() + "-" + entity);
                        jmi.addActionListener(new ActionListener() {
                            //                            @Override
                            public void actionPerformed(ActionEvent actionEvent) {

                                JMenuItem source = (JMenuItem) actionEvent.getSource();
                                StringTokenizer st = new StringTokenizer(source.getText(), "-");
                                String menuID = st.nextToken();
                                int id = Integer.parseInt(menuID);
                                selectObject(id);
                            }
                        });
                        popup.add(jmi);
                    }

                    if (c != 1) {
                        double x = MouseInfo.getPointerInfo().getLocation().getX();
                        double y = MouseInfo.getPointerInfo().getLocation().getY();

                        popup.show(viewer, (int) Math.round(x), (int) Math.round(y));
                    } else if (id != -1) {
                        selectObject(id);
                    }
                }
            }

            //            @Override
            public void objectsRollover(ViewComponent view, List<RenderedObject> objects) {
            }
        });


//        tree.addTreeSelectionListener(new TreeSelectionListener() {
//            public void valueChanged(TreeSelectionEvent e) {
//                try {
//                    DefaultMutableTreeNode node = (DefaultMutableTreeNode) tree.getLastSelectedPathComponent();
//                    /* if nothing is selected */
//                    if (node == rootNode) {
//                        selectedTreeObject = null;
//                        prepareExtraDataPanel();
//                        return;
//                    }
//                    selectedTreeObject = null;
//                    if (node == null)
//                        return;
//
//                    selectedTreeObject = node.getUserObject();
//                    expandNode(node);
//                    treeNodes = node.getPath();
//                    tree.updateUI();
//                    repaintSelectedObject();
//                } catch (Exception ignore) {
//                }
//            }
//        });
//
//        tree.addMouseListener(new MouseListener() {
//
//            public void mouseClicked(MouseEvent e) {
//                try {
//                    if (e.getButton() == MouseEvent.BUTTON3) {
////                        updateDataObject();
//                        prepareExtraDataPanel();
//                        treeNodes = null;
//                        StaticViewProperties.objectToPaint.clear();
//                    } else {
//                        if (e.getClickCount() == 2) {
//                            DefaultMutableTreeNode node = (DefaultMutableTreeNode) tree.getLastSelectedPathComponent();
//                            if (node != null)
//                                if (!tree.isCollapsed(new TreePath(node.getPath()))) {
//                                    tree.expandPath(new TreePath(node.getPath()));
//                                } else {
//                                    tree.collapsePath(new TreePath(node.getPath()));
//                                }
//                        } else {
//                            StaticViewProperties.objectToPaint.clear();
//                        }
//                    }
//                } catch (Exception ex) {
////                    ex.printStackTrace();
//                }
//            }
//
//            public void mousePressed(MouseEvent e) {
//            }
//
//            public void mouseReleased(MouseEvent e) {
//            }
//
//            public void mouseEntered(MouseEvent e) {
//            }
//
//            public void mouseExited(MouseEvent e) {
//            }
//        });

    }

    protected void handleTimeStepProcess(final KVTimestep t) {
        if (pause)
            return;
//        Collection<StandardEntity> buildings=model.getEntitiesOfType(StandardEntityURN.BUILDING);
//
//        for(StandardEntity entity: buildings){
//            Building building= (Building) entity;
//            building.setFieryness(0);
//        }
//        for(Integer i:dataMap.keySet() ){
//
//            if(i<t.getTime()){
//                super.handleTimestep(dataMap.get(i));
//            }else{
//                break;
//            }
//        }
        super.handleTimestep(t);


        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                time = t.getTime();
                timeLabel.setText("Time: " + t.getTime());
                scoreValue = format.format(scoreFunction.score(model, new Timestep(t.getTime())));
                scoreLabel.setText("    Score: " + scoreValue);
                if (selectedObject != null && shouldUpdateAgentData) {
                    selectObject(selectedObject.getID().getValue());
                }
                viewer.view(model, t.getCommands());
//                if (t.getTime() == 300) {
//                printOutCivilianInformations();
//                }

                viewer.repaint();
            }
        });
    }

//    LearnerIO learnerIO_Log = new LearnerIO("data/ambulance/" + "Test" + ".dta", true);
//
//    private void printOutCivilianInformations() {
//        long totalHp = 0;
//        int totalRescuedCivs = 0;
//        String data = "";
//        String filename = "data/ambulance/" + "Test" + ".dta";
//        String filename_Log = "data/ambulance/" + "Test" + "_Market_Log.dta";
//        Civilian civilian;
//        for (StandardEntity st : model.getEntitiesOfType(StandardEntityURN.CIVILIAN)) {
//            civilian = (Civilian) st;
//            if (civilian.getPosition(model) instanceof Refuge) {
//                totalHp += civilian.getHP();
//                totalRescuedCivs++;
//            }
//        }
//
//        data += " " + time + " " + totalRescuedCivs + " " + totalHp + "              " + scoreValue;
////        if (time == 300) {
////            LearnerIO learnerIO = new LearnerIO("data/ambulance/" + "Test" + ".dta", true);
////            learnerIO.simplePrintToFile(filename, " " + totalRescuedCivs + " " + totalHp + "              " + scoreValue);
////        }
//
//        if (time == 1) {
//
//            learnerIO_Log.simplePrintToFile(filename, "___________________________________________________\n");
//
//        }
//        learnerIO_Log.simplePrintToFile(filename, data);
//
//
//    }

    @Override
    protected void handleTimestep(final KVTimestep t) {
        dataMap.put(t.getTime(), t);
        if (!isSetToManual) {
            handleTimeStepProcess(t);
        } else {
            KVTimestep timestep = dataMap.get(++currentTime);
            if (timestep != null)
                handleTimeStepProcess(timestep);
        }
    }

    protected void manualTimeStep(int timeStep) {
        KVTimestep kvTimestep = dataMap.get(timeStep);
        if (kvTimestep != null) {
            currentTime = timeStep - 1;
            isSetToManual = true;
            handleTimestep(kvTimestep);
        } else {
            Logger.error("Null KVTimeStep");
        }
    }

    @Override
    public String toString() {
        return "MRLs viewer";
    }


    private void prepareExtraDataPanel() {
        if (selectedObject == null) {
            return;
        }
        rootNode.removeAllChildren();

        expandNode(rootNode);
        tree.updateUI();
//        paintField.updateUI();
    }

    private void expandNode(DefaultMutableTreeNode node) {
        Object nodeInfo = node.getUserObject();
        if (!node.children().hasMoreElements()) {
            for (Method m : nodeInfo.getClass().getMethods()) {
                if (m.getAnnotation(AttributeField.class) != null) {

                    DefaultMutableTreeNode childNode = new DefaultMutableTreeNode(m.getName());
                    node.add(childNode);
                    Object obj;
                    try {
                        obj = m.invoke(nodeInfo);
                        if (obj != null) {
                            if (obj instanceof ArrayList) {
                                for (Object o : (ArrayList) obj) {
                                    childNode.add(new DefaultMutableTreeNode(o));
                                }
                            } else {
                                childNode.add(new DefaultMutableTreeNode(obj));
                            }
                        }
                    } catch (Exception ignore) {
                    }

                }
            }

        }

    }

    private ScoreFunction makeScoreFunction() {
        String className = config.getValue(Constants.SCORE_FUNCTION_KEY);
        ScoreFunction result = instantiate(className, ScoreFunction.class);
        result.initialise(model, config);
        return result;
    }
}
