package mrl.viewer.layers;

import javolution.util.FastMap;
import javolution.util.FastSet;
import math.geom2d.conic.Circle2D;
import mrl.common.ConvexHull;
import mrl.common.clustering.ConvexObject;
import mrl.viewer.StaticViewProperties;
import mrl.world.object.MrlBuilding;
import rescuecore2.misc.gui.ScreenTransform;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.view.Icons;
import rescuecore2.view.RenderedObject;
import rescuecore2.worldmodel.EntityID;

import javax.swing.*;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.util.*;
import java.util.List;

/**
 * Created by IntelliJ IDEA.
 * User: sajjadsalehi
 * Date: 12/16/11
 * Time: 6:59 PM
 * Edited by Mostafa Shabani
 * Rided by Vahid hooshangi
 */
public class MrlConvexHullLayer extends MrlPreRoutingPartitionsLayer {
    private static final Color ConColour = Color.CYAN;
    private static final Color BORDER_DIRECTION_ENTIY_Colour = Color.orange;
    private static final Color BORDER_ENTIY_Colour = Color.WHITE;
    private static final Color BORDER_MAP_Colour = Color.RED;
    private static final Color TARGET_COLOR = Color.PINK;
    private static final Color STAND_NEAR_TARGET = new Color(Color.BLUE.getRed(),Color.BLUE.getBlue(), Color.BLUE.getGreen(), 100);
    private static final Color SMALL_BORDER_COLOUR = Color.magenta;
    private static final Color BIG_BORDER_COLOUR = Color.yellow;
    private static final Stroke STROKE = new BasicStroke(3, BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER);
    private static final Stroke STROKE_DIRECTION = new BasicStroke(5, BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER);
    public static final Map<EntityID, List<ConvexObject>> CONVEX_HULLS_MAP = new FastMap<EntityID, List<ConvexObject>>();
    public static Point CENTER_POINT;
    public static Map<EntityID, StandardEntity> BUILDINGS = new FastMap<EntityID, StandardEntity>();//TODO :just for test
    public static List<Polygon> BIG_BORDER_HULLS = new ArrayList<Polygon>();
    public static Map<EntityID,List<StandardEntity>> BORDER_BUILDINGS =  Collections.synchronizedMap(new FastMap<EntityID, List<StandardEntity>>());
    public static Map<EntityID, List<StandardEntity>> BORDER_DIRECTION_BUILDINGS = Collections.synchronizedMap(new FastMap<EntityID, List<StandardEntity>>());
    public static Map<EntityID, Set<StandardEntity>> BEST_PLACE_TO_STAND = Collections.synchronizedMap(new FastMap<EntityID, Set<StandardEntity>>());
    public static Map<EntityID, List<Building>> MY_TARGET = Collections.synchronizedMap(new FastMap<EntityID, List<Building>>());
    public static Map<EntityID, List<StandardEntity>> STAND = Collections.synchronizedMap(new FastMap<EntityID, List<StandardEntity>>());
    public static int EXTINGUISH_RANGE;

    public static Map<EntityID, Set<EntityID>> BORDER_MAP_BUILDINGS = Collections.synchronizedMap(new FastMap<EntityID, Set<EntityID>>());
    public static ConvexHull BIG_Whole_BORDER_HULL = new ConvexHull();
    public static Polygon SMALL_Whole_BORDER_HULL = new Polygon();

    public static List<Polygon> SMALL_BORDER_HULLS = new ArrayList<Polygon>();

    private boolean borderEntitiesInfo;
    private BorderEntitiesAction borderEntitiesAction;

    private boolean borderDirectionsInfo;
    private BorderDirectionEntitiesAction borderDirectionEntitiesAction;

    private boolean borderHullsInfo;
    private BorderHullsAction borderHullsAction;

    private boolean othersInfo;
    private OthersAction othersAction;

    private boolean borderMapBuildingsInfo;
    private BorderMapBuildingsAction borderMapBuildingsAction;

    private boolean showTargetInfo;
    private ShowTargetAction showTargetAction;

    public MrlConvexHullLayer() {
        borderEntitiesInfo = false;
        borderEntitiesAction = new BorderEntitiesAction();
        borderHullsInfo = false;
        borderHullsAction = new BorderHullsAction();
        borderDirectionsInfo = false;
        borderDirectionEntitiesAction = new BorderDirectionEntitiesAction();
        othersInfo = false;
        othersAction = new OthersAction();
        borderMapBuildingsInfo = false;
        borderMapBuildingsAction = new BorderMapBuildingsAction();
        showTargetInfo = false;
        showTargetAction = new ShowTargetAction();
    }

    @Override
    public String getName() {
        return "Convex Hull";
    }


    @Override
    public List<JMenuItem> getPopupMenuItems() {
        List<JMenuItem> result = new ArrayList<JMenuItem>();
        result.add(new JMenuItem(borderEntitiesAction));
        result.add(new JMenuItem(borderHullsAction));
        result.add(new JMenuItem(othersAction));
        result.add(new JMenuItem(borderDirectionEntitiesAction));
        result.add(new JMenuItem(borderMapBuildingsAction));
        result.add(new JMenuItem(showTargetAction));
        return result;
    }

    @Override
    public Collection<RenderedObject> render(Graphics2D g, ScreenTransform t, int width, int height) {

        for (StandardEntity entity : world.getEntitiesOfType(StandardEntityURN.BUILDING)) {
            BUILDINGS.put(entity.getID(), entity);
        }

        List<Integer> xs = new ArrayList<Integer>();
        List<Integer> ys = new ArrayList<Integer>();
        Collection<RenderedObject> list = new ArrayList<RenderedObject>();


        List<ConvexObject> convexObjects = new ArrayList<ConvexObject>();
        List<Polygon> smallPolygons = new ArrayList<Polygon>();
        List<Polygon> bigPolygons = new ArrayList<Polygon>();
        List<StandardEntity> borderBuildings = new ArrayList<StandardEntity>();
        List<StandardEntity> borderDirectionBuildings = new ArrayList<StandardEntity>();
        Set<EntityID> borderMapBuildings = new FastSet<EntityID>();
        List<Building> target = new ArrayList<Building>();
        Set<StandardEntity> pathToFire = new HashSet<StandardEntity>();
        List<StandardEntity> standEntity = new ArrayList<StandardEntity>();
        //Building target;
//        ConvexHull borderMapConvex = new ConvexHull();

        if (StaticViewProperties.selectedObject != null) {
            try {
                if (CONVEX_HULLS_MAP.get(StaticViewProperties.selectedObject.getID()) != null) {
                    convexObjects = Collections.synchronizedList(CONVEX_HULLS_MAP.get(StaticViewProperties.selectedObject.getID()));
                }
                if (SMALL_BORDER_HULLS != null ) {
                    smallPolygons = Collections.synchronizedList(SMALL_BORDER_HULLS);
                }
                if (BIG_BORDER_HULLS != null) {
                bigPolygons = Collections.synchronizedList(BIG_BORDER_HULLS);
                }
                if (BORDER_BUILDINGS.get(StaticViewProperties.selectedObject.getID()) != null) {
                borderBuildings = Collections.synchronizedList(BORDER_BUILDINGS.get(StaticViewProperties.selectedObject.getID()));
                }
                if (BORDER_DIRECTION_BUILDINGS.get(StaticViewProperties.selectedObject.getID()) != null) {
                borderDirectionBuildings = Collections.synchronizedList(BORDER_DIRECTION_BUILDINGS.get(StaticViewProperties.selectedObject.getID()));
                }
                if (BORDER_MAP_BUILDINGS.get(StaticViewProperties.selectedObject.getID()) != null) {
                borderMapBuildings = Collections.synchronizedSet(BORDER_MAP_BUILDINGS.get(StaticViewProperties.selectedObject.getID()));
                }
                if (MY_TARGET.get(StaticViewProperties.selectedObject.getID()) != null) {
                target = Collections.synchronizedList(MY_TARGET.get(StaticViewProperties.selectedObject.getID()));
                }
                if (BEST_PLACE_TO_STAND.get(StaticViewProperties.selectedObject.getID()) != null) {
                pathToFire = Collections.synchronizedSet(BEST_PLACE_TO_STAND.get(StaticViewProperties.selectedObject.getID()));
                }
                if (STAND.get(StaticViewProperties.selectedObject.getID()) != null) {
                standEntity = Collections.synchronizedList(STAND.get(StaticViewProperties.selectedObject.getID()));
                }
                //borderMapConvex = Collections.synchronizedMap(BORDER_HULL.get(StaticViewProperties.selectedObject.getID()));
            } catch (NullPointerException ignored) {
                ignored.printStackTrace();
            }
        }


        if (borderEntitiesInfo) {
            RenderBorderEntitiesAction(g, t, borderBuildings);
        }

        if (borderHullsInfo) {
            RenderBorderHullsAction(g, t, smallPolygons, bigPolygons);
        }

        if (borderMapBuildingsInfo) {
            renderBorderMapBuildings(g, t, borderMapBuildings);
        }

        if (borderDirectionsInfo)
        {
            RenderBorderDirectionEntitiesAction(g,t, borderDirectionBuildings);
        }
        if (showTargetInfo) {
//            if (!target.isEmpty() && !standEntity.isEmpty()) {
                renderTarget(g, t, target/*.get(target.size()-1)*/,standEntity/*.get(standEntity.size()-1)*/ , pathToFire);
//            }
        }

        for (ConvexObject convexObject : convexObjects) {
            xs.clear();
            ys.clear();

            for (int x : convexObject.getConvexPolygon().xpoints) {
                xs.add(t.xToScreen(x));
            }
            for (int y : convexObject.getConvexPolygon().ypoints) {
                ys.add(t.yToScreen(y));
            }

            Polygon poly = new Polygon(listToArray(xs), listToArray(ys), convexObject.getConvexPolygon().npoints);
            g.setColor(ConColour);
            g.setStroke(STROKE);
            g.draw(poly);

            if (othersInfo)
            {
                RenderOthers(g,t,convexObject);
            }
        }

        return list;
    }

    public void renderTarget(Graphics2D g, ScreenTransform t, List<Building> targetList, List<StandardEntity> standList, Set<StandardEntity> stations){
        g.setStroke(STROKE);
        List<Integer> xs = new ArrayList<Integer>();
        List<Integer> ys = new ArrayList<Integer>();
        Polygon poly;

        Building building;
        Road road;
        g.setColor(STAND_NEAR_TARGET);
        
        for (StandardEntity entity : stations)
        {
            xs.clear();
            ys.clear();

            if (entity instanceof Building) {
                building = (Building) entity;
                for (int i = 0; i < building.getApexList().length; i += 2) {
                    xs.add(t.xToScreen(building.getApexList()[i]));
                    ys.add(t.yToScreen(building.getApexList()[i + 1]));
                }
                poly = new Polygon(listToArray(xs), listToArray(ys), building.getApexList().length / 2);

                g.fill(poly);
            }
            else if (entity instanceof Road) {
                road = (Road) entity;
                for (int i = 0; i < road.getApexList().length; i += 2) {
                    xs.add(t.xToScreen(road.getApexList()[i]));
                    ys.add(t.yToScreen(road.getApexList()[i + 1]));
                }
                poly = new Polygon(listToArray(xs), listToArray(ys), road.getApexList().length / 2);

                g.fill(poly);
            }
        }

        xs.clear();
        ys.clear();
        if (!standList.isEmpty())
        {
            StandardEntity stand = standList.get(standList.size()-1);
        g.setColor(Color.YELLOW);
            if (stand != null)
            {
                if (stand instanceof Road)
                {
                    Road rd = (Road) stand;
                    for (int i = 0; i < rd.getApexList().length; i += 2) {
                        xs.add(t.xToScreen(rd.getApexList()[i]));
                        ys.add(t.yToScreen(rd.getApexList()[i + 1]));
                    }
                    poly = new Polygon(listToArray(xs), listToArray(ys), rd.getApexList().length/2);
                    g.fill(poly);
                }
                else if (stand instanceof Building)
                {
                    Building bl = (Building) stand;
                    for (int i = 0; i < bl.getApexList().length; i += 2) {
                        xs.add(t.xToScreen(bl.getApexList()[i]));
                        ys.add(t.yToScreen(bl.getApexList()[i + 1]));
                    }
                    poly = new Polygon(listToArray(xs), listToArray(ys), bl.getApexList().length/2);
                    g.fill(poly);
                }
            }
        }

        xs.clear();
        ys.clear();
        g.setColor(TARGET_COLOR);
        if (!targetList.isEmpty())
        {
            Building target = targetList.get(targetList.size()-1);
            if (target != null) {
                for (int i = 0; i < target.getApexList().length; i += 2) {
                    xs.add(t.xToScreen(target.getApexList()[i]));
                    ys.add(t.yToScreen(target.getApexList()[i + 1]));
                }
                poly = new Polygon(listToArray(xs), listToArray(ys), target.getApexList().length/2);
                g.fill(poly);
                int d = t.xToScreen(target.getX() + EXTINGUISH_RANGE ) - t.xToScreen(target.getX());
                Circle2D circle2D = new Circle2D(t.xToScreen(target.getX()), t.yToScreen(target.getY()), d, true);
                circle2D.draw(g);
            }
        }
    }

    public void renderBorderMapBuildings(Graphics2D g, ScreenTransform t, Set<EntityID> borderMapBuildings) {
        Building building;
        g.setColor(BORDER_MAP_Colour);
        g.setStroke(STROKE_DIRECTION);
        List<Integer> xs = new ArrayList<Integer>();
        List<Integer> ys = new ArrayList<Integer>();
        Polygon poly;

        poly = BIG_Whole_BORDER_HULL.convex();
        g.draw(poly);
        
        poly = SMALL_Whole_BORDER_HULL;
        g.draw(poly);

        for (EntityID entityID : borderMapBuildings) {
            StandardEntity entity = world.getEntity(entityID);
            building = (Building) entity;

            xs.clear();
            ys.clear();

            for (int i = 0; i < building.getApexList().length; i += 2) {
                xs.add(t.xToScreen(building.getApexList()[i]));
                ys.add(t.yToScreen(building.getApexList()[i + 1]));
            }
            poly = new Polygon(listToArray(xs), listToArray(ys), building.getApexList().length / 2);

            g.draw(poly);
        }


    }

    public static int[] listToArray(List<Integer> list) {
        int[] array = new int[list.size()];
        for (int i = 0; i < list.size(); i++)
            array[i] = list.get(i);
        return array;
    }


    private void RenderBorderEntitiesAction(Graphics2D g, ScreenTransform t, List<StandardEntity> border_entities) {
        Building building;
        g.setColor(BORDER_ENTIY_Colour);
        g.setStroke(STROKE);
        List<Integer> xs = new ArrayList<Integer>();
        List<Integer> ys = new ArrayList<Integer>();
        Collection<RenderedObject> list = new ArrayList<RenderedObject>();
        Polygon poly;
        for (StandardEntity entity : border_entities) {
            building = (Building) entity;

            xs.clear();
            ys.clear();

            for (int i = 0; i < building.getApexList().length; i += 2) {
                xs.add(t.xToScreen(building.getApexList()[i]));
                ys.add(t.yToScreen(building.getApexList()[i + 1]));
            }
            poly = new Polygon(listToArray(xs), listToArray(ys), building.getApexList().length / 2);

            g.draw(poly);
        }
    }

    private void RenderBorderDirectionEntitiesAction(Graphics2D g, ScreenTransform t, List<StandardEntity> border_direction_entities) {
        Building building;
        g.setColor(BORDER_DIRECTION_ENTIY_Colour);
        g.setStroke(STROKE_DIRECTION);
        List<Integer> xs = new ArrayList<Integer>();
        List<Integer> ys = new ArrayList<Integer>();
        Polygon poly;
        for (StandardEntity entity : border_direction_entities) {
            building = (Building) entity;

            xs.clear();
            ys.clear();

            for (int i = 0; i < building.getApexList().length; i += 2) {
                xs.add(t.xToScreen(building.getApexList()[i]));
                ys.add(t.yToScreen(building.getApexList()[i + 1]));
            }
            poly = new Polygon(listToArray(xs), listToArray(ys), building.getApexList().length / 2);

            g.draw(poly);
        }
    }

    private void RenderOthers(Graphics2D g, ScreenTransform t, ConvexObject convexObject)
    {
        List<Integer> xs = new ArrayList<Integer>();
        List<Integer> ys = new ArrayList<Integer>();
        if (convexObject.FIRST_POINT != null) {
            Set<Line2D> lines = Collections.synchronizedSet((convexObject.CONVEX_INTERSECT_LINES == null ? new FastSet<Line2D>() : convexObject.CONVEX_INTERSECT_LINES));
            if (lines != null) {
                g.setColor(Color.RED);
                g.setStroke(new BasicStroke(9));
                for (Line2D line : lines) {
                    g.drawLine(t.xToScreen(line.getX1()), t.yToScreen(line.getY1()), t.xToScreen(line.getX2()), t.yToScreen(line.getY2()));
                }
            }

            g.setStroke(new BasicStroke(1, BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL));
            g.setColor(Color.YELLOW);
            Circle2D circle2D = new Circle2D(t.xToScreen(CENTER_POINT.x), t.yToScreen(CENTER_POINT.y), 10, true);
            circle2D.draw(g);
            circle2D = new Circle2D(t.xToScreen(convexObject.CONVEX_POINT.x), t.yToScreen(convexObject.CONVEX_POINT.y), 10, true);
            circle2D.draw(g);
            g.drawLine(t.xToScreen(CENTER_POINT.x), t.yToScreen(CENTER_POINT.y), t.xToScreen(convexObject.CONVEX_POINT.x), t.yToScreen(convexObject.CONVEX_POINT.y));

            g.setStroke(new BasicStroke(1.8f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL));
            g.setColor(Color.GREEN);
            g.drawLine(t.xToScreen(convexObject.FIRST_POINT.x), t.yToScreen(convexObject.FIRST_POINT.y), t.xToScreen(convexObject.SECOND_POINT.x), t.yToScreen(convexObject.SECOND_POINT.y));

            if (convexObject.CONVEX_INTERSECT_POINTS != null)
            {
                for (Point2D point : convexObject.CONVEX_INTERSECT_POINTS)
                {
                    circle2D = new Circle2D(t.xToScreen(point.getX()), t.yToScreen(point.getY()), 7, true);
                    circle2D.draw(g);
                }
            }
            g.setStroke(new BasicStroke(1.7f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL, 0, new float[]{9}, 0));
            g.drawLine(t.xToScreen(convexObject.FIRST_POINT.x), t.yToScreen(convexObject.FIRST_POINT.y), t.xToScreen(convexObject.CONVEX_POINT.x), t.yToScreen(convexObject.CONVEX_POINT.y));
            g.drawLine(t.xToScreen(convexObject.SECOND_POINT.x), t.yToScreen(convexObject.SECOND_POINT.y), t.xToScreen(convexObject.CONVEX_POINT.x), t.yToScreen(convexObject.CONVEX_POINT.y));
            g.drawLine(t.xToScreen(convexObject.FIRST_POINT.x), t.yToScreen(convexObject.FIRST_POINT.y), t.xToScreen(convexObject.OTHER_POINT1.x), t.yToScreen(convexObject.OTHER_POINT1.y));
            g.drawLine(t.xToScreen(convexObject.SECOND_POINT.x), t.yToScreen(convexObject.SECOND_POINT.y), t.xToScreen(convexObject.OTHER_POINT2.x), t.yToScreen(convexObject.OTHER_POINT2.y));
            g.drawLine(t.xToScreen(convexObject.OTHER_POINT1.x), t.yToScreen(convexObject.OTHER_POINT1.y), t.xToScreen(convexObject.OTHER_POINT2.x), t.yToScreen(convexObject.OTHER_POINT2.y));
        }
        if (convexObject.DIRECTION_POLYGON != null) {
            Polygon polys = convexObject.DIRECTION_POLYGON;

            xs.clear();
            ys.clear();
            for (int x : polys.xpoints) {
                xs.add(t.xToScreen(x));
            }
            for (int y : polys.ypoints) {
                ys.add(t.yToScreen(y));
            }
            Polygon polygon = new Polygon(listToArray(xs), listToArray(ys), polys.npoints);

            g.setStroke(new BasicStroke(3f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL));
            g.setColor(Color.BLACK);
            /*g.draw(polygon);*/
        }
    }

    private void RenderBorderHullsAction(Graphics2D g, ScreenTransform t, List<Polygon> smallBorderHulls, List<Polygon> bigBorderHulls) {
        g.setStroke(STROKE);
        g.setColor(SMALL_BORDER_COLOUR);
        Collection<RenderedObject> list = new ArrayList<RenderedObject>();
        List<Integer> xs = new ArrayList<Integer>();
        List<Integer> ys = new ArrayList<Integer>();
        for (Polygon polygon : smallBorderHulls) {
            xs.clear();
            ys.clear();

            for (int x : polygon.xpoints) {
                xs.add(t.xToScreen(x));
            }
            for (int y : polygon.ypoints) {
                ys.add(t.yToScreen(y));
            }

            Polygon poly = new Polygon(listToArray(xs), listToArray(ys), polygon.npoints);
            g.draw(poly);
        }

        g.setColor(BIG_BORDER_COLOUR);

        for (Polygon polygon : bigBorderHulls) {
            xs.clear();
            ys.clear();

            for (int x : polygon.xpoints) {
                xs.add(t.xToScreen(x));
            }
            for (int y : polygon.ypoints) {
                ys.add(t.yToScreen(y));
            }

            Polygon poly = new Polygon(listToArray(xs), listToArray(ys), polygon.npoints);
            g.draw(poly);
        }
    }

    private final class BorderEntitiesAction extends AbstractAction {
        public BorderEntitiesAction() {
            super("border entities");
            update();
        }


        @Override
        public void actionPerformed(ActionEvent e) {
            setBorderEntitiesInfo(!borderEntitiesInfo);
            component.repaint();
        }

        void update() {
            putValue(Action.SELECTED_KEY, Boolean.valueOf(borderEntitiesInfo));
            putValue(Action.SMALL_ICON, borderEntitiesInfo ? Icons.TICK : Icons.CROSS);
        }

    }

    public void setBorderEntitiesInfo(boolean render) {
        borderEntitiesInfo = render;
        borderEntitiesAction.update();
    }


    private final class BorderHullsAction extends AbstractAction {

        public BorderHullsAction() {
            super("border hulls");
            update();
        }

        @Override
        public void actionPerformed(ActionEvent e) {
            setBorderHullsInfo(!borderHullsInfo);
            component.repaint();
        }

        void update() {
            putValue(Action.SELECTED_KEY, Boolean.valueOf(borderHullsInfo));
            putValue(Action.SMALL_ICON, borderHullsInfo ? Icons.TICK : Icons.CROSS);
        }

    }

    public void setBorderHullsInfo(boolean render) {
        borderHullsInfo = render;
        borderHullsAction.update();
    }

    private final class OthersAction extends AbstractAction
    {

        public OthersAction() {
            super("Direction Info");
            update();
        }

        @Override
        public void actionPerformed(ActionEvent e) {
            setOthersInfo(!othersInfo);
            component.repaint();
        }

        public void update()
        {
            putValue(Action.SELECTED_KEY,Boolean.valueOf(othersInfo));
            putValue(Action.SMALL_ICON, othersInfo ? Icons.TICK : Icons.CROSS);
        }
    }

    public void setOthersInfo(boolean render)
    {
        othersInfo = render;
        othersAction.update();
    }

    private class BorderDirectionEntitiesAction extends AbstractAction
    {

        public BorderDirectionEntitiesAction()
        {
            super("Border Dir Info");
            update();
        }

        private void update() {
            putValue(Action.SELECTED_KEY,Boolean.valueOf(borderDirectionsInfo));
            putValue(Action.SMALL_ICON, borderDirectionsInfo ? Icons.TICK : Icons.CROSS);
        }

        @Override
        public void actionPerformed(ActionEvent e) {
            setBorderDirectionsInfo(!borderDirectionsInfo);
            component.repaint();
        }
    }

    public void setBorderDirectionsInfo(boolean render)
    {
        borderDirectionsInfo = render;
        borderDirectionEntitiesAction.update();
    }

    private class BorderMapBuildingsAction extends AbstractAction {

        public BorderMapBuildingsAction() {
            super("Border Map Info");
            update();
        }

        public void update() {
            putValue(Action.SELECTED_KEY,Boolean.valueOf(borderMapBuildingsInfo));
            putValue(Action.SMALL_ICON, borderMapBuildingsInfo ? Icons.TICK : Icons.CROSS);
        }

        @Override
        public void actionPerformed(ActionEvent e) {
            setBorderMapBuildingsInfo(!borderMapBuildingsInfo);
            component.repaint();
        }
    }

    public void setBorderMapBuildingsInfo (boolean render) {
        borderMapBuildingsInfo = render;
        borderMapBuildingsAction.update();
    }

    private class ShowTargetAction extends AbstractAction {

        public ShowTargetAction() {
            super("Show Target");
            update();
        }

        public void update() {
            putValue(Action.SELECTED_KEY,Boolean.valueOf(showTargetInfo));
            putValue(Action.SMALL_ICON, showTargetInfo ? Icons.TICK : Icons.CROSS);
        }

        @Override
        public void actionPerformed(ActionEvent e) {
            setShowTargetInfo(!showTargetInfo);
            component.repaint();
        }
    }

    public void setShowTargetInfo(boolean render) {
        showTargetInfo = render;
        showTargetAction.update();
    }
}
