package mrl.viewer.layers;

import mrl.viewer.StandardEntityToPaint;
import mrl.viewer.StaticViewProperties;
import rescuecore2.config.Config;
import rescuecore2.misc.gui.ScreenTransform;
import rescuecore2.standard.entities.Edge;
import rescuecore2.standard.entities.Road;
import rescuecore2.view.Icons;

import javax.swing.*;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.util.*;
import java.util.List;

/**
 * Created by Mostafa Shabani.
 * Date: Dec 12, 2010
 * Time: 4:42:03 PM
 */
public class MrlRoadLayer extends MrlAreaLayer<Road> {
    private static final Color ROAD_EDGE_COLOUR = Color.GRAY.darker();
    private static final Color ROAD_SHAPE_COLOUR = new Color(185, 185, 185);

    private static final Stroke WALL_STROKE = new BasicStroke(2f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER);
    private static final Stroke ENTRANCE_STROKE = new BasicStroke(0.3f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER);

    private Action showIdsAction;
    private Action showRoadScaleAction;

    /**
     * Construct a road rendering layer.
     */
    public MrlRoadLayer() {
        super(Road.class);
    }

    @Override
    public String getName() {
        return "Roads";
    }

    @Override
    protected void paintShape(Road r, Polygon shape, Graphics2D g) {
        StandardEntityToPaint entityToPaint = StaticViewProperties.getPaintObject(r);
        if (entityToPaint != null) {
            g.setColor(entityToPaint.getColor());
            g.fill(shape);
            return;

        } else {
            if (r == StaticViewProperties.selectedObject) {
                g.setColor(Color.MAGENTA);
            } else {
                g.setColor(ROAD_SHAPE_COLOUR);
            }
        }
        g.fill(shape);
    }

    @Override
    protected void paintEdge(Edge e, Graphics2D g, ScreenTransform t) {
        g.setColor(ROAD_EDGE_COLOUR);
        g.setStroke(e.isPassable() ? ENTRANCE_STROKE : WALL_STROKE);
        g.drawLine(t.xToScreen(e.getStartX()),
                t.yToScreen(e.getStartY()),
                t.xToScreen(e.getEndX()),
                t.yToScreen(e.getEndY()));
    }

    @Override
    public void initialise(Config config) {
        showRIds = false;
        showRoadScale = false;
        showIdsAction = new ShowIdsAction();
        showRoadScaleAction = new ShowRoadScaleAction();
    }

    private final class ShowIdsAction extends AbstractAction {
        public ShowIdsAction() {
            super("show ids");
            putValue(Action.SELECTED_KEY, Boolean.valueOf(showRIds));
            putValue(Action.SMALL_ICON, showRIds ? Icons.TICK : Icons.CROSS);
        }

        @Override
        public void actionPerformed(ActionEvent e) {
            showRIds = !showRIds;
            putValue(Action.SELECTED_KEY, Boolean.valueOf(showRIds));
            putValue(Action.SMALL_ICON, showRIds ? Icons.TICK : Icons.CROSS);
            component.repaint();
        }
    }

    private final class ShowRoadScaleAction extends AbstractAction {
        public ShowRoadScaleAction() {
            super("show road scale");
            putValue(Action.SELECTED_KEY, Boolean.valueOf(showRoadScale));
            putValue(Action.SMALL_ICON, showRoadScale ? Icons.TICK : Icons.CROSS);
        }

        @Override
        public void actionPerformed(ActionEvent e) {
            showRoadScale = !showRoadScale;
            putValue(Action.SELECTED_KEY, Boolean.valueOf(showRoadScale));
            putValue(Action.SMALL_ICON, showRoadScale ? Icons.TICK : Icons.CROSS);
            component.repaint();
        }
    }


    @Override
    public java.util.List<JMenuItem> getPopupMenuItems() {
        java.util.List<JMenuItem> result = new ArrayList<JMenuItem>();
        result.add(new JMenuItem(showIdsAction));
        result.add(new JMenuItem(showRoadScaleAction));

        return result;
    }
}
