package mrl.world.object;

import javolution.util.FastSet;
import mrl.common.MRLConstants;
import mrl.common.Util;
import mrl.helper.RoadHelper;
import mrl.world.MrlWorld;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.worldmodel.EntityID;

import java.awt.geom.Line2D;
import java.util.*;

/**
 * Created by Mostafa Shabani.
 * Date: 4/27/11
 * Time: 12:28 PM
 */
public class MrlBuilding {
    public double BUILDING_VALUE;
    private boolean probablyOnFire = false;
    private Building selfBuilding;
    private List<MrlBuilding> connectedBuilding;
    private List<Float> connectedValues;
    private Hashtable connectedBuildingsTable;
    private List<EntityID> neighbourIdBuildings;
    private List<EntityID> neighbourFireBuildings;
    private Collection<Wall> walls;
    private double totalWallArea;
    private ArrayList<Wall> allWalls;
    private List<Entrance> entrances;
    private Integer zoneId;
    private double cellCover;
    private boolean visited = false;
    private boolean shouldCheckInside;
    private Set<Human> humans;
    private boolean isReachable;
    private MrlWorld world;
    private int lastUpdateTime;

//    protected List<EntityID> buildingNeighbours = new ArrayList<EntityID>();


    public MrlBuilding(StandardEntity entity, MrlWorld world) {
        selfBuilding = (Building) entity;
        connectedBuildingsTable = new Hashtable(30);
        neighbourIdBuildings = new ArrayList<EntityID>();
        neighbourFireBuildings = new ArrayList<EntityID>();
        connectedBuilding = new ArrayList<MrlBuilding>();
        entrances = new ArrayList<Entrance>();
        humans = new FastSet<Human>();
        this.isReachable = true;
        this.world = world;
        lastUpdateTime = 0;

        initWalls(world);
        initSimulatorValues();
    }

    public void addMrlBuildingNeighbour(MrlBuilding mrlBuilding) {
        allWalls.addAll(mrlBuilding.getWalls());
//        connectedBuilding.add(mrlBuilding);
    }

    public void initWalls(MrlWorld world) {

        int fx = selfBuilding.getApexList()[0];
        int fy = selfBuilding.getApexList()[1];
        int lx = fx;
        int ly = fy;
        Wall w;
        walls = new ArrayList<Wall>();
        allWalls = new ArrayList<Wall>();

        for (int n = 2; n < selfBuilding.getApexList().length; n++) {
            int tx = selfBuilding.getApexList()[n];
            int ty = selfBuilding.getApexList()[++n];
            w = new Wall(lx, ly, tx, ty, this, world.rayRate);
            if (w.validate()) {
                walls.add(w);
                totalWallArea += FLOOR_HEIGHT * 1000 * w.length;
            }
            lx = tx;
            ly = ty;
        }

        w = new Wall(lx, ly, fx, fy, this, world.rayRate);
        walls.add(w);
        totalWallArea = totalWallArea / 1000000d;

    }

    public void initWallValues(MrlWorld world) {

        int totalRays = 0;

        for (Wall wall : walls) {
            wall.findHits(world, this);
            totalRays += wall.rays;
        }
//        int c = 0;
        connectedBuilding = new ArrayList<MrlBuilding>();
        connectedValues = new ArrayList<Float>();
        float base = totalRays;

        for (Enumeration e = connectedBuildingsTable.keys(); e.hasMoreElements(); /*c++*/) {
            MrlBuilding b = (MrlBuilding) e.nextElement();
            Integer value = (Integer) connectedBuildingsTable.get(b);
            connectedBuilding.add(b);
            connectedValues.add(value.floatValue() / base);
//            buildingNeighbours.add(b.getSelfBuilding().getID());
        }

    }

    public List<Entrance> getEntrances() {
        return entrances;
    }

    public void addEntrance(Entrance entrance) {
        this.entrances.add(entrance);
    }

    public void setConnectedBuilding(List<MrlBuilding> connectedBuilding) {
        this.connectedBuilding = connectedBuilding;
    }

    public List<MrlBuilding> getConnectedBuilding() {
        return connectedBuilding;
    }

    public void setConnectedValues(List<Float> connectedValues) {
        this.connectedValues = connectedValues;
    }

    public List<Float> getConnectedValues() {
        return connectedValues;
    }

    public Collection<Wall> getWalls() {
        return walls;
    }

    public Hashtable getConnectedBuildingsTable() {
        return connectedBuildingsTable;
    }

    public ArrayList<Wall> getAllWalls() {
        return allWalls;
    }

    public void setNeighbourIdBuildings(List<EntityID> neighbourIdBuildings) {
        this.neighbourIdBuildings = neighbourIdBuildings;
    }

    public void setNeighbourFireBuildings(List<EntityID> neighbourFireBuildings) {
        this.neighbourFireBuildings = neighbourFireBuildings;
    }

    public List<EntityID> getNeighbourIdBuildings() {
        return neighbourIdBuildings;
    }

    public List<EntityID> getNeighbourFireBuildings() {
        return neighbourFireBuildings;
    }

    public Integer getZoneId() {
        return zoneId;
    }

    public void setZoneId(Integer zoneId) {
        this.zoneId = zoneId;
    }

    public double getCellCover() {
        return cellCover;
    }

    public void setCellCover(double cellCover) {
        this.cellCover = cellCover;
    }

    public boolean shouldCheckInside() {
        return shouldCheckInside;
    }

    public void setShouldCheckInside(boolean shouldCheckInside) {
        this.shouldCheckInside = shouldCheckInside;
    }

    public boolean isBurning() {
        return getEstimatedFieryness() > 0 && getEstimatedFieryness() < 4;
    }

    public double getBuildingRadiation() {
        double value = 0;
//        double totalArea = 0;
        MrlBuilding b;

        for (int c = 0; c < connectedValues.size(); c++) {
            b = connectedBuilding.get(c);
            if (!b.isBurning()) {
                value += (connectedValues.get(c));

            }
        }
        return value * getEstimatedTemperature() / 1000;
    }

    public double getNeighbourRadiation() {
        double value = 0;
//        double totalArea = 0;
        MrlBuilding b;
        int index;

        for (MrlBuilding building : connectedBuilding) {
            if (building.isBurning()) {
                index = building.getConnectedBuilding().indexOf(this);
                if (index >= 0) {
                    value += (building.getConnectedValues().get(index) * building.getEstimatedTemperature());
                }
            }
        }

        return value / 10000;
    }

    public double getBuildingAreaTempValue() {
        return Util.gauss2mf(selfBuilding.getTotalArea() * getEstimatedTemperature(), 10000, 30000, 20000, 40000);
    }

    public boolean isAllEntrancesOpen(MrlWorld world) {
        RoadHelper roadHelper = world.getHelper(RoadHelper.class);
        for (Entrance en : entrances) {
            if (roadHelper.isOpenOrNotSeen(this.getID(), en.getNeighbour().getID()))
                return false;
        }
        return true;
    }

    public boolean isOneEntranceOpen(MrlWorld world) {
        RoadHelper roadHelper = world.getHelper(RoadHelper.class);
//        Building building = world.getEntity(getID(), Building.class);
//        for (EntityID nID : building.getNeighbours()) {
//            StandardEntity entity = world.getEntity(nID);
//            if (entity instanceof Road) {
//                if (roadHelper.isOpenOrNotSeen(this.getID(), entity.getID())) {
//                    return true;
//                }
//            } else {
//                return true;
//            }
//        }
        for (Entrance entrance : getEntrances()) {
            Road road = world.getEntity(entrance.getID(), Road.class);
            for (Building building : entrance.getBuildings()) {
                if (road.getNeighbours().contains(building.getID())) {
                    if (roadHelper.isOpenOrNotSeen(building.getID(), entrance.getID())) {
                        return true;
                    }
                }
            }
        }

        return false;
    }

    public boolean isOneEntranceSurlyOpen(MrlWorld world) {
        RoadHelper roadHelper = world.getHelper(RoadHelper.class);
        for (Entrance en : entrances) {
            if (!roadHelper.isSeenAndBlocked(this.getID(), en.getNeighbour().getID()))
                return true;
        }
        return false;
    }

    public boolean isAllEntrancesBlocked(MrlWorld world) {
        return !isOneEntranceOpen(world);
    }


    ///////////////////////////////////FIRE SIMULATOR PROPERTIES////////////////////////////////////
    static final int FLOOR_HEIGHT = 3;
    static float RADIATION_COEFFICIENT = 0.011f;
    static final double STEFAN_BOLTZMANN_CONSTANT = 0.000000056704;

    private int startTime = -1;
    private float fuel;
    private float initFuel = -1;
    private float volume;
    private double energy;
    private float prevBurned;
    private float capacity;
    private int waterQuantity;
    private boolean wasEverWatered = false;
    private boolean flammable = true;

    public static float woodIgnition = 47;
    public static float steelIgnition = 47;
    public static float concreteIgnition = 47;
    public static float woodCapacity = 1.1f;
    public static float steelCapacity = 1.0f;
    public static float concreteCapacity = 1.5f;
    public static float woodEnergy = 2400;
    public static float steelEnergy = 800;
    public static float concreteEnergy = 350;

    public void initSimulatorValues() {
        volume = selfBuilding.getGroundArea() * selfBuilding.getFloors() * FLOOR_HEIGHT;
        fuel = getInitialFuel();
        capacity = (volume * getThermoCapacity());
        energy = 0;
        initFuel = -1;
        prevBurned = 0;
    }

    public float getInitialFuel() {
        if (initFuel < 0) {
            initFuel = (getFuelDensity() * volume);
        }
        return initFuel;
    }

    private float getThermoCapacity() {
        switch (selfBuilding.getBuildingCode()) {
            case 0:
                return woodCapacity;
            case 1:
                return steelCapacity;
            default:
                return concreteCapacity;
        }
    }

    private float getFuelDensity() {
        switch (selfBuilding.getBuildingCode()) {
            case 0:
                return woodEnergy;
            case 1:
                return steelEnergy;
            default:
                return concreteEnergy;
        }
    }

    public float getIgnitionPoint() {
        switch (selfBuilding.getBuildingCode()) {
            case 0:
                return woodIgnition;
            case 1:
                return steelIgnition;
            default:
                return concreteIgnition;
        }
    }

    public float getConsume(double bRate) {
        if (fuel == 0) {
            return 0;
        }
        float tf = (float) (getEstimatedTemperature() / 1000f);
        float lf = fuel / getInitialFuel();
        float f = (float) (tf * lf * bRate);
        if (f < 0.005f)
            f = 0.005f;
        return getInitialFuel() * f;
    }

    public double getEstimatedTemperature() {
        double rv = energy / capacity;
        if (Double.isNaN(rv)) {
            new RuntimeException().printStackTrace();
        }
        if (rv == Double.NaN || rv == Double.POSITIVE_INFINITY || rv == Double.NEGATIVE_INFINITY)
            rv = Double.MAX_VALUE * 0.75;
        return rv;
    }

    public int getEstimatedFieryness() {
        if (!isFlammable())
            return 0;
        if (getEstimatedTemperature() >= getIgnitionPoint()) {
            if (fuel >= getInitialFuel() * 0.66)
                return 1;   // burning, slightly damaged
            if (fuel >= getInitialFuel() * 0.33)
                return 2;   // burning, more damaged
            if (fuel > 0)
                return 3;    // burning, severly damaged
        }
        if (fuel == getInitialFuel())
            if (wasEverWatered)
                return 4;   // not burnt, but watered-damaged
            else
                return 0;   // not burnt, no water damage
        if (fuel >= getInitialFuel() * 0.66)
            return 5;        // extinguished, slightly damaged
        if (fuel >= getInitialFuel() * 0.33)
            return 6;        // extinguished, more damaged
        if (fuel > 0)
            return 7;        // extinguished, severely damaged
        return 8;           // completely burnt down
    }

    public double getRadiationEnergy() {
        double t = getEstimatedTemperature() + 293; // Assume ambient temperature is 293 Kelvin.
        double radEn = (t * t * t * t) * totalWallArea * RADIATION_COEFFICIENT * STEFAN_BOLTZMANN_CONSTANT;
        if (radEn == Double.NaN || radEn == Double.POSITIVE_INFINITY || radEn == Double.NEGATIVE_INFINITY)
            radEn = Double.MAX_VALUE * 0.75;
        if (radEn > getEnergy()) {
            radEn = getEnergy();
        }
        return radEn;
    }

    public void resetOldPassability(int resetTime) {
        if (world.getTime() - lastUpdateTime > resetTime) {
            setReachable(true);
        }
    }

    public int getRealTemperature() {
        return selfBuilding.getTemperature();
    }

    public Building getSelfBuilding() {
        return selfBuilding;
    }

    public float getVolume() {
        return volume;
    }

    public float getCapacity() {
        return capacity;
    }

    public double getEnergy() {
        return energy;
    }

    public void setEnergy(double v) {
        energy = v;
    }

    public float getPrevBurned() {
        return prevBurned;
    }

    public void setPrevBurned(float consumed) {
        prevBurned = consumed;
    }

    public boolean isFlammable() {
        return flammable;
    }

    public void setFlammable(boolean flammable) {
        this.flammable = flammable;
    }

    public float getFuel() {
        return fuel;
    }

    public void setFuel(float fuel) {
        this.fuel = fuel;
    }

    public int getWaterQuantity() {
        return waterQuantity;
    }

    public void setWaterQuantity(int i) {
        if (i > waterQuantity) {
            wasEverWatered = true;
        }
        waterQuantity = i;
    }

    public void increaseWaterQuantity(int i) {
        waterQuantity += i;
    }

    public int getStartTime() {
        return startTime;
    }

    public void setStartTime(int startTime) {
        this.startTime = startTime;
    }

    public void setWasEverWatered(boolean wasEverWatered) {
        this.wasEverWatered = wasEverWatered;
    }


    // --------------- TEMPORARY SIMULATION --------------------------
    private float tempFuel;
    private double tempEnergy;
    private float tempPrevBurned;
    private boolean tempFlammable = false;


    public void initTempSimulatorValues() {
        tempFuel = fuel;
        tempEnergy = energy;
        tempPrevBurned = prevBurned;
        tempFlammable = flammable;
    }

    public float getTempConsume(double bRate) {
        if (tempFuel == 0) {
            return 0;
        }
        float tf = (float) (getTempEstimatedTemperature() / 1000f);
        float lf = tempFuel / getInitialFuel();
        float f = (float) (tf * lf * bRate);
        if (f < 0.005f)
            f = 0.005f;
        return getInitialFuel() * f;
    }

    public double getTempEstimatedTemperature() {
        double rv = tempEnergy / capacity;
        if (Double.isNaN(rv)) {
            new RuntimeException().printStackTrace();
        }
        if (rv == Double.NaN || rv == Double.POSITIVE_INFINITY || rv == Double.NEGATIVE_INFINITY)
            rv = Double.MAX_VALUE * 0.75;
        return rv;
    }

    public int getTempEstimatedFieryness() {
        if (!isFlammable())
            return 0;
        if (getTempEstimatedTemperature() >= getIgnitionPoint()) {
            if (tempFuel >= getInitialFuel() * 0.66)
                return 1;   // burning, slightly damaged
            if (tempFuel >= getInitialFuel() * 0.33)
                return 2;   // burning, more damaged
            if (tempFuel > 0)
                return 3;    // burning, severly damaged
        }
        if (tempFuel == getInitialFuel())
//            if (wasEverWatered)
//                return 4;   // not burnt, but watered-damaged
//            else
            return 0;   // not burnt, no water damage
        if (tempFuel >= getInitialFuel() * 0.66)
            return 5;        // extinguished, slightly damaged
        if (tempFuel >= getInitialFuel() * 0.33)
            return 6;        // extinguished, more damaged
        if (tempFuel > 0)
            return 7;        // extinguished, severely damaged
        return 8;           // completely burnt down
    }

    public double getTempRadiationEnergy() {
        double t = getTempEstimatedTemperature() + 293; // Assume ambient temperature is 293 Kelvin.
        double radEn = (t * t * t * t) * totalWallArea * RADIATION_COEFFICIENT * STEFAN_BOLTZMANN_CONSTANT;
        if (radEn == Double.NaN || radEn == Double.POSITIVE_INFINITY || radEn == Double.NEGATIVE_INFINITY)
            radEn = Double.MAX_VALUE * 0.75;
        if (radEn > tempEnergy) {
            radEn = tempEnergy;
        }
        return radEn;
    }

    public float getTempFuel() {
        return tempFuel;
    }

    public void setTempFuel(float tempFuel) {
        this.tempFuel = tempFuel;
    }

    public double getTempEnergy() {
        return tempEnergy;
    }

    public void setTempEnergy(double tempEnergy) {
        this.tempEnergy = tempEnergy;
    }

    public float getTempPrevBurned() {
        return tempPrevBurned;
    }

    public void setTempPrevBurned(float tempPrevBurned) {
        this.tempPrevBurned = tempPrevBurned;
    }

    public boolean isTempFlammable() {
        return tempFlammable;
    }

    public void setTempFlammable(boolean tempFlammable) {
        this.tempFlammable = tempFlammable;
    }

    public boolean isPutOff() {
        return getEstimatedFieryness() > 4 && getEstimatedFieryness() < 8;
    }

    public boolean isBurned() {
        return getEstimatedFieryness() == 8;
    }

    public void setProbablyOnFire(boolean probablyOnFire) {
        this.probablyOnFire = probablyOnFire;
    }

    public boolean isProbablyOnFire() {
        return probablyOnFire;
    }

    @Override
    public String toString() {
        return "MrlBuilding[" + selfBuilding.getID().getValue() + "]";
    }

    public EntityID getID() {
        return selfBuilding.getID();
    }

    public boolean isVisited() {
        return visited;
    }

    public void setVisited() {
        this.visited = true;
    }

    public boolean isReachable() {
        return isReachable;
    }

    public void setReachable(boolean reachable) {
        lastUpdateTime = world.getTime();
        isReachable = reachable;
        if (MRLConstants.LAUNCH_VIEWER) {
            if (world.getPlatoonAgent() != null) {
                if (!MrlWorld.BLOCKED_BUILDINGS.containsKey(world.getPlatoonAgent().getID())) {
                    MrlWorld.BLOCKED_BUILDINGS.put(world.getPlatoonAgent().getID(), new HashSet<EntityID>());
                }
                if (reachable) {
                    MrlWorld.BLOCKED_BUILDINGS.get(world.getPlatoonAgent().getID()).remove(getID());
                } else {
                    MrlWorld.BLOCKED_BUILDINGS.get(world.getPlatoonAgent().getID()).add(getID());
                }
            }
        }
    }

    public boolean intersectToLine2D(Line2D line) {
        for (Wall wall : getWalls()) {
            if (wall.getLine().intersectsLine(line))
                return true;
        }
        return false;
    }

//    public MrlBuilding buildingToMrlBuilding( Building building){
//
//        for (Property property:building.getProperties()){
//            this.
//        }
//    }


    /**
     * A set of containing humans
     *
     * @return set of humans
     */
    public Set<Human> getHumans() {
        return humans;
    }
}
