package mrl.world.object;

import javolution.util.FastMap;
import mrl.common.MRLConstants;
import mrl.common.Util;
import mrl.helper.EdgeHelper;
import mrl.helper.RoadHelper;
import mrl.police.MrlPoliceForce;
import mrl.util.PolygonUtil;
import mrl.world.MrlWorld;
import mrl.world.routing.graph.Graph;
import mrl.world.routing.graph.MyEdge;
import mrl.world.routing.graph.Node;
import rescuecore2.misc.Pair;
import rescuecore2.misc.geometry.Line2D;
import rescuecore2.misc.geometry.Point2D;
import rescuecore2.misc.gui.ScreenTransform;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;

import java.util.*;
import java.awt.*;
import java.util.List;

/**
 * Created by: Mahdi Taherian
 * User: mrl
 * Date: 5/17/12
 * Time: 8:25 PM
 */
public class MrlRoad {

    public static Map<Road, Set<Blockade>> VIEWER_ROAD_BLOCKADES = new FastMap<Road, Set<Blockade>>();
    private Map<MrlEdge, HashSet<MrlEdge>> reachableEdges;
    private Set<MrlBlockade> veryImportantBlockades;
    private Set<MrlBlockade> importantBlockades;
    private List<MrlBlockade> mrlBlockades;
    private List<MrlEdge> blockedEdges;
    private List<MrlEdge> passableMrlEdges;
    private Polygon transformedPolygon;
    private List<Point2D> apexPoints;
    private List<MrlRoad> childRoads;
    private HashSet<MrlEdge> openEdges;
    private List<MrlEdge> mrlEdges;
    private boolean isReachable;
    private int totalRepairCost;
    private boolean isPassable;
    private int lastSeenTime;
    private List<Edge> edges;
    private Polygon polygon;
    private MrlWorld world;
    private boolean isSeen;
    private int repairTime;
    private Road parent;

    public MrlRoad(Road road, MrlWorld world, List<MrlEdge> mrlEdges) {
        initialize(road, world, mrlEdges);
    }

    public MrlRoad(Road road, MrlWorld world) {
        this.parent = road;
        initialize(road, world, createMrlEdges(road.getEdges()));
    }

    private void initialize(Road road, MrlWorld world, List<MrlEdge> mrlEdges) {
        this.parent = road;
        this.world = world;
        this.edges = new ArrayList<Edge>(road.getEdges());
        this.apexPoints = new ArrayList<Point2D>();
        this.mrlBlockades = new ArrayList<MrlBlockade>();
        this.childRoads = new ArrayList<MrlRoad>();
        this.blockedEdges = new ArrayList<MrlEdge>();
        this.reachableEdges = new FastMap<MrlEdge, HashSet<MrlEdge>>();
        this.isPassable = true;
        this.isReachable = true;
        this.lastSeenTime = 0;
        this.totalRepairCost = 0;
        this.repairTime = 0;
        this.importantBlockades = new HashSet<MrlBlockade>();
        this.veryImportantBlockades = new HashSet<MrlBlockade>();
        passableMrlEdges = new ArrayList<MrlEdge>();
        for (MrlEdge mrlEdge : mrlEdges) {
            if (mrlEdge.isPassable()) {
                passableMrlEdges.add(mrlEdge);
            }
        }
        setSeen(false);

        setMrlEdges(mrlEdges);
        this.openEdges = new HashSet<MrlEdge>(mrlEdges);
        resetReachableEdges();
    }

    public void update() {
        reset();
        setMrlBlockades();
//        setBlockadesFromViewer();
        for (MrlEdge mrlEdge : mrlEdges) {
            if (mrlEdge.isPassable()) {
                if (mrlEdge.isTooSmall()) {
                    mrlEdge.setBlocked(true);
                }
                mrlEdge.setOpenPart(mrlEdge.getLine());
                List<MrlBlockade> blockedStart = new ArrayList<MrlBlockade>();
                List<MrlBlockade> blockedEnd = new ArrayList<MrlBlockade>();
                for (MrlBlockade mrlBlockade : mrlBlockades) {
                    if (Util.distance(mrlBlockade.getPolygon(), mrlEdge.getStart()) < MRLConstants.AGENT_PASSING_THRESHOLD) {
                        blockedStart.add(mrlBlockade);
                    }
                    if (Util.distance(mrlBlockade.getPolygon(), mrlEdge.getEnd()) < MRLConstants.AGENT_PASSING_THRESHOLD) {
                        blockedEnd.add(mrlBlockade);
                    }
                    setMrlEdgeOpenPart(mrlEdge, mrlBlockade);
                }
                if (mrlBlockades.size() == 1) {
                    if (Util.containsEach(blockedEnd, blockedStart)) {
                        mrlBlockades.get(0).addBlockedEdges(mrlEdge);
                        mrlEdge.setBlocked(true);
                    }
                } else {
                    for (MrlBlockade block1 : blockedStart) {
                        for (MrlBlockade block2 : blockedEnd) {
//                            double distance = Util.distance(block1.getPolygon(), block2.getPolygon());
                            if (Util.isPassable(block1.getPolygon(), block2.getPolygon(), MRLConstants.AGENT_PASSING_THRESHOLD)) {
                                mrlEdge.setBlocked(true);
                                block1.addBlockedEdges(mrlEdge);
                                block2.addBlockedEdges(mrlEdge);
                            }
                        }
                    }
                }
                if (mrlEdge.isBlocked()) {
                    blockedEdges.add(mrlEdge);
                }
                isPassable = getReachableEdges(mrlEdge) != null && !getReachableEdges(mrlEdge).isEmpty();
            } else {
                for (MrlBlockade mrlBlockade : mrlBlockades) {
                    double distance = Util.distance(mrlEdge.getLine(), mrlBlockade.getPolygon());
                    if (distance < MRLConstants.AGENT_PASSING_THRESHOLD) {
                        mrlEdge.setBlocked(true);
                        mrlBlockade.addBlockedEdges(mrlEdge);
                    }
                }
            }
//            boolean isOtherSideBlocked = mrlEdge.isOtherSideBlocked(world);
//            if (isOtherSideBlocked) {
//                mrlEdge.setBlocked(true);
//            }
        }

        //check too small edge passably
        for (MrlEdge mrlEdge : mrlEdges) {
            if (mrlEdge.isPassable() && mrlEdge.isTooSmall()) {
                mrlEdge.setBlocked(true);
            }
        }

        openEdges.removeAll(blockedEdges);
        if (world.getPlatoonAgent() != null && world.getPlatoonAgent() instanceof MrlPoliceForce) {
            updateRepairCost();
//            updateBlockadesValue();
        } else {
            updateNodesPassably();
        }
    }


    private void updateNodesPassably() {
        if (world.getPlatoonAgent() == null)
            return;
        Graph graph = world.getPlatoonAgent().getPathPlanner().getGraph();
        for (MrlEdge mrlEdge : passableMrlEdges) {
            Node node = graph.getNode(mrlEdge.getMiddle());
            if (node == null) {
                continue;
            }

            if (mrlEdge.isBlocked() || mrlEdge.isOtherSideBlocked(world)) {
                node.setPassable(false, world.getTime());
            } else {
                node.setPassable(true, world.getTime());
            }
        }
    }

    private void setMrlEdgeOpenPart(MrlEdge mrlEdge, MrlBlockade mrlBlockade) {
        List<Point2D> pointList = Util.getPoint2DList(mrlBlockade.getPolygon().xpoints, mrlBlockade.getPolygon().ypoints);
        Point2D centerPoint = null;
        boolean isBlockedStart = false, isBlockedEnd = false;
        for (Point2D point : pointList) {
            if (Util.contains(mrlEdge.getLine(), point, 1)) {
                if (point.equals(mrlEdge.getLine().getOrigin())) {
                    isBlockedStart = true;
                } else if (point.equals(mrlEdge.getLine().getEndPoint())) {
                    isBlockedEnd = true;
                } else {
                    centerPoint = point;
                }
            }
        }
        if (centerPoint != null) {
            if (isBlockedEnd && isBlockedStart) {
                //kollan basast!!!
            } else if (isBlockedStart) {
                mrlEdge.setOpenPart(new Line2D(mrlEdge.getOpenPart().getEndPoint(), centerPoint));
            } else if (isBlockedEnd) {
                mrlEdge.setOpenPart(new Line2D(mrlEdge.getOpenPart().getOrigin(), centerPoint));
            }
        }
    }

    public HashSet<MrlEdge> getReachableEdges(MrlEdge from) {
        return reachableEdges.get(from);
    }

    public void addReachableEdges(MrlEdge from, MrlEdge to) {
        reachableEdges.get(from).add(to);
        reachableEdges.get(to).add(from);
    }

    public void removeReachableEdges(MrlEdge from, MrlEdge to) {
        reachableEdges.get(from).remove(to);
        reachableEdges.get(to).remove(from);
    }

    private void setApexPoint() {
        apexPoints.clear();
        for (MrlEdge mrlEdge : mrlEdges) {
            if (mrlEdge == null) {
                System.out.println("(MrlRoad.class ==> mrlEdge == null)");
                continue;
            }
            if (!apexPoints.contains(mrlEdge.getStart()))
                apexPoints.add(mrlEdge.getStart());
            else if (!apexPoints.contains(mrlEdge.getEnd()))
                apexPoints.add(mrlEdge.getEnd());
        }
        createPolygon();
    }

    public List<MrlEdge> getMrlEdgesTo(EntityID neighbourID) {
        List<MrlEdge> mrlEdgeList = new ArrayList<MrlEdge>();
        for (MrlEdge mrlEdge : mrlEdges) {
            if (mrlEdge.isPassable() && mrlEdge.getNeighbours().second().equals(neighbourID)) {
                mrlEdgeList.add(mrlEdge);
            }
        }
        return mrlEdgeList;
    }

    private void createPolygon() {
        int count = apexPoints.size();
        int xs[] = new int[count];
        int ys[] = new int[count];
        for (int i = 0; i < count; i++) {
            xs[i] = (int) apexPoints.get(i).getX();
            ys[i] = (int) apexPoints.get(i).getY();
        }
        polygon = new Polygon(xs, ys, count);
    }

    /**
     * ye polygon migirim bad khat be kahte polygon ro ba in road moghayese mikonim(baraye mohasebeye passably)
     * MrlEdge har khat ro migirim va dakhele yek list mirizim
     * hala ba in liste edge ha ye mrlroad misazim ba parente hamin road...
     *
     * @param polygon polygon on the road that we want to convert to another road(useful for split)
     * @return new mrlRoad
     */
    private MrlRoad convertToRoad(Polygon polygon) {
        List<MrlEdge> mrlEdgeList = new ArrayList<MrlEdge>();
        for (int i = 0; i < polygon.npoints; i++) {
            int j = (i + 1) % polygon.npoints;
            Point2D p1 = new Point2D(polygon.xpoints[i], polygon.ypoints[i]);
            Point2D p2 = new Point2D(polygon.xpoints[j], polygon.ypoints[j]);
            Point2D m1 = Util.getMiddle(p1, p2);
            MrlEdge mrlEdgeTemp = getEdgeInPoint(m1);
            MrlEdge mrlEdge;
            if (mrlEdgeTemp != null)
                mrlEdge = new MrlEdge(mrlEdgeTemp.isPassable(), p1, p2);
            else
                mrlEdge = new MrlEdge(true, p1, p2);
            mrlEdgeList.add(mrlEdge);
        }
        MrlRoad newRoad = new MrlRoad(parent, world, mrlEdgeList);
        return newRoad;
    }

    private void resetReachableEdges() {
        HashSet<MrlEdge> edgesInstead;
        for (MrlEdge mrlEdge : mrlEdges) {
            if (mrlEdge.isPassable()) {
                edgesInstead = new HashSet<MrlEdge>(passableMrlEdges);
                edgesInstead.remove(mrlEdge);
                reachableEdges.put(mrlEdge, edgesInstead);
            }
        }
    }

    /**
     * yek point migirim va beine edge haye roademoon iterate mikonim ta bebinim kodoom MrlEdge in noghte ro dare
     * age hich kodoom in noghte ro nadashtand null bar migardoonim
     *
     * @param point point that we want found edge on it
     * @return MrlEdge which point on it.
     */
    public MrlEdge getEdgeInPoint(Point2D point) {
        for (MrlEdge mrlEdge : mrlEdges) {
            if (Util.contains(mrlEdge.getLine(), point, 1.0)) {
                return mrlEdge;
            }
        }
        return null;
    }

    public MrlEdge getMrlEdge(Edge edge) {
        Point2D middle = Util.getPoint(EdgeHelper.getEdgeMiddle(edge));
        return getEdgeInPoint(middle);
    }

    /**
     * aval polygone road ro migirim bad be vasileye tabe split mikonim nesbat be khat
     * hala 2ta polygon darim
     * har polygon ro tabdil be road mikonim va return mikonim
     *
     * @param line2D khate monas'sef
     * @return 2 mrl roade jadid ba parent_e hamin road
     */
    public Pair<MrlRoad, MrlRoad> splitRoad(rescuecore2.misc.geometry.Line2D line2D) {
        Pair<Polygon, Polygon> splitPolygon = PolygonUtil.split(parent.getApexList(), line2D);
        MrlRoad road1 = convertToRoad(splitPolygon.first());
        MrlRoad road2 = convertToRoad(splitPolygon.second());
        childRoads.add(road1);
        childRoads.add(road2);
        return new Pair<MrlRoad, MrlRoad>(road1, road2);
    }

    private void setMrlEdges(List<MrlEdge> edges) {
        mrlEdges = new ArrayList<MrlEdge>();
        mrlEdges.addAll(edges);
        setApexPoint();
    }

    private List<MrlEdge> createMrlEdges(List<Edge> edges) {
        List<MrlEdge> mrlEdges = new ArrayList<MrlEdge>();
        for (Edge edge : edges) {
            mrlEdges.add(new MrlEdge(edge, parent.getID()));
        }
        return mrlEdges;
    }

    public Polygon getPolygon() {
        return polygon;
    }

    private void addBlockade(MrlBlockade blockade) {
        mrlBlockades.add(blockade);
    }

    private void setBlockadesFromViewer() {
        System.out.println("REMOVE IT(From MrlRoad.class)");
        if (!VIEWER_ROAD_BLOCKADES.containsKey(parent)) {
            return;
        }
        mrlBlockades.clear();
        for (Blockade blockade : VIEWER_ROAD_BLOCKADES.get(parent)) {
            Polygon blockPolygon = PolygonUtil.retainPolygon(getPolygon(), Util.getPolygon(blockade.getApexes()));
            if (blockPolygon == null || blockPolygon.npoints < 3) {
                continue;
            }
            MrlBlockade newBlockade = new MrlBlockade(this, blockade, blockPolygon);
            addBlockade(newBlockade);
        }
    }

    private void setMrlBlockades() {
        totalRepairCost = 0;
        mrlBlockades.clear();
        if (!parent.isBlockadesDefined()) {
            return;
        }
        try {
            for (EntityID blockID : parent.getBlockades()) {
                Blockade blockade = (Blockade) world.getEntity(blockID);
                Polygon blockPolygon = PolygonUtil.retainPolygon(getPolygon(), Util.getPolygon(blockade.getApexes()));
                if (blockPolygon == null) {
                    continue;
                }
                MrlBlockade newBlockade = new MrlBlockade(this, blockade, blockPolygon);
                addBlockade(newBlockade);
            }
        } catch (Exception ex) {
            //TODO: handle it
//            System.err.println("null exception");
        }
    }

    private void updateRepairCost() {
        totalRepairCost = 0;
        for (MrlBlockade mrlBlockade : mrlBlockades) {
            totalRepairCost += mrlBlockade.getRepairCost();
        }
        int repairRate = world.getConfig().getIntValue("clear.repair.rate");
        repairTime = (int) Math.ceil(totalRepairCost / repairRate);
    }

    private void updateBlockadesValue() {
        if (this.getMrlBlockades().size() == 0) {
            return;
        }
        importantBlockades.clear();
        for (int e1 = 0; e1 < passableMrlEdges.size() - 1; e1++) {
            MrlEdge from = getMrlEdges().get(e1);
            for (int e2 = e1; e2 < passableMrlEdges.size(); e2++) {
                MrlEdge to = getMrlEdges().get(e2);
                updateBlockadesValue(from, to);
            }
        }

    }

    private void updateBlockadesValue(MrlEdge from, MrlEdge to) {
        Pair<List<MrlEdge>, List<MrlEdge>> edgesBetween = world.getHelper(RoadHelper.class).getEdgesBetween(this, from, to, false);
        for (MrlBlockade blockade : this.getMrlBlockades()) {
            blockade.setValue(BlockadeValue.WORTHLESS);
            if (blockade.getBlockedEdges().contains(from) || blockade.getBlockedEdges().contains(to)) {
                blockade.setValue(BlockadeValue.VERY_IMPORTANT);
                importantBlockades.add(blockade);
                veryImportantBlockades.add(blockade);
                continue;
            }

            if (Util.containsEach(blockade.getBlockedEdges(), edgesBetween.first()) &&
                    Util.containsEach(blockade.getBlockedEdges(), edgesBetween.second())) {
                blockade.setValue(BlockadeValue.VERY_IMPORTANT);
                veryImportantBlockades.add(blockade);
                importantBlockades.add(blockade);
            }
        }

        for (int i = 0; i < this.getMrlBlockades().size() - 1; i++) {
            List<MrlEdge> blockedEdges = new ArrayList<MrlEdge>();
            MrlBlockade blockade1 = this.getMrlBlockades().get(i);
            if (blockade1.getValue().equals(BlockadeValue.VERY_IMPORTANT))
                continue;
            blockedEdges.addAll(blockade1.getBlockedEdges());
            for (int j = i + 1; j < this.getMrlBlockades().size(); j++) {
                MrlBlockade blockade2 = this.getMrlBlockades().get(j);
                if (blockade2.getValue().equals(BlockadeValue.VERY_IMPORTANT))
                    continue;
                blockedEdges.addAll(blockade2.getBlockedEdges());
//                if (Util.distance(blockade1.getPolygon(), blockade2.getPolygon()) < MRLConstants.AGENT_PASSING_THRESHOLD) {
                if (Util.isPassable(blockade1.getPolygon(), blockade2.getPolygon(), MRLConstants.AGENT_PASSING_THRESHOLD)) {
                    if (Util.containsEach(blockedEdges, edgesBetween.first()) &&
                            Util.containsEach(blockedEdges, edgesBetween.second())) {
                        if (blockade1.getRepairCost() > blockade2.getRepairCost()) {
                            importantBlockades.add(blockade2);
                            blockade1.setValue(BlockadeValue.IMPORTANT_WITH_HIGH_REPAIR_COST);
                            blockade2.setValue(BlockadeValue.IMPORTANT_WITH_LOW_REPAIR_COST);
                        } else {
                            importantBlockades.add(blockade1);
                            blockade1.setValue(BlockadeValue.IMPORTANT_WITH_LOW_REPAIR_COST);
                            blockade2.setValue(BlockadeValue.IMPORTANT_WITH_HIGH_REPAIR_COST);
                        }
                    }
                }
            }
        }

        rescuecore2.misc.geometry.Line2D myEdgeLine = new rescuecore2.misc.geometry.Line2D(from.getMiddle(), to.getMiddle());
        for (MrlBlockade blockade : this.getMrlBlockades()) {
            if (blockade.getValue().equals(BlockadeValue.WORTHLESS)) {
                if (Util.intersections(blockade.getPolygon(), myEdgeLine).size() > 0) {
                    blockade.setValue(BlockadeValue.ORNERY);
                }
            }
        }
    }

    public Set<MrlBlockade> getObstacles(MrlEdge from, MrlEdge to) {
        Set<MrlBlockade> obstacles = new HashSet<MrlBlockade>();
        Pair<List<MrlEdge>, List<MrlEdge>> edgesBetween = world.getHelper(RoadHelper.class).getEdgesBetween(this, from, to, false);
        for (MrlBlockade blockade : this.getMrlBlockades()) {
            if (blockade.getBlockedEdges().contains(from) || blockade.getBlockedEdges().contains(to)) {
                obstacles.add(blockade);
                continue;
            }

            if (Util.containsEach(blockade.getBlockedEdges(), edgesBetween.first()) &&
                    Util.containsEach(blockade.getBlockedEdges(), edgesBetween.second())) {
                obstacles.add(blockade);
            }
        }

        for (int i = 0; i < this.getMrlBlockades().size() - 1; i++) {
            List<MrlEdge> blockedEdges = new ArrayList<MrlEdge>();
            MrlBlockade blockade1 = this.getMrlBlockades().get(i);
            blockedEdges.addAll(blockade1.getBlockedEdges());
            for (int j = i + 1; j < this.getMrlBlockades().size(); j++) {
                MrlBlockade blockade2 = this.getMrlBlockades().get(j);
                blockedEdges.addAll(blockade2.getBlockedEdges());
//                if (Util.distance(blockade1.getPolygon(), blockade2.getPolygon()) < MRLConstants.AGENT_PASSING_THRESHOLD) {
                if (Util.isPassable(blockade1.getPolygon(), blockade2.getPolygon(), MRLConstants.AGENT_PASSING_THRESHOLD)) {
                    if (Util.containsEach(blockedEdges, edgesBetween.first()) &&
                            Util.containsEach(blockedEdges, edgesBetween.second())) {
                        if (blockade1.getRepairCost() > blockade2.getRepairCost()) {
                            obstacles.add(blockade2);
                        } else {
                            obstacles.add(blockade1);
                        }
                    }
                }
            }
        }
        return obstacles;
    }

    /**
     * agar 1 road 1 modat zamane khassi 2bare dide nashod ya payami dar ertebat ba un naresid reset mishe
     * be in soorat ke tamame yalhaye dakhele un + node haye un passable mishand va
     * edge haye un azx halate block kharej mishand
     * <p/>
     * zamane reset shodan bayad az meghdare repairCost/repair_rate (meghdar zamini ke bara pak kardane road lazeme) +
     * yek meghdare threshold baraye etminan be dast miad
     */
    public void resetOldPassably() {
        if (!isSeen() || world.getPlatoonAgent() == null) {
            return;
        }
        if (isTimeToReset()) {
            reset();
        }
    }

    private boolean isTimeToReset() {
        int resetTime = getRepairTime() + MRLConstants.ROAD_PASSABLY_RESET_TIME;
        return world.getTime() - lastSeenTime > resetTime;
    }

    public void reset() {
        blockedEdges.clear();
        openEdges.addAll(mrlEdges);
        mrlBlockades.clear();
        isPassable = true;
        isReachable = true;
        setLastSeenTime(world.getTime());
        if (world.getPlatoonAgent() == null) {
            return;
        }
        Graph graph = world.getPlatoonAgent().getPathPlanner().getGraph();
        for (MrlEdge mrlEdge : mrlEdges) {
            mrlEdge.setBlocked(false);
            Area neighbour = world.getEntity(mrlEdge.getNeighbours().second(), Area.class);
            if (mrlEdge.isPassable()) {
                Node node = graph.getNode((mrlEdge.getMiddle()));
                if (node == null) {
                    System.out.println("node == null in " + this.getID());
                    continue;
                }
                if (neighbour instanceof Road) {
                    MrlRoad mrlRoad = world.getMrlRoad(neighbour.getID());
                    MrlEdge neighbourEdge = mrlRoad.getEdgeInPoint(mrlEdge.getMiddle());
                    if (neighbourEdge != null && !neighbourEdge.isBlocked()) {
                        node.setPassable(true, world.getTime());
                    }
                } else {
                    node.setPassable(true, world.getTime());
                }
            }
        }
        resetReachableEdges();
        for (MyEdge myEdge : graph.getMyEdgesInArea(getID())) {
            myEdge.setPassable(true);
        }
    }

    public List<MrlBlockade> getMrlBlockades() {
        return mrlBlockades;
    }

    public boolean isNeedUpdate() {
        if (parent.isBlockadesDefined() && parent.getBlockades().size() != getMrlBlockades().size()) {
            return true;
        }
        Blockade blockade;
        for (MrlBlockade mrlBlockade : getMrlBlockades()) {
            blockade = mrlBlockade.getParent();
            if (blockade == null || !parent.getBlockades().contains(mrlBlockade.getParent().getID()) || blockade.getRepairCost() != mrlBlockade.getRepairCost()) {
                return true;
            }
        }
        return false;
    }

    public List<MrlEdge> getMrlEdges() {
        return mrlEdges;
    }

    public Road getParent() {
        return parent;
    }

    public List<Edge> getEdges() {
        return edges;
    }

    public List<MrlRoad> getChildRoads() {
        return childRoads;
    }

    public boolean isPassable() {
        return isPassable;
    }

    public List<MrlEdge> getBlockedEdges() {
        return blockedEdges;
    }

    public HashSet<MrlEdge> getOpenEdges() {
        return openEdges;
    }

    public EntityID getID() {
        return parent.getID();
    }

    public boolean isReachable() {
        return isReachable;
    }

    public void setReachable(boolean reachable) {
        isReachable = reachable;
    }

    public void setSeen(boolean seen) {
        this.isSeen = seen;
    }

    public boolean isSeen() {
        return isSeen;
    }

    public void setLastSeenTime(int lastSeenTime) {
        this.lastSeenTime = lastSeenTime;
    }

    public int getLastSeenTime() {
        return lastSeenTime;
    }

    public int getRepairTime() {
        return repairTime;
    }

    public Set<MrlBlockade> getImportantBlockades() {
        return importantBlockades;
    }

    public Set<MrlBlockade> getVeryImportantBlockades() {
        return veryImportantBlockades;
    }

    /**
     * @return transformed polygon just for viewer
     */
    public Polygon getTransformedPolygon() {
        return transformedPolygon;
    }

    /**
     * transform polygon for viewer
     *
     * @param t viewer screen transform
     */
    public void createTransformedPolygon(ScreenTransform t) {

        int count = apexPoints.size();
        int xs[] = new int[count];
        int ys[] = new int[count];
        int i = 0;
        for (Point2D point2D : apexPoints) {
            xs[i] = t.xToScreen(point2D.getX());
            ys[i] = t.yToScreen(point2D.getY());
            i++;
        }
        transformedPolygon = new Polygon(xs, ys, count);
    }
}
