package mrl.world.routing.pathPlanner;

import mrl.common.CommandException;
import mrl.common.MRLConstants;
import mrl.common.Util;
import mrl.helper.RoadHelper;
import mrl.platoon.MrlPlatoonAgent;
import mrl.police.MrlPoliceForce;
import mrl.world.MrlWorld;
import mrl.world.object.*;
import mrl.world.routing.a_star.A_Star;
import mrl.world.routing.graph.Graph;
import mrl.world.routing.graph.Node;
import rescuecore2.misc.Pair;
import rescuecore2.misc.geometry.Line2D;
import rescuecore2.misc.geometry.Point2D;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;

import java.util.ArrayList;
import java.util.List;
import java.awt.*;

/**
 * @author : Mahdi Taherian
 *         Date: 5/12/12
 *         Time: 12:26 PM
 */
public class EscapeHere implements MRLConstants {

    private MrlWorld world;
    private PathPlanner pathPlanner;
    private Graph graph;
    private RoadHelper roadHelper;
    private MrlPlatoonAgent platoonAgent;
    private EntityID lastPositionID;
    private Point lastPositionCoordinate;
    private int stuckThreshold;
    private Area target;
    private A_Star a_star;
    private EntityID nextArea;
    private EntityID position;
    private List<EntityID> planMove;
    private List<Pair<Integer, Integer>> escapeCoordination;
    private EscapeState escapeState = EscapeState.DEFAULT;
    private List<EntityID> newPlanMove;
    private EntityID lastStuckAreaID;
    private EntityID lastStuckRoad;
    private int tryCount;


    public EscapeHere(PathPlanner pathPlanner) {
        this.pathPlanner = pathPlanner;
        this.world = pathPlanner.getWorld();
        this.graph = this.pathPlanner.getGraph();
        this.roadHelper = this.world.getHelper(RoadHelper.class);
        this.platoonAgent = world.getPlatoonAgent();
        this.lastPositionID = world.getSelfPosition().getID();
        this.lastPositionCoordinate = new Point(world.getSelfLocation().first(), world.getSelfLocation().second());
        this.stuckThreshold = 500;
        this.a_star = new A_Star(this.pathPlanner);
        this.planMove = new ArrayList<EntityID>();
        this.escapeCoordination = new ArrayList<Pair<Integer, Integer>>();
        this.planMove = new ArrayList<EntityID>();
        this.lastStuckAreaID = new EntityID(0);
        this.lastStuckRoad = new EntityID(0);
        tryCount = 0;
    }

    private boolean isLockedByBlockade() {
        boolean isLocked = false;
        if (world.getSelfPosition() instanceof Building) {
            return false;
        }
        if (Util.isOnBlockade(world) /*|| isNearBlockade()*/)
            isLocked = true;

        return isLocked;
    }

    public boolean amIMotionLess() {
        return !pathPlanner.getLastMovePlan().contains(world.getSelfPosition().getID()) ||
                setBuried() ||
                escapeState.equals(EscapeState.ESCAPE_BY_COORDINATION) ||
                stuckCondition();
    }

    /**
     * moshakhas mikonad ke aya agent bi harekat boode ya kheir
     *
     * @return true age agent bi harekat bashe...
     */
    private boolean stuckCondition() {
        if (pathPlanner.lastMoveTime < 2)
            return false;
        position = world.getSelfPosition().getID();
//        if (platoonAgent instanceof MrlPoliceForce) {
//            if (world.getSelfPosition() instanceof Road) {
//                return false;
//            }
//        }
        if (pathPlanner.getLastMovePlan().size() > 0) {
            EntityID target = pathPlanner.getLastMovePlan().get(pathPlanner.getLastMovePlan().size() - 1);
            if (target.equals(world.getSelfPosition().getID())) {
                return false;
            }
        }

        Point positionCoordinate = new Point(world.getSelfLocation().first(), world.getSelfLocation().second());
        double moveDistance = Util.distance(lastPositionCoordinate, positionCoordinate);
        if (lastPositionID.equals(position) && moveDistance <= stuckThreshold) {
            return true;
        }
        lastPositionID = position;
        lastPositionCoordinate = positionCoordinate;
        return false;
    }

    public EscapeState escape(Area target) throws CommandException {
        if (escapeState.equals(EscapeState.BURIED))
            return escapeState;
        StandardEntity positionEntity = world.getSelfPosition();
        if (escapeCoordination.size() > 0 && target.equals(this.target)) {
            moveOnCoordination();
        }
        escapeState = EscapeState.DEFAULT;
        this.target = target;
        if (!canIEscapeHere() || tryCount > 10) {
            if (DEBUG_PLANNER) {
                world.printData("I AM STUCK HERE: " + position + " (EscapeHere.class)");
            }
            platoonAgent.setStuck(true);
            escapeState = EscapeState.STUCK;
            return escapeState;
        }
//        platoonAgent.setStuck(false);
        planMove = pathPlanner.getLastMovePlan();
        newPlanMove = new ArrayList<EntityID>(a_star.getShortestPath((Area) positionEntity, target, false));
        if (newPlanMove.isEmpty()) {
            return EscapeState.UNREACHABLE;
        }

        if (planMove.isEmpty() || (newPlanMove.size() == planMove.size() && !newPlanMove.containsAll(planMove))) {
            return EscapeState.TRY_NEW_A_STAR;
        }
        if (lastStuckAreaID.equals(position)) {
            tryCount++;
        } else {
            tryCount = 0;
            lastStuckAreaID = position;
        }
        nextArea = getNextPosition();
        if (positionEntity instanceof Road) {
            escape((Road) positionEntity);
        } else {
            if (positionEntity instanceof Building) {
                escape((Building) positionEntity);
            }
        }
        if (escapeState.equals(EscapeState.ESCAPE_BY_COORDINATION)) {
            moveOnCoordination();
        }

        if (tryCount >= 2) {
            //try to move to target that have no escape result...
            return EscapeState.UNREACHABLE;
        }

        return EscapeState.FAILED;
    }

    public List<EntityID> newPlanMove() {
        return newPlanMove;
    }

    private EntityID getNextPosition() {

        if (planMove.size() == 0)
            return null;
        int posIndex = planMove.indexOf(position);
        if (posIndex + 1 == planMove.size()) {
            return planMove.get(posIndex);
        }
        Area nextArea = world.getEntity(position, Area.class);
        if (nextArea.getNeighbours().contains(this.nextArea)) {
            return planMove.get(posIndex + 1);
        } else {
            return nextArea.getNeighbours().get(0);
        }

    }

    /**
     * agar agent dar dakhele yek building gir karde bood bayad aval be markaze building move bezane bad be noghteei kharej az entrance.
     *
     * @param building building that agent is stuck in it.
     */
    private void escape(Building building) {
        escapeCoordination.clear();
        MrlBuilding mrlBuilding = world.getMrlBuilding(building.getID());
        if (!mrlBuilding.isOneEntranceOpen(world)) {
            escapeState = EscapeState.STUCK;
            return;
        }
        if (nextArea != null) {
            //Add Building Center Coordination
            Point2D escapePoint = Util.getPoint(world.getEntity(position, Area.class).getLocation(world));
            escapeCoordination.add(new Pair<Integer, Integer>((int) escapePoint.getX(), (int) escapePoint.getY()));

            //Add Building Entrance Coordination
            Point2D edgeMiddle = null;
            if (world.getEntity(nextArea, Area.class) instanceof Road) {
                MrlRoad mrlRoad = world.getMrlRoad(nextArea);
                for (MrlEdge mrlEdge : mrlRoad.getMrlEdgesTo(position)) {
                    if (mrlEdge.isPassable() && !mrlEdge.isBlocked() && Util.lineLength(mrlEdge.getOpenPart()) > AGENT_PASSING_THRESHOLD) {
                        edgeMiddle = Util.getMiddle(mrlEdge.getOpenPart());
                        break;
                    }
                }
            }
            if (edgeMiddle == null) {
                Edge edge = world.getEntity(position, Area.class).getEdgeTo(nextArea);
                edgeMiddle = Util.getMiddle(edge.getLine());
            }
            Line2D escapeLine = Util.improveLine(new Line2D(escapePoint, edgeMiddle), MRLConstants.AGENT_SIZE * 2);
            escapePoint = escapeLine.getEndPoint();
            escapeCoordination.add(new Pair<Integer, Integer>((int) escapePoint.getX(), (int) escapePoint.getY()));

            escapeState = EscapeState.ESCAPE_BY_COORDINATION;
        }
    }

    /**
     * agar target dashtim va rahi ham baraye residan be un dashtim bayad betoonim be un beresim
     * hal age blockadi mane az residan be target shod bayad azash oboor konim
     * be in soorat ke ghesmate baze masir ro peida mikonim bad be markaze un move mizanim
     *
     * @param road the road that agent is stuck in it.
     */
    private void escape(Road road) {
        if (target == null) {
            return;
        }
        MrlRoad mrlRoad = world.getMrlRoad(road.getID());
        escapeCoordination.clear();
        int index = planMove.indexOf(mrlRoad.getID());
        if (index < 0 || planMove.size() < 2) {
            return;
        }
        EntityID destinationID = planMove.get(index + 1);
        Node node;
        if (!lastStuckRoad.equals(position)) {
            //agar in roadi ke dar an gir kardim haman area ghabli ke gir kardim nabood
            lastStuckRoad = position;
            for (MrlEdge mrlEdge : mrlRoad.getMrlEdgesTo(destinationID)) {
                node = graph.getNode(mrlEdge.getMiddle());
                if (!mrlEdge.isPassable() || mrlEdge.isBlocked() || node == null || !node.isPassable()) {
                    continue;
                }
                //be center ghesmate baze Edge + agent size be samte jolo move to point mizanim..
                Point2D escapePoint = Util.getMiddle(mrlEdge.getOpenPart());
                Line2D improved = Util.improveLine(new Line2D(Util.getPoint(world.getSelfLocation()), escapePoint), AGENT_SIZE);
                escapeCoordination.add(new Pair<Integer, Integer>((int) improved.getEndPoint().getX(), (int) improved.getEndPoint().getY()));
                escapeState = EscapeState.ESCAPE_BY_COORDINATION;
                return;
            }
        } else {
            if (world.getEntity(destinationID, Area.class) instanceof Road) {

                MrlRoad destinationRoad = world.getMrlRoad(destinationID);
                for (MrlEdge mrlEdge : destinationRoad.getMrlEdgesTo(mrlRoad.getID())) {
                    node = graph.getNode(mrlEdge.getMiddle());
                    if (!mrlEdge.isPassable() || mrlEdge.isBlocked() || node == null || !node.isPassable())
                        continue;
                    //be center ghesmate baze Edge + agent size be samte jolo move to point mizanim..
                    Point2D escapePoint = Util.getMiddle(mrlEdge.getOpenPart());
                    Line2D improved = Util.improveLine(new Line2D(Util.getPoint(world.getSelfLocation()), escapePoint), AGENT_SIZE);
                    escapeCoordination.add(new Pair<Integer, Integer>((int) improved.getEndPoint().getX(), (int) improved.getEndPoint().getY()));
                    escapeState = EscapeState.ESCAPE_BY_COORDINATION;
                    return;
                }
            }
        }
    }

    /**
     * vaghti yek listi az mokhtasat baraye move dashtim in tabe doone doone az aval beheshoon move mizane
     *
     * @throws CommandException move command may throw an command exception
     */
    private void moveOnCoordination() throws CommandException {
        Pair<Integer, Integer> escape = escapeCoordination.remove(0);
        if (escapeCoordination.size() == 0) {
            escapeState = EscapeState.DEFAULT;
        }
        pathPlanner.moveToPoint(position, escape.first(), escape.second());
    }

    /**
     * age agente gheire police buried bood ya rooye blockade oftade bood ya tamame neighbour haye un
     * edge passableshoon be positione feli baste bood, nemitoonim kharej shim...
     *
     * @return true agar ma ghableiate kharej shodan az positione felimoon ro dashte bashim...
     */
    private boolean canIEscapeHere() {
        return platoonAgent instanceof MrlPoliceForce || !isLockedByBlockade() && !isIsolated();
    }

    /**
     * hesab mikonim ke aya in positioni ke toosh hastim neighbour hash baste hast ya kheir
     * age hameye neighbour hash baste boodand migim in position isole hastesh...
     * agar khode road ham tamame edge hash baste bashe isole mishe
     *
     * @return true agar neighboure ghabele oboori dashte bashim
     */
    private boolean isIsolated() {
        return roadHelper.getReachableNeighbours(world.getEntity(position, Area.class)).size() == 0;
    }

    private boolean setBuried() {
        boolean isBuried = (world.getSelfHuman().isBuriednessDefined() && world.getSelfHuman().getBuriedness() > 0);
        if (isBuried)
            escapeState = EscapeState.BURIED;
        return isBuried;
    }

    private boolean isNearBlockade() {
        StandardEntity entity = world.getSelfPosition();
        return entity instanceof Road && getBlockadesNearby(world.getMrlRoad(entity.getID())).size() > 0;
    }

    private List<MrlBlockade> getBlockadesNearby(MrlRoad road) {
        List<MrlBlockade> blockadesNearby = new ArrayList<MrlBlockade>();
        Pair<Integer, Integer> position = world.getSelfLocation();
        double distance;
        if (road.getMrlBlockades().size()>0) {
            for (MrlBlockade mrlBlockade : road.getMrlBlockades()) {
                distance = Util.distance(mrlBlockade.getPolygon(), new rescuecore2.misc.geometry.Point2D(position.first(), position.second()));
                if (distance <= AGENT_SIZE / 4) {
                    blockadesNearby.add(mrlBlockade);
                }
            }
        }
        return blockadesNearby;
    }
}
